/****************************************************************************\
*   BCONSOLE - this source file is part of Binary vIEW project (BIEW)        *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1995, 2000 Nick Kurshev                            *
*   License:              See below                                          *
*   Author and developer: Nick Kurshev                                       *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   bconsole.c                                         *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  This file contain low level BIEW console functions.                       *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/
#include <stdarg.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <limits.h>
#include <time.h>

#include "pmalloc.h"
#include "biewutil.h"
#include "tstrings.h"
#include "bconsole.h"
#include "kbd_code.h"
#include "biewlib.h"

unsigned short Width,Height;
extern TWindow * ErrorWnd;

#define TRUE 1
#define FALSE 0

#define MAXINPUT 79

#define FORMFEED 12

static int KB_Buff[64];
static unsigned char KB_freq = 0;

void initBConsole( unsigned long vio_flg,unsigned long twin_flg )
{
  twInit(vio_flg,twin_flg);
  if(tvioWidth < 80 || tvioHeight < 3)
  {
    printm("Size of video buffer is too small\n");
    twDestroy();
    exit(EXIT_FAILURE);
  }
  Width = tvioWidth;
  Height = tvioHeight;
}

void termBConsole( void )
{
  twDestroy();
}

static int __NEAR__ getkey(int hard)
/* read the next keyboard character */
{
 int key;

 if(KB_freq)  key = KB_Buff[--KB_freq];
 else         key = GetEvent(hard ? hard > 1 ? drawAsmEdPrompt : drawEditPrompt : drawEmptyPrompt,NULL);
 return key;
}

static int __NEAR__ ungotkey(int keycode)
{
  if(KB_freq > sizeof(KB_Buff)/sizeof(int)-1) return 0;
  KB_Buff[KB_freq++] = keycode;
  return 1;
}

/* Allows the user to edit a string with only certain characters allowed -
   Returns TRUE if ESC was not pressed, FALSE is ESC was pressed.
*/

int editstring(char *s,char *legal,int maxlength)
{
  return eeditstring(s,legal,&maxlength,1,NULL,__ESS_ENABLEINSERT,NULL,0,0) != KE_ESCAPE;
}

int xeditstring(char *s,char *legal,int maxlength)
{
  return eeditstring(s,legal,&maxlength,1,NULL,__ESS_ENABLEINSERT,NULL,0,0);
}

static int __NEAR__ isSpace(int val)
{
  return (val+1)%3 ? 0 : 1;
}

static int __NEAR__ isFirstD(int pos) { return isSpace(pos-1); }
static int __NEAR__ isSecondD(int pos) { return isSpace(pos+1); }

static unsigned char insert = UCHAR_MAX;

int eeditstring(char *s, char *legal, int *maxlength,int y,int *stx,int attr,char *undo,char ncol,char hcol)
{
 int c, len = attr & __ESS_HARDEDIT ? *maxlength : attr & __ESS_WANTRETURN && !undo ? ncol : strlen(s), pos = len;
 int lastkey,func_getkeys;
 char ashex = attr & __ESS_ASHEX;
 unsigned char freq = attr & __ESS_HARDEDIT ? 1 : 0;
 if(stx) pos = *stx;
 if(!(attr & __ESS_HARDEDIT))
 {
   if(attr & __ESS_ENABLEINSERT) __vioSetCursorType(insert ? CUR_NORM : CUR_SOLID);
   else                          __vioSetCursorType(CUR_NORM);
 }
 do
 {
  int i;
  Loop:
  if(!(attr & __ESS_NOREDRAW))
  {
    if(!undo) twDirectWrite(1,y,s,len);
    else
    {
      for(i = 0;i < len;i++)
      {
        if(s[i] == undo[i]) twTextColor(ncol);
        else                twTextColor(hcol);
        twDirectWrite(i+1,y,&s[i],1);
      }
    }
    if(!(attr & __ESS_HARDEDIT))
    {
      twGotoXY(len + 1,y);
      if(ashex)
        if(isSecondD(pos) && pos >= len)
        {
          twPutChar('.');
          twGotoXY(len + 2,y);
        }
      for(i = len; i < *maxlength;i++) twPutChar((attr & __ESS_FILLER_7BIT) == __ESS_FILLER_7BIT ? ' ' : '');
    }
  }
  twGotoXY(pos + 1, y);
  func_getkeys = attr & __ESS_HARDEDIT ? attr & 0x0020 ? 2 : 1 : 0;
  new_keycode:
  c = getkey(func_getkeys);
  lastkey = c;
  switch(c)
  {
   case KE_MOUSE:
   case KE_SHIFTKEYS:
                      goto new_keycode;
   case KE_HOME : if(!(ashex && isSecondD(pos))) pos = 0; break;
   case KE_END  : if(!(ashex && isSecondD(pos))) pos = len; break;
   case KE_LEFTARROW : if(pos) { if(!ashex) pos--; else if(!isSecondD(pos)) { pos -=3; if(pos < 0) pos = 0; } }
                       break;
   case KE_RIGHTARROW: if(pos < len) { if(!ashex) pos++; else if(!isSecondD(pos)) { pos += 3; if(pos >= len) pos = len; } }
                       break;
   case KE_BKSPACE   : if (pos > 0 && !(attr & __ESS_HARDEDIT))
                       {
                       memmove(&s[pos-1], &s[pos], len - pos + 1);
                       pos--; len--;
                       if(ashex)
                         if(isSpace(pos) && pos > 2)
                         {
                           memmove(&s[pos-2], &s[pos], len - pos + 2);
                           pos-=2; len-=2;
                         }
                       }
                       break;
   case KE_DEL : if (pos < len && !(attr & __ESS_HARDEDIT) && !ashex)
                    {
                      memmove(&s[pos], &s[pos+1], len - pos);
                      len--;
                    }
                    break;
   case KE_CTL_BKSPACE:
                   if(undo)
                   {
                     if(pos) pos--; if(ashex) if(isSpace(pos) && pos) pos--;
                     s[pos] = undo[pos];
                   }
                   break;
   case KE_INS :    if(attr & __ESS_ENABLEINSERT &&
                       !(attr & __ESS_HARDEDIT) &&
                       !ashex)
                            insert = insert ? 0 : UCHAR_MAX;
                    if(attr & __ESS_ENABLEINSERT) __vioSetCursorType(insert ? CUR_NORM : CUR_SOLID);
                    else                          __vioSetCursorType(CUR_NORM);

                    break;
   case KE_ENTER  : break;
   case KE_ESCAPE : len = 0; break;
   default     :
                  /* FunKey trough int 16 returned XX00 exclude Ctrl0-CtrlZ */
                  if(!(c & 0x00FF) || c == KE_TAB || c == KE_BKSPACE) { c = KE_ENTER; break; }
                  if ( legal == 0 || strchr(legal, c) != NULL )
                  {
                    attr &= ~__ESS_NOREDRAW;
                    if(!freq && !(attr & __ESS_WANTRETURN)) { freq = 1; s[0] = 0; len = pos = 0; ungotkey(c); break; }
                    if(pos < *maxlength)
                    {
                     if(insert)
                     {
                      if(len < *maxlength)
                      {
                        memmove(&s[pos + 1], &s[pos], len - pos + 1);
                        len++;
                      }
                     }
                     else if (pos >= len) len++; if(len > *maxlength) len = *maxlength;
                     s[pos++] = c;
                       if(ashex)
                       {
                         if(isSpace(pos))
                         {
                           if (pos >= len) len++;
                           if(len > *maxlength) len = *maxlength;
                           if(attr & __ESS_HARDEDIT) s[pos] = !((pos + 1) % 12) ? '-' : ' ';
                           else s[pos] = ' ';
                           pos++;
                         }
                         if(isSecondD(pos) && pos < len) s[pos] = '.';
                       }
                    }
                    else s[pos - 1] = c;
                  }
                  break;
  } /* switch */
  if(!freq) freq = 1;
  if(!(attr & __ESS_HARDEDIT)) s[len] = 0;
  if(ashex)
    if(!isFirstD(pos))
      if(c != KE_ESCAPE) goto Loop;
  if(attr & __ESS_WANTRETURN) if(!(ashex && isSecondD(pos))) break;
 }
 while ( (c != KE_ENTER) && (c != KE_ESCAPE));
 if(!(attr & __ESS_HARDEDIT)) __vioSetCursorType(CUR_OFF);
 if(!(attr & __ESS_WANTRETURN && attr & __ESS_NOTUPDATELEN)) *maxlength = len;
 if(stx) *stx = pos;
 return lastkey;
} /* editstring */

TWindow *PleaseWaitWnd( void )
{
   TWindow *w,*usd;
   usd = twUsedWin();
   w = CrtHlpWnd(" Busy ",14,1);
   twGotoXY(1,1); twPutS("Please Wait...");
   twUseWin(usd);
   return w;
}

void MemOutBox(char *user_msg)
{
  ErrMessageBox(user_msg," Not enough memory! ");
}

TWindow *PercentWnd(char *text,char *title)
{
  TWindow *ret,*usd;
  static time_t sttime;
  usd = twUsedWin();
  ret = WindowOpen(1,1,53,6,TWS_FRAMEABLE | TWS_NLS866);
  twCentredWin(ret,NULL);
  twSetColor(Black,LightGray);
  twSetFrame(ret,TW_UP3D_FRAME,Black,LightGray);
  twSetTitle(ret,title,TW_TMODE_CENTER,Black,LightGray);
  twSetFooter(ret," [ ESC ] - Abort ",TW_TMODE_RIGHT,Black,LightGray);
  twClearWin();
  twGotoXY(1,1); twPutS(text);
  twinDraw3DFrame(1,2,52,4,FRAME_DOWN);
  twShowWin(ret);
  twUseWin(usd);
  time(&sttime);
  twSetUsrData(ret,&sttime);
  return ret;
}

int ShowPercentInWnd(TWindow *pw,unsigned percents)
{
  TWindow *usd;
  unsigned cells,remaind,i;
  time_t *sttime,curtime,deltat;
  struct tm *tm;
  char outb[50];
  usd = twUsedWin();
  twUseWin(pw);
  if(percents > 100) percents = 100;
  cells = percents/2;
  remaind = percents%2;
  memset(outb,TW_DEF_FILLER,sizeof(outb));
  for(i = 0;i < cells;i++) outb[i] = '';
  if(remaind) outb[i] = '';
  twDirectWrite(2,3,outb,sizeof(outb));
  sttime = twGetUsrData(pw);
  time(&curtime);
  deltat = curtime - (*sttime);
  tm = localtime(&deltat);
  twGotoXY(1,5);
  twPrintF("Elapsed time: %02u:%02u:%02u",tm->tm_hour,tm->tm_min,tm->tm_sec);
  twUseWin(usd);
  return !IsKbdTerminate();
}

TWindow * WindowOpen(int x1,int y1,int x2,int y2,int flags)
{
  TWindow *ret;
  ret = twCreateWin(x1,y1,x2-x1+1,y2-y1+1,flags);
  if(!ret) { ErrMessageBox("Out of memory",NULL); exit(EXIT_FAILURE); }
  return ret;
}

void CloseWnd(TWindow *w)
{
   twDestroyWin(w);
}

TWindow * CreateWindowDD(char * title,int x2,int y2,char col,char back,char frcol,char frback)
{
 TWindow *win;
 win = WindowOpen(0,0,x2,y2,TWS_FRAMEABLE);
 twCentredWin(win,NULL);
 twSetColor(col,back);
 twClearWin();
 twSetFrame(win,TW_DOUBLE_FRAME,frcol,frback);
 if(title) twSetTitle(win,title,TW_TMODE_CENTER,frcol,frback);
 twShowWin(win);
 return win;
}

TWindow * CreateWindowDDnls(char * title,int x2,int y2,char col,char back,char frcol,char frback)
{
 TWindow *win;
 win = WindowOpen(0,0,x2,y2,TWS_FRAMEABLE | TWS_NLS866);
 twCentredWin(win,NULL);
 twSetColor(col,back);
 twClearWin();
 twSetFrame(win,TW_DOUBLE_FRAME,frcol,frback);
 if(title) twSetTitle(win,title,TW_TMODE_CENTER,frcol,frback);
 twShowWin(win);
 return win;
}

TWindow * CrtHlpWnd( char * title,int width , int height )
{
  return CreateWindowDD(title,width,height,Black,LightGray,Black,LightGray);
}

TWindow * CrtHlpWndnls( char * title,int width , int height )
{
  return CreateWindowDDnls(title,width,height,Black,LightGray,Black,LightGray);
}

static void __NEAR__ __MB(char * text,char * title,char col,char back,char cfr,char bkfr)
{
 unsigned slen,tlen;
 TWindow * prv;
 slen = strlen(text) + 3;
 tlen = strlen(title) + 2;
 slen = min(max(slen,tlen)+1,78);
 twResizeWin(ErrorWnd,slen,3);
 twCentredWin(ErrorWnd,NULL);
 twSetFrame(ErrorWnd,TW_DOUBLE_FRAME,cfr,bkfr);
 twSetTitle(ErrorWnd,title,TW_TMODE_CENTER,cfr,bkfr);
 prv = twUsedWin();
 twUseWin(ErrorWnd);
 twSetColor(col,back);
 twClearWin();
 twShowWin(ErrorWnd);
 twGotoXY(2,1);
 twPutS(text);
 twUseWin(prv);
}

void TMessageBox(char * text,char * title)
{
 __MB(text,title,Black,LightGray,Black,LightGray);
 while(GetEvent(drawEmptyPrompt,ErrorWnd) != KE_ESCAPE);
 twHideWin(ErrorWnd);
 twResizeWin(ErrorWnd,Width,Height);
}

void NotifyBox(char * text,char * title)
{
 if(title == NULL) title = NOTE_MSG;
 __MB(text,title,White,Green,Yellow,Green);
 while(GetEvent(drawEmptyPrompt,ErrorWnd) != KE_ESCAPE);
 twHideWin(ErrorWnd);
 twResizeWin(ErrorWnd,Width,Height);
}

void ShowBox(char * text,char * title)
{
  __MB(text,title,Black,LightGray,Black,LightGray);
}

void ErrMessageBox(char * text,char * title)
{
 if(title == NULL) title = ERROR_MSG;
 __MB(text,title,Yellow,Red,LightRed,Red);
 while(__kbdGetKey(0) != KE_ESCAPE);
 twHideWin(ErrorWnd);
 twResizeWin(ErrorWnd,Width,Height);
}

void errnoMessageBox(char *text,char *title,int __errno__)
{
  char stmp[256];
  sprintf(stmp,"%s: %i (%s)",text,__errno__,strerror(__errno__));
  ErrMessageBox(stmp,title);
}

static void __NEAR__ PaintLine(unsigned i,char *name,unsigned width,int isOrdinal,int useAcc)
{
  char buffer[__TVIO_MAXSCREENWIDTH];
  memset(buffer,TW_DEF_FILLER,sizeof(buffer));
  if(isOrdinal)
  {
    char * endptr;
    endptr = strrchr(name,'');
    if(endptr)
    {
      unsigned len;
      len = endptr - name;
      strncpy((char *)buffer,name,len);
      buffer[width - 6] = '@';
      strncpy((char *)&buffer[width - 5],&name[len + 1],5);
    }
  }
  else strncpy((char *)buffer,name,width);
  if(useAcc)
  {
    char *st,*ends,*ptr;
    char ch;
    twGotoXY(3,i+1);
    st = buffer;
    ends = buffer+width;
    while(1)
    {
      ptr = strchr(st,'~');
      if(ptr)
      {
        int outlen;
        outlen = ptr-st;
        twDirectWrite(twWhereX(),twWhereY(),st,outlen);
        twGotoXY(twWhereX()+outlen,twWhereY());
        st = ptr;
        ch = *(++st);
        if(ch != '~')
        {
          Color col;
          col = twTextColor(Red);
          twPutChar(ch);
          twTextColor(col);
        }
        st++;
      }
      else
      {
        twDirectWrite(twWhereX(),twWhereY(),st,ends-st);
        break;
      }
    }
  }
  else  twDirectWrite(3,i+1,buffer,width);
}

static void __NEAR__ Paint(TWindow *win,char * * names,unsigned nlist,unsigned start,unsigned height,unsigned width,int isOrdinal,int useAcc)
{
 unsigned i;
 twUseWin(win);
 twFreezeWin(win);
 width -= 3;
 twSetColor(Black,LightGray);
 for(i = 0;i < height;i++)
 {
   if(i == 0)
   {
     twGotoXY(1,i + 1);
     if(start)  twPutChar('\x1E');
     else       twPutChar(' ');
   }
   if(i == height - 1)
   {
     twGotoXY(1,i + 1);
     if(start + height < nlist)  twPutChar('\x1F');
     else                        twPutChar(' ');
   }
   twGotoXY(2,i + 1); twPutChar('');
   PaintLine(i,names[i + start],width,isOrdinal,useAcc);
 }
 twRefreshWin(win);
}

extern void drawEmptyPrompt( void );
extern void drawEmptyListPrompt( void );
extern void drawListPrompt( void );
extern void drawOrdListPrompt( void );
extern void drawSearchListPrompt( void );

static char byNam;

typedef char *lpstr;

static int __NEAR__ __searchtext(char *str,char *tmpl,unsigned searchlen,int Case,int Word)
{
  unsigned slen,tlen;
  int __found;
  int cond;
  int start;
  char fbuff[21];
  char cbuff[21],ch,ch1;
  __found = 0;
  slen = strlen(str);
  tlen = searchlen;
  cond = 0;
  for(start = 0;start < slen;start++)
  {
   ch = str[start];
   ch1 = tmpl[0];
   if(!Case) { ch = toupper(ch); ch1 = toupper(ch1); }
   if(ch == ch1)
   {
    if(tlen > 1)
    {
      int tmp;
      tmp = min(tlen - 1,slen-start);
      memcpy(fbuff,&str[start+1],tmp);
      memcpy(cbuff,&tmpl[1],tmp);
      if(!Case)
      {
         memupr((void *)fbuff,tmp);
         memupr((void *)cbuff,tmp);
      }
      if(memcmp(fbuff,cbuff,tmp) == 0) cond = 1;
    }
    else cond = 1;
   }
   if(Word && cond)
   {
     if(start) ch = str[start - 1];
     else      ch = ' ';
     if(start + tlen < slen) ch1 = str[start + tlen];
     else      ch1 = ' ';
     if(!(isseparate(ch) && isseparate(ch1))) cond = 0;
   }
   if(cond) { __found = 1; break; }
  }
  return __found;
}


int listcompare(const void __HUGE__ *v1,const void __HUGE__ *v2)
{
  if(byNam)
  {
    lpstr s1, s2;
    s1 = *((lpstr __HUGE__ *)v1);
    s2 = *((lpstr __HUGE__ *)v2);
    return stricmp(s1,s2);
  }
  else
  {
    lpstr s1,s2;
    lpstr o1,o2;

    s1 = *((lpstr __HUGE__ *)v1);
    s2 = *((lpstr __HUGE__ *)v2);
    o1 = strrchr(s1,'');
    o2 = strrchr(s2,'');
    if(o1 && o2)
    {
      unsigned long ord1,ord2;
      char buff1[7], buff2[7];
       strncpy(buff1,o1 + 1,6);
       strncpy(buff2,o2 + 1,6);
       ord1 = atol(buff1);
       ord2 = atol(buff2);
       if(ord1 < ord2) return -1;
       else if(ord1 == ord2) return 0;
            else return 1;
    }
    else          return stricmp(s1,s2);
  }
}

#define LB_SELECTIVE  0x01
#define LB_SORTABLE   0x02

static int __NEAR__ __ListBox(char * * names,unsigned nlist,unsigned defsel,char * title,int assel)
{
 TWindow * wlist;
 unsigned mwidth = strlen(title);
 unsigned i,j,height;
 char *acctable = 0;
 int ret,width;
 int start,ostart,cursor,ocursor,sf,scursor;
 int isOrdinal;
 isOrdinal = 1;
 scursor = -1;
 if((assel & LB_USEACC) == LB_USEACC)
 {
   acctable = PMalloc(nlist*sizeof(char));
   if(!acctable)
   {
     MemOutBox("Displaying list");
     return -1;
   }
   memset(acctable,0,nlist*sizeof(char));
   for(i = 0;i < nlist;i++)
   {
     unsigned len;
     len = strlen(names[i]);
     for(j = 0;j < len;j++)
     {
       if(names[i][j] == '~' && names[i][j+1] != '~')
       {
         acctable[i] = toupper(names[i][j+1]);
         break;
       }
     }
   }
 }
 if(!strrchr(names[0],'')) isOrdinal = 0;
 if(!isOrdinal) for(i = 0;i < nlist;i++) { j = strlen(names[i]); if(j > mwidth) mwidth = j; }
 else                                    mwidth = tvioWidth-7;
 mwidth += 4;
 if(mwidth > tvioWidth-2) mwidth = tvioWidth-2;
 height = nlist < tvioHeight - 4 ? nlist : tvioHeight - 4;
 wlist = CreateWindowDDnls(title,mwidth-1,height,Black,LightGray,Black,LightGray);
 if((assel & LB_SELECTIVE) == LB_SELECTIVE) twSetFooter(wlist," [ENTER] - Go ",TW_TMODE_RIGHT,White,Black);
 restart:
 ostart = start = cursor = ocursor = 0;
 if(defsel != -1 && defsel < nlist)
 {
    cursor = defsel;
    while(cursor > height) { start += height; cursor -= height; }
    ostart = start;
    ocursor = cursor;
 }
 Paint(wlist,names,nlist,start,height,mwidth,isOrdinal,(assel & LB_USEACC) == LB_USEACC);
 width = mwidth - 3;
 if((assel & LB_SELECTIVE) == LB_SELECTIVE)
 {
   twSetColor(White,Black);
   PaintLine(cursor,names[cursor + start],width,isOrdinal,(assel & LB_USEACC) == LB_USEACC);
 }
 sf = 0;
 for(;;)
 {
   unsigned ch;
   ch = GetEvent(isOrdinal ? drawOrdListPrompt : (assel & LB_SORTABLE) ? drawListPrompt : drawSearchListPrompt,wlist);   if(ch == KE_ESCAPE) { ret = -1; break; }
   if(ch == KE_ENTER)  { ret = start + cursor; break; }
   if(ch != KE_F(7)) scursor = -1;
   switch(ch)
   {
     case KE_F(2):
     case KE_F(3):
              if(isOrdinal || (assel & LB_SORTABLE))
              {
                byNam = ch == KE_F(2);
                if(!isOrdinal && !byNam) break;
                HQSort(names,nlist,sizeof(char *),listcompare);
                goto restart;
              }
              break;
     case KE_F(4): /* save contest to a disk */
              {
                char ofname[256];
                ofname[0] = 0;
                if(GetStringDlg(ofname," Save info to the file : "," [ENTER] - Proceed ",NAME_MSG))
                {
                  FILE * out;
                  out = fopen(ofname,"wt");
                  if(out)
                  {
                    unsigned i;
                    int j;
                    strncpy(ofname,title,sizeof(ofname));
                    ofname[sizeof(ofname)-1] = '\0';
                    if(GetStringDlg(ofname," User comments : "," [ENTER] - Proceed "," Description : "))
                    {
                      fprintf(out,"%s\n\n",ofname);
                    }
                    for(i = 0;i < nlist;i++)
                    {
                      char *p;
                      p = strchr(names[i],'');
                      if(p)
                      {
                        *p = 0;
                        fprintf(out,names[i]);
                        for(j = p - names[i];j < 50;j++) fprintf(out," ");
                        fprintf(out," @%s",p+1);
                        if(p) *p = '';
                      }
                      else fprintf(out,names[i]);
                      fprintf(out,"\n");
                    }
                    fclose(out);
                  }
                  else errnoMessageBox("Can not access file",NULL,errno);
                }
              }
              break;
     case KE_F(7): /* perform binary search in list */
             {
               static char searchtxt[21] = "";
               static unsigned char searchlen = 0;
               static unsigned char CaseSens = 0,WordOnly = 0,Backwarding = 0;
               if(SearchDialog(searchtxt,&searchlen,&CaseSens,&WordOnly,&Backwarding))
               {
                  int direct,found;
                  int i;
                  int endsearch,startsearch;
                  searchtxt[searchlen] = 0;
                  endsearch = Backwarding ? -1 : nlist;
                  direct = Backwarding ? -1 : 1;
                  startsearch = (assel & LB_SELECTIVE) == LB_SELECTIVE ?
                                 cursor + start :
                                 scursor != -1 ?
                                   scursor :
                                   start;
                  if(startsearch > (int)(nlist-1)) startsearch = nlist-1;
                  if(startsearch < 0) startsearch = 0;
                  if((assel & LB_SELECTIVE) == LB_SELECTIVE || scursor != -1)
                  {
                    Backwarding ? startsearch-- : startsearch++;
                  }
                  found = 0;
                  for(i = startsearch;i != endsearch;i+=direct)
                  {
                     if(__searchtext(names[i],searchtxt,searchlen,CaseSens,WordOnly))
                     {
                        start = scursor = i;
                        if(start > nlist - height) start = nlist - height;
                        ostart = start - 1;
                        if((assel & LB_SELECTIVE) == LB_SELECTIVE) cursor = scursor - start;
                        found = 1;
                        break;
                     }
                  }
                  if(!found) scursor = -1;
                  if(scursor == -1) ErrMessageBox(STR_NOT_FOUND,SEARCH_MSG);
               }
             }
             break;
     case KE_DOWNARROW : if((assel & LB_SELECTIVE) == LB_SELECTIVE) cursor ++; else start ++; break;
     case KE_UPARROW   : if((assel & LB_SELECTIVE) == LB_SELECTIVE) cursor --; else start--; break;
     case KE_PGDN   : start += height; break;
     case KE_PGUP   : start -= height; break;
     case KE_CTL_PGDN  : start = nlist - height; cursor = height; break;
     case KE_CTL_PGUP  : start = cursor = 0; break;
     default :
              /* Try accelerate choose */
              if((assel & LB_USEACC) == LB_USEACC)
              {
                 if((unsigned char)(ch & 0x00FF) > 31)
                 {
                   ch = toupper(ch & 0x00FF);
                   for(i = 0;i < nlist;i++)
                   {
                     if(ch == acctable[i]) { ret = i; goto Done; }
                   }
                 }
              }
   }
   if((assel & LB_SELECTIVE) == LB_SELECTIVE)
   {
     if(cursor < 0) { cursor = 0; start--; }
     if(cursor > height - 1) { cursor = height - 1; start++; }
   }
   if(start < 0) start = 0;
   if(start > nlist - height) start = nlist - height;
   if(start != ostart)
   {
     ostart = start;
     Paint(wlist,names,nlist,start,height,mwidth,isOrdinal,(assel & LB_USEACC) == LB_USEACC);
     sf = 1;
   }
   if((cursor != ocursor || sf) && (assel & LB_SELECTIVE) == LB_SELECTIVE)
   {
     twSetColor(Black,LightGray);
     PaintLine(ocursor,names[ocursor + start],width,isOrdinal,(assel & LB_USEACC) == LB_USEACC);
     twSetColor(White,Black);
     PaintLine(cursor,names[cursor + start],width,isOrdinal,(assel & LB_USEACC) == LB_USEACC);
     ocursor = cursor;
     sf = 0;
   }
   if(scursor >= 0)
   {
     twSetColor(Black,Green);
     if(scursor >= start && scursor < start + height) PaintLine(scursor - start,names[scursor],width,isOrdinal,(assel & LB_USEACC) == LB_USEACC);
   }
 }
 Done:
 CloseWnd(wlist);
 if(acctable) PFREE(acctable);
 return ret;
}

int CommonListBox(char * *names,unsigned nlist,char *title,int acc,unsigned defsel)
{
  return __ListBox(names,nlist,defsel,title,acc);
}

void DisplayBox(char * *names,unsigned nlist,char *title)
{
  __ListBox(names,nlist,-1,title,0); /* not sortable & not selective */
}

void ListBox(char * * names,unsigned nlist,char * title)
{
  __ListBox(names,nlist,-1,title,LB_SORTABLE);
}

int SelListBox(char * * names,unsigned nlist,char * title,unsigned defsel)
{
  return __ListBox(names,nlist,defsel,title,LB_SELECTIVE | LB_SORTABLE);
}

int SelBox(char * * names,unsigned nlist,char * title,unsigned defsel)
{
  return __ListBox(names,nlist,defsel,title,LB_SELECTIVE);
}

int SelBoxA(char * * names,unsigned nlist,char * title,unsigned defsel)
{
  return __ListBox(names,nlist,defsel,title,LB_SELECTIVE | LB_USEACC);
}

int PageBox(int width,int height,void * * __obj,unsigned nobj,pagefunc func)
{
 TWindow * wlist;
 int start,ostart,ret;
 wlist = CreateWindowDD(0,width-1,height,Black,LightGray,Black,LightGray);
 ostart = start = 0;
 (*func)(wlist,__obj,start,nobj);
 for(;;)
 {
   unsigned ch;
   ch = GetEvent(drawEmptyPrompt,wlist);
   if(ch == KE_ESCAPE) { ret = -1; break; }
   if(ch == KE_ENTER) { ret = start; break; }
   switch(ch)
   {
     case KE_PGDN : start ++; break;
     case KE_PGUP   : start --; break;
     case KE_CTL_PGDN : start = nobj - 1; break;
     case KE_CTL_PGUP : start = 0; break;
     default : break;
   };
   if(start < 0) start = 0;
   if(start > nobj - 1) start = nobj - 1;
   if(start != ostart)
   {
     ostart = start;
     (*func)(wlist,__obj,start,nobj);
   }
 }
 CloseWnd(wlist);
 return ret;
}




