/****************************************************************************\
*   BIEWHELP - this source file is part of Binary vIEW project (BIEW)        *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1995, 2000 Nick Kurshev                            *
*   License:              See below                                          *
*   Author and developer: Nick Kurshev                                       *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   biewhelp.c                                         *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  This file contain LZSS-based help system functions.                       *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/
#include <string.h>
#include <stdlib.h>
#include <limits.h>
#include <time.h>
#include <stdio.h>
#include <errno.h>
#if __BITNESS__ == 16
#include <alloc.h>
#endif

#include "pmalloc.h"
#include "bmfile.h"
#include "bconsole.h"
#include "biewhelp.h"
#include "biewlib.h"
#include "tools/lzss/lzssutil.c"
#include "setup.h"
static BGLOBAL bHelp = NULL;
static BIEW_HELP_ITEM bhi;

int hlpOpen( int interact )
{
  char *help_name;
  char hlp_id[sizeof(BIEW_HELP_VER)];
  help_name = biewGetHelpName();
  bHelp = biewOpen(help_name,BBIO_SMALL_CACHE_SIZE);
  if(bHelp == &bNull)
  {
    if(interact) errnoMessageBox("Can not open help file",NULL,errno);
    return 0;
  }
  bioSetOptimization(bHelp,BIO_OPT_RANDOM);
  bioSeek(bHelp,0L,SEEK_SET);
  bioReadBuffer(bHelp,hlp_id,sizeof(hlp_id));
  if(memcmp(hlp_id,BIEW_HELP_VER,sizeof(BIEW_HELP_VER)) != 0)
  {
    if(interact)
    {
      ErrMessageBox("Wrong version of help",NULL);
    }
    bioClose(bHelp);
    bHelp = &bNull;
    return 0;
  }
  return 1;
}

void hlpClose( void )
{
  if(bHelp != &bNull && bHelp != NULL) bioClose(bHelp);
}

static int find_item(unsigned long item_id)
{
  unsigned long i,nsize,lval;
  char sout[HLP_SLONG_LEN];
  bioSeek(bHelp,HLP_VER_LEN,SEEK_SET);
  bioReadBuffer(bHelp,sout,sizeof(sout));
  nsize = strtoul(sout,NULL,16);
  for(i = 0;i < nsize;i++)
  {
    bioReadBuffer(bHelp,&bhi,sizeof(BIEW_HELP_ITEM));
    lval = strtoul(bhi.item_id,NULL,16);
    if(lval == item_id) return 1;
  }
  return 0;
}

unsigned long   hlpGetItemSize(unsigned long item_id)
{
  if(find_item(item_id))
  {
    unsigned long ret;
    ret = strtoul(bhi.item_decomp_size,NULL,16);
    return ret;
  }
  else
  {
    ErrMessageBox("Find: Item not found",NULL);
  }
  return 0;
}

int             hlpLoadItem(unsigned long item_id, void __HUGE__* buffer)
{
  unsigned long hlp_off,hlp_size;
  if(bHelp == &bNull) return 0;
  if(find_item(item_id))
  {
    hlp_off = strtoul(bhi.item_off,NULL,16);
    hlp_size = strtoul(bhi.item_length,NULL,16);
    if(!Decode(bHelp,buffer,hlp_off,hlp_size))
    {
      MemOutBox("Help uncompression");
      return 0;
    }
    return 1;
  }
  ErrMessageBox("Load: Item not found",NULL);
  return 0;
}

void hlpDisplay( unsigned long item_id )
{
  char **str_ptr = 0;
  char __HUGE__ *data = 0;
  char *title;
  char ch,ch1;
  unsigned long data_size,i;
  unsigned nstr;
  if(!hlpOpen(1)) return;
  data_size = hlpGetItemSize(item_id);
  if(!data_size) goto hlp_bye;
  data = PHMalloc(data_size+1);
  if(!data)
  {
    mem_off:
    MemOutBox("Loading help");
    if(data) PHFREE(data);
    if(str_ptr) PFREE(str_ptr);
    goto hlp_bye;
  }
  nstr = 0;
  data[data_size] = 0;
  if(hlpLoadItem(item_id,data))
  {
    if(!(str_ptr = PMalloc(sizeof(char *)))) goto mem_off;
    title = &data[0];
    for(i = 0;i < data_size;i++)
    {
      ch = data[i];
      if(ch == '\n' || ch == '\r')
      {
        data[i] = 0;
        if(!(str_ptr = PRealloc(str_ptr,(nstr+1)*sizeof(char *)))) goto mem_off;
        ch1 = data[i+1];
        if((ch1 == '\n' || ch1 == '\r') && ch != ch1) ++i;
        str_ptr[nstr++] = &data[i+1];
      }
      if(nstr > UINT_MAX-2) goto mem_off;
    }
    CommonListBox(str_ptr,nstr,title,0,-1);
    PFREE(str_ptr);
  }
  PHFREE(data);
  hlp_bye:
  hlpClose();
}

