/****************************************************************************\
*   BIEWUTIL - this header file is part of Binary vIEW project (BIEW)        *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1995, 2000 Nick Kurshev                            *
*   License:              See below                                          *
*   Author and developer: Nick Kurshev                                       *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   biewutil.h                                         *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  This file contain prototypes of BIEW utilities.                           *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/

#ifndef __BIEWUTIL__H
#define __BIEWUTIL__H

#ifdef __cplusplus
extern "C" {
#endif

#ifndef __TWIN_H
#include "twin.h"
#endif

#ifndef __SYS_DEP_H
#include "_sys_dep.h"
#endif

#ifndef __BBIO_H
#include "bbio.h"
#endif

#ifndef __FILE_INI_RUNTIME_SUPPORT_SYSTEM__
#include "file_ini.h"
#endif

#include "biewtype.h"

#define NOT_EXE 0
#define DOS_SYS 1
#define OLD_EXE 2
#define NEW_EXE 3

typedef struct tagDOSDRIVER
{
  tWord ddAttribute;
  tWord ddStrategyOff;
  tWord ddInterruptOff;
  char         ddName[8];
}DOSDRIVER;

typedef struct tagMZHEADER
{
 tWord mzPartLastPage;
 tWord mzPageCount;
 tWord mzRelocationCount;
 tWord mzHeaderSize;
 tWord mzMinMem;
 tWord mzMaxMem;
 tWord mzRelocationSS;
 tWord mzExeSP;
 tWord mzCheckSumm;
 tWord mzExeIP;
 tWord mzRelocationCS;
 tWord mzTableOffset;
 tWord mzOverlayNumber;
}MZHEADER;

typedef union tagDOSHDR
{
  DOSDRIVER                 drv;
  MZHEADER                  mz;
}DOSHDR;

extern DOSHDR oldh;

extern char legalchars[];
extern char FType;
extern unsigned long headshift;
extern int DumpMode;
extern unsigned char EditMode;

extern unsigned long lastbyte;
extern char * ini_name;
extern unsigned long OldCurrFilePos; /* means previous File position */

extern int                GetBool(int _bool);
extern void **            FAllocPtrPtr(unsigned num);
extern void *             FAllocPtr(unsigned size);
extern void               FFreeArr(void **arr,unsigned n);
extern void               CriticalExit(int code);

extern void               SelectTool( void );
extern void               SelectSysInfo( void );
extern int                isValidIniArgs( void ); /* return !0 if LastOpenFile == Current open file */

                                         /* convert space characters to tabs */
extern void               Space2Tab(char *dest,const char *src);
extern void               TrimTrailingSpace(char *str);

extern unsigned long      IsNewExe(void);

extern char * Get2Digit(unsigned char);
extern char * Get2SignDig(signed char);
extern char * Get4Digit(unsigned short);
extern char * Get4SignDig(signed short);
extern char * Get8Digit(unsigned long);
extern char * Get8SignDig(signed long);
extern char * Get2DigitS(unsigned char *);
extern char * Get4DigitS(unsigned char *);
extern char * Get8DigitS(unsigned char *);
extern char * GetBinary(char val);

extern int    ExpandHex(char * dest,const unsigned char * src,int size,char hard);
extern void   CompressHex(unsigned char * dest,const char * src,int sizedest,int usespace);
extern void   ExtHelp(void);
extern int    Summ(unsigned char *array,unsigned size);
extern void   drawPrompt( void );
extern void   About( void );

extern void  ShowSysInfo( void );
extern unsigned long ShowSysHeader( void );
extern unsigned long ShowMZHeader( void );

extern void  PaintTitle( void );
extern void  MainLoop( void );

extern unsigned long FoundTextSt,FoundTextEnd;
extern int  isHOnLine(unsigned long cp,int width);

#define HLS_NORMAL               0x0000
#define HLS_USE_DOUBLE_WIDTH     0x0001
#define HLS_USE_BUFFER_AS_VIDEO  0x0002
extern void HiLightSearch(TWindow *out,unsigned long cfp,char minx,
                          char maxx,char y,char *buff,unsigned flags);

/* Class array */

typedef struct tag_memArray
{
  unsigned nItems;
  void **  data;
  unsigned nSize;
}memArray;

extern memArray *ma_Build( int maxitems, int interact );
extern int       ma_AddString(memArray *obj,char *data,int interact);
extern int       ma_AddData(memArray *obj,void *data,unsigned size,int interact);
extern void      ma_Destroy(memArray *obj);
extern int       ma_Display(memArray *obj,char *title,int flg,unsigned defsel);

/* Class linear huge array */
typedef struct tag_linearArray
{
  unsigned long   nItems;
  void __HUGE__ * data;
  unsigned long   nSize;
  unsigned        itemSize;
}linearArray;

extern linearArray *la_Build( unsigned long maxitems,unsigned size_of_item,
                              int interact );
extern int          la_AddData(linearArray *obj,void *data,int interact);
extern void         la_Destroy(linearArray *obj);
extern void         la_Sort(linearArray *obj,
                            int (* compare)(const void __HUGE__ *e1, const void __HUGE__ *e2));
extern void __HUGE__ *la_Find(linearArray *obj,
                            const void *key,
                            int (* compare)(const void __HUGE__ *e1, const void __HUGE__ *e2));
extern unsigned long la_FindNearest(linearArray *obj,
                            const void *key,
                            int (* compare)(const void __HUGE__ *e1, const void __HUGE__ *e2));

extern unsigned biewReadProfileString(hIniProfile *ini,
                                      const char *section,
                                      const char *subsection,
                                      const char *_item,
                                      const char *def_value,
                                      char *buffer,
                                      unsigned cbBuffer);

extern int biewWriteProfileString(hIniProfile *ini,
                                  const char *section,
                                  const char *subsection,
                                  const char *item,
                                  const char *value);

#ifdef __cplusplus
}
#endif

#endif


