/****************************************************************************\
*   BIN_UTIL - this source file is part of Binary vIEW project (BIEW)        *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1995, 2000 Nick Kurshev                            *
*   License:              See below                                          *
*   Author and developer: Nick Kurshev                                       *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   bin_util.c                                         *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  This file contain common functions of plugins/bin of BIEW project.        *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/
#include <limits.h>
#include <stdio.h>
#include <string.h>
#include <stddef.h>
#include <errno.h>
#include <stdlib.h>

#include "bin_util.h"
#include "reg_form.h"
#include "biewutil.h"
#include "bmfile.h"
#include "bconsole.h"
#include "tstrings.h"

unsigned fmtActiveState = 0;

linearArray *PubNames = NULL;

void fmtSetState(int state)
{
   if(state == PS_ACTIVE) { if(fmtActiveState < UINT_MAX) fmtActiveState++; }
   else  { if(fmtActiveState) fmtActiveState--; }
}

int fmtComparePubNames(const void __HUGE__ *v1,const void __HUGE__ *v2)
{
  struct PubName __HUGE__ *pnam1,__HUGE__ *pnam2;
  pnam1 = (struct PubName __HUGE__ *)v1;
  pnam2 = (struct PubName __HUGE__ *)v2;
  if(pnam1->pa < pnam2->pa) return -1;
  else
    if(pnam1->pa > pnam2->pa) return 1;
    else                      return 0;
}

int fmtFindPubName(BGLOBAL fmt_cache,char *buff,unsigned cb_buff,
                   unsigned long pa,
                   ReadPubNameList fmt_readlist,
                   ReadPubName fmt_readpub)
{
  struct PubName *ret,key;
  key.pa = pa;
  if(!PubNames) (*fmt_readlist)(fmt_cache,1);
  ret = la_Find(PubNames,&key,fmtComparePubNames);
  if(ret)
  {
    (*fmt_readpub)(fmt_cache,ret->nameoff,buff,cb_buff);
    return 1;
  }
  return 0;
}

unsigned long fmtGetPubSym(BGLOBAL fmt_cache,char *str,unsigned cb_str,
                           unsigned *func_class,unsigned long pa,int as_prev,
                           ReadPubNameList fmt_readlist,
                           ReadPubName fmt_readpub)
{
  unsigned long cfpos,ret_addr;
  unsigned i,idx;
  struct PubName key;
  cfpos = BMGetCurrFilePos();
  if(!PubNames) (*fmt_readlist)(fmt_cache,0);
  if(!PubNames->nItems) return 0;
  ret_addr = 0L;
  idx = UINT_MAX;
  *func_class = SC_GLOBAL;
  key.pa = pa;
  i = (unsigned)la_FindNearest(PubNames,&key,fmtComparePubNames);
  if(as_prev) idx = i;
  else
  {
    while(((struct PubName __HUGE__ *)PubNames->data)[i].pa <= pa) i++;
    idx = i;
  }
  if(idx < PubNames->nItems);
  {
    ret_addr = ((struct PubName __HUGE__ *)PubNames->data)[idx].pa;
    if(!idx && pa < ret_addr && as_prev)
    {
      ret_addr = 0;
    }
    else
      (*fmt_readpub)(fmt_cache,
                     ((struct PubName __HUGE__ *)PubNames->data)[idx].nameoff,
                     str,cb_str);
      str[cb_str-1] = 0;
  }
  BMSeek(cfpos,BIO_SEEK_SET);
  return ret_addr;
}

static BGLOBAL __NEAR__ ReopenSeek(unsigned long dist)
{
 BGLOBAL handle;
 char sout[256];
 handle = bioDup(BMbioHandle());
 if(handle != &bNull) bioSeek(handle,dist,BIO_SEEK_SET);
 else
 {
    sprintf(sout,"Can not read file: error code = %u (%s)\n",errno, strerror(errno));
    ErrMessageBox(sout,NULL);
 }
 return handle;
}

int fmtShowList( GetNumItems gni,ReadItems ri,char * title,int flags,unsigned * ordinal)
{
 int ret,bool;
 BGLOBAL handle;
 unsigned nnames;
 memArray * obj;
 TWindow* w;
 ret = -1;
 if((handle = ReopenSeek(0)) == &bNull) return ret;
 nnames = gni ? (*gni)(handle) : -1;
 if(!(obj = ma_Build(nnames,1))) goto exit;
 w = PleaseWaitWnd();
 bool = (*ri)(handle,obj,nnames);
 CloseWnd(w);
 if(bool)
 {
   if(!obj->nItems) { NotifyBox(NOT_ENTRY,title); goto exit; }
   if(flags)
   {
     ret = ma_Display(obj,title,flags,-1);
     if(ordinal && ret != -1)
     {
       char * cptr;
       char buff[10];
       cptr = strrchr(obj->data[ret],'');
       cptr++;
       strcpy(buff,cptr);
       *ordinal = atoi(buff);
     }
   }
   else    { ret = -1; ma_Display(obj,title,LB_SORTABLE,-1); }
 }
 ma_Destroy(obj);
 exit:
 bioClose(handle);
 return ret;
}
