/****************************************************************************\
*   CODEGUID - this source file is part of Binary vIEW project (BIEW)        *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1995, 2000 Nick Kurshev                            *
*   License:              See below                                          *
*   Author and developer: Nick Kurshev                                       *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   codeguid.c                                         *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  This file contain code navigation routines.                               *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/
#include <string.h>
#include <stdio.h>
#include <limits.h>

#include "kbd_code.h"
#include "bmfile.h"
#include "biewutil.h"
#include "disasm.h"
#include "bconsole.h"
#include "codeguid.h"
#include "reg_form.h"
#include "__os_dep.h"
#include "pmalloc.h"

#define BACK_ADDR_SIZE 256
#define GO_ADDR_SIZE   37

static unsigned long *BackAddr;
static int BackAddrPtr = -1;
static unsigned long *GoAddr;
static unsigned int  *GoLineNums;
static int GoAddrPtr = -1;
static unsigned char Alarm = 0;

extern int DisasmCurrLine;

static char image[] = "=>[X]";

int initCodeGuider( void )
{
  BackAddr = PMalloc(sizeof(unsigned long)*BACK_ADDR_SIZE);
  if(!BackAddr) return 0;
  GoAddr = PMalloc(sizeof(unsigned long)*GO_ADDR_SIZE);
  if(!GoAddr) { PFree(BackAddr); return 0; }
  GoLineNums = PMalloc(sizeof(unsigned long)*GO_ADDR_SIZE);
  if(!GoLineNums) { PFree(BackAddr); PFree(GoAddr); return 0; }
  return 1;
}

void termCodeGuider( void )
{
  PFree(GoLineNums);
  PFree(GoAddr);
  PFree(BackAddr);
}

static char * __NEAR__ gidBuildKeyStr( void )
{
  if(GoAddrPtr < 10) image[3] = GoAddrPtr + '0';
  else               image[3] = GoAddrPtr - 10 + 'A';
  return image;
}

void GidResetGoAddress( int keycode )
{
  Alarm = 0;
  if(keycode == KE_DOWNARROW)
  {
    int i;
    if(GoAddrPtr >= 0)
    {
      if(GoLineNums[0] == 0)
      {
        for(i = 1;i <= GoAddrPtr;i++)
        {
          GoAddr[i - 1] = GoAddr[i];
          GoLineNums[i - 1] = GoLineNums[i];
        }
        GoAddrPtr--;
      }
      for(i = 0;i <= GoAddrPtr;i++)
      {
        char dig;
        GoLineNums[i]--;
        if(i < 10) dig = i + '0';
        else       dig = i - 10 + 'A';
        twDirectWrite(tvioWidth-1,GoLineNums[i]+2,&dig,1);
      }
    }
  }
  else if(keycode == KE_UPARROW)
       {
          int i;
          Alarm = UCHAR_MAX;
          if(GoAddrPtr >= 0)
          {
            for(i = 0;i <= GoAddrPtr;i++) GoLineNums[i]++;
            if(GoLineNums[GoAddrPtr] >= Height)  GoAddrPtr--;
          }
       }
       else  GoAddrPtr = -1;
}

extern int disPanelMode;
extern char DisasmPrepareMode;

void GidAddGoAddress(char *str,unsigned long addr)
{
  int len,where;
  if(DisasmPrepareMode) return;
  len = strlen((char *)str);
  where = (disPanelMode == PANMOD_FULL ? tvioWidth :
           disPanelMode == PANMOD_MEDIUM ? tvioWidth-10 : tvioWidth-41) - 5;
  if(Alarm)
  {
     int i;
      if(GoAddrPtr < GO_ADDR_SIZE - 2) GoAddrPtr++;
      memmove(&GoAddr[1],&GoAddr[0],GoAddrPtr*sizeof(long));
      memmove(&GoLineNums[1],&GoLineNums[0],GoAddrPtr*sizeof(int));
      GoAddr[0] = addr;
      GoLineNums[0] = DisasmCurrLine;
      if(len < where)
      {
        int i;
        for(i = len;i < where;i++) str[i] = ' ';
        str[i] = 0;
      }
      strcat(str,image);
      str[where + 3] = '0';
      for(i = 1;i <= GoAddrPtr;i++)
      {
        char dig;
        if(i < 10) dig = i + '0';
        else       dig = i - 10 + 'A';
        twDirectWrite(tvioWidth-1,GoLineNums[i]+1,&dig,1);
      }
  }
  else
  if(GoAddrPtr < GO_ADDR_SIZE - 2)
  {
     GoAddrPtr++;
     GoAddr[GoAddrPtr] = addr;
     GoLineNums[GoAddrPtr] = DisasmCurrLine;
     if(len < where)
     {
       int i;
       for(i = len;i < where;i++) str[i] = ' ';
       str[i] = 0;
     }
     strcpy((char *)&str[where],(char *)gidBuildKeyStr());
  }
}

void GidAddBackAddress( void )
{
  if(BackAddrPtr >= BACK_ADDR_SIZE - 2)
  {
      int i;
      for(i = 0;i < BackAddrPtr;i++) BackAddr[i] = BackAddr[i + 1];
      BackAddrPtr--;
  }
  BackAddrPtr++;
  BackAddr[BackAddrPtr] = BMGetCurrFilePos();
}

unsigned long GidGetGoAddress(unsigned keycode)
{
  unsigned long ret;
  keycode &= 0x00FF;
  if(keycode == 0x08)
  {
      if(BackAddrPtr >= 0) ret = BackAddr[BackAddrPtr--];
      else                 ret = BMGetCurrFilePos();
  }
  else
  {
      int ptr;
      if(keycode >= 'a') keycode = keycode + 'A' - 'a';
      if(keycode < 58) ptr = keycode - '0';
      else             ptr = keycode - 'A' + 10;
      if(ptr <= GoAddrPtr)
      {
        GidAddBackAddress();
        ret = GoAddr[ptr];
      }
      else                 ret = BMGetCurrFilePos();
  }
  return ret;
}

char * GidEncodeAddress(unsigned long cfpos,int AddressDetail)
{
  static char addr[11];
  strcpy(addr,Get8Digit(cfpos)); strcat(addr,": ");
  if(AddressDetail)
  {
    if(FType == DOS_SYS)
    {
       if(cfpos < sizeof(DOSDRIVER) + 4) sprintf(addr,"SYSH:%s",Get4Digit(cfpos));
    }
    else
    if(FType >= OLD_EXE && !(headshift && detectedFormat->AddressResolving))
    {
       if(cfpos < sizeof(MZHEADER)) sprintf(addr,"MZH :%s",Get4Digit(cfpos));
       if(oldh.mz.mzRelocationCount)
       {
         if(cfpos >= sizeof(MZHEADER) && cfpos < sizeof(MZHEADER) + (oldh.mz.mzRelocationCount << 2))
         {
            sprintf(addr,"MZRl:%s",Get4Digit(cfpos - sizeof(MZHEADER)));
         }
       }
    }
    if(FType > OLD_EXE && detectedFormat->AddressResolving)
    {
       detectedFormat->AddressResolving(addr,cfpos);
    }
  }
  return addr;
}

