/****************************************************************************\
*   EDITORS - this source file is part of Binary vIEW project (BIEW)         *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1995, 2000 Nick Kurshev                            *
*   License:              See below                                          *
*   Author and developer: Nick Kurshev                                       *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   editors.c                                          *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  This file contain low level editor implementation of BIEW project.        *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/
#include <string.h>
#include <errno.h>

#include "pmalloc.h"
#include "bmfile.h"
#include "kbd_code.h"
#include "tstrings.h"
#include "disasm.h"
#include "bconsole.h"
#include "biewutil.h"
#include "biewhelp.h"
#include "editor.h"
#include "__os_dep.h"

long edit_cp = 0;
struct tag_emem EditorMem;

int edit_x,edit_y;
unsigned char edit_XX = 0;

void ExtHelp( void )
{
 TWindow * using = twUsedWin();
 hlpDisplay(2);
 twUseWin(using);
}

void PaintETitle( int shift,int use_shift )
{
  TWindow * using = twUsedWin();
  char byte,obyte;
  twUseWin(TitleWnd);
  twFreezeWin(TitleWnd);
  twGotoXY(1,1);
  twClrEOL();
  twPrintF("%08lX: ",edit_cp + shift);
  if(!use_shift)
  {
    unsigned eidx;
    eidx = edit_y*EditorMem.width+edit_x;
    byte  = EditorMem.buff[eidx];
    obyte = EditorMem.save[eidx];
  }
  else
  {
     byte  = EditorMem.buff[shift];
     obyte = EditorMem.save[shift];
  }
  if(byte != obyte) twTextColor(Red);
  twPrintF("%c %02XH %sH %sB "
           ,byte ? byte : ' '
           ,byte & 0x00FF
           ,Get2SignDig(byte)
           ,GetBinary(byte));
  twTextColor(Black);
  if(byte != obyte)
  {
    twPrintF("ORIGINAL: %c %02XH %sH %sB "
             ,obyte ? obyte : ' '
             ,obyte & 0x00FF
             ,Get2SignDig(obyte)
             ,GetBinary(obyte));
  }
  else
    twPrintF("                                ");
  twPrintF("MASK: %sH"
           ,Get2Digit(edit_XX));
  twRefreshWin(TitleWnd);
  twUseWin(using);
}

void editInitBuffs(unsigned width, int *err_state )
{
 unsigned long flen,cfp,ssize;
 unsigned i,msize;
 msize = tvioWidth*tvioHeight;
 EditorMem.buff = PMalloc(msize);
 EditorMem.save = PMalloc(msize);
 EditorMem.alen = PMalloc(tvioHeight);
 if((!EditorMem.buff) || (!EditorMem.save) || (!EditorMem.alen))
 {
   if(EditorMem.buff) PFREE(EditorMem.buff);
   if(EditorMem.save) PFREE(EditorMem.save);
   if(EditorMem.alen) PFREE(EditorMem.alen);
   MemOutBox("Editor initialization");
   *err_state = 1;
   return;
 }
 *err_state = 0;
 memset(EditorMem.buff,TW_DEF_FILLER,msize);
 memset(EditorMem.save,TW_DEF_FILLER,msize);
 flen = BMGetFLength();
 edit_cp = cfp = BMGetCurrFilePos();
 EditorMem.size = (unsigned)((unsigned long)msize > (flen-cfp) ? (flen-cfp) : msize);
 EditorMem.width = width;
 BMReadBufferEx(EditorMem.buff,EditorMem.size,cfp,BM_SEEK_SET);
 memcpy(EditorMem.save,EditorMem.buff,EditorMem.size);
 BMSeek(cfp,BM_SEEK_SET);
 /* initialize EditorMem.alen */
 ssize = flen-cfp;
 for(i = 0;i < tvioHeight;i++)
 {
    if(ssize >= width) EditorMem.alen[i] = width;
    else               EditorMem.alen[i] = ssize;
    ssize -= min(ssize,width);
 }
 return;
}

void editDestroyBuffs( void )
{
   PFREE(EditorMem.buff);
   PFREE(EditorMem.save);
   PFREE(EditorMem.alen);
}

void CheckBounds( void )
{
   if(edit_y < 0) edit_y = 0;
   if(edit_y > Height - 1) edit_y = Height - 1;
   if(!EditorMem.alen[edit_y]) edit_y--;
   if(edit_x >= EditorMem.alen[edit_y]) edit_x = EditorMem.alen[edit_y] - 1;
}

void CheckYBounds( void )
{
 if(edit_y < 0) edit_y = 0;
 if(edit_y > Height - 1) edit_y = Height - 1;
 while(!EditorMem.alen[edit_y]) edit_y--;
}

void CheckXYBounds( void )
{
   CheckYBounds();
   if(edit_x < 0) edit_x = EditorMem.alen[--edit_y]*2;
   if(edit_x >= EditorMem.alen[edit_y]*2) { edit_x = 0; edit_y++; }
   CheckYBounds();
}

void editSaveContest( void )
{
  if(!BMWriteBuffEx(edit_cp,BM_SEEK_SET,(void *)EditorMem.buff,EditorMem.size))
  {
      errnoMessageBox("Can not write to file",NULL,errno);
  }
}

int edit_defaction(int lastbyte)
{
 int redraw;
  redraw = 0;
   switch(lastbyte)
   {
     case KE_UPARROW  : edit_y--; break;
     case KE_DOWNARROW: edit_y++; break;
     case KE_ENTER:
     case KE_LEFTARROW:
     case KE_RIGHTARROW: break;
     case KE_F(3)     :
                      Get2DigitDlg(INIT_MASK,INPUT_MASK,&edit_XX);
                      break;
     default: redraw = 1; break;
   }
 return redraw;
}

int editDefAction(int lastbyte)
{
 int redraw = 1;
 int eidx;
 eidx = edit_y*EditorMem.width+edit_x;
   switch(lastbyte)
   {
     case KE_F(4)     : EditorMem.buff[eidx] = ~EditorMem.buff[eidx]; break;
     case KE_F(5)     : EditorMem.buff[eidx] |= edit_XX; break;
     case KE_F(6)     : EditorMem.buff[eidx] &= edit_XX; break;
     case KE_F(7)     : EditorMem.buff[eidx] ^= edit_XX; break;
     case KE_F(8)     : EditorMem.buff[eidx]  = edit_XX; break;
     case KE_F(9)     : EditorMem.buff[eidx] = EditorMem.save[eidx]; break;
     default        : redraw = edit_defaction(lastbyte); edit_x--; break;
   }
   edit_x++;
   return redraw;
}

int FullEdit(TWindow * txtwnd,char ncol,char hcol)
{
 int i,j,mlen;
 unsigned int lastbyte;
 int redraw,flags;
 char attr = __ESS_HARDEDIT | __ESS_WANTRETURN;
 twTextColor(ncol);
 __MsSetState(0);
 twFreezeWin(twUsedWin());
 for(i = 0;i < Height;i++)
 {
   for(j = 0;j < EditorMem.alen[i];j++)
   {
     unsigned eidx;
     eidx = i*EditorMem.width+j;
     twTextColor(EditorMem.buff[eidx] == EditorMem.save[eidx] ? ncol : hcol );
     twDirectWrite(j + 1,i + 1,&EditorMem.buff[eidx],1);
   }
   if(EditorMem.alen[i] + 1 < EditorMem.width)
   {
      twGotoXY(EditorMem.alen[i] + 1,i + 1); twClrEOL();
   }
 }
 twRefreshWin(twUsedWin());
 __MsSetState(1);
 PaintETitle(edit_y*EditorMem.width + edit_x,0);
 __vioSetCursorType(CUR_NORM);
 redraw = 1;
 while(1)
 {
   unsigned eidx;
   eidx = edit_y*EditorMem.width;
   mlen = EditorMem.alen[edit_y];
   flags = attr;
   if(!redraw) flags |= __ESS_NOREDRAW;
   lastbyte = eeditstring((char *)&EditorMem.buff[eidx],NULL,&mlen,edit_y + 1,
                          &edit_x,flags,(char *)&EditorMem.save[eidx],ncol,hcol);
   switch(lastbyte)
   {
     case KE_F(1)   : ExtHelp(); continue;
     case KE_F(2)   : editSaveContest();
     case KE_F(10)  :
     case KE_ESCAPE : return lastbyte;
     case KE_TAB : if(txtwnd) return lastbyte;
     default     : redraw = editDefAction(lastbyte); break;
   }
   CheckBounds();
   PaintETitle(edit_y*EditorMem.width + edit_x,0);
 }
 __vioSetCursorType(CUR_OFF);
 return lastbyte;
}



