/****************************************************************************\
*   BBIO - this source file is part of Binary vIEW project (BIEW)            *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1995, 2000 Nick Kurshev                            *
*   License:              See below                                          *
*   Author and developer: Nick Kurshev                                       *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   lib/bbio.c                                         *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  This file contain implementation of Buffering binary input/ouput routines.*
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/
#include "bbio.h"
#include <sys/stat.h>
#include <stddef.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include "pmalloc.h"
#include "__os_dep.h"
#include "_sys_dep.h"

#ifdef HAVE_MMAN_H
#include <sys/mman.h>
#endif

#define MK_FPTR(x) (x)

typedef struct tagBFILE
{
 int             handle;    /* file handle */
 int             openmode;  /* mode,that OsOpen this file */
 char *          FileName;  /* Real file name of opened file */
 int             updated;   /* 1 if buffer contain data, that not pesent in file */
 long            FilePos;   /* current logical position in file */
 long            FLength;   /* real length of the file */
#ifndef HAVE_MMAN_H
 char *          MBuffer;   /* NULL - not buffered i/o */
 int             optimize;  /* seek optimization */
 unsigned        MBufLen;   /* length data, actually contains in buffer */
 unsigned        MBufSize;  /* real size of buffer */
 long            FBufStart; /* logical position of mirror the buffer onto file */
#endif
}BFILE;

/* notes: all function with prefix=>__ assume, that buffer present */

#define __isOutOfBuffer(obj,pos)\
        (int)((unsigned long)pos < ((BFILE *)obj)->FBufStart ||\
              (unsigned long)pos >= ((BFILE *)obj)->FBufStart +\
               ((BFILE *)obj)->MBufSize)

#define __isOutOfContents(obj,pos)\
        (int)((unsigned long)pos < ((BFILE *)obj)->FBufStart ||\
              (unsigned long)pos >= ((BFILE *)obj)->FBufStart +\
               ((BFILE *)obj)->MBufLen)

static int __NEAR__ __fill(BFILE  *obj,long pos)
{
  int ret;
  void  * mbuff;
  unsigned long remaind;
  if(pos < 0) pos = 0;
  if(pos > obj->FLength) { pos = obj->FLength; obj->MBufLen = 0; return 0; }
  obj->FBufStart = pos;
  remaind = obj->FLength - obj->FBufStart;
  obj->MBufLen = (unsigned long)obj->MBufSize < remaind ? obj->MBufSize : (unsigned)remaind;
  mbuff = MK_FPTR(obj->MBuffer);
  __OsSeek(obj->handle,obj->FBufStart,SEEKF_START);
  ret = __OsRead(obj->handle,mbuff,obj->MBufLen) == obj->MBufLen;
  return ret;
}

static int __NEAR__ __flush(BFILE  *obj)
{
  int ret;
  void  * mbuff;
  ret = 1;
  if(!obj->updated)
  {
    mbuff = MK_FPTR(obj->MBuffer);
    __OsSeek(obj->handle,obj->FBufStart,SEEKF_START);
    if(obj->MBufLen) ret = __OsWrite(obj->handle,mbuff,obj->MBufLen) == obj->MBufLen;
    else             ret = 1;
    if(ret)
    {
      obj->updated = 1;
    }
  }
  return ret;
}

static int __NEAR__ __seek(BFILE  *obj,long pos,unsigned char origin)
{
 int ret,rret;
 ret = 1;
 switch(origin)
 {
    case BIO_SEEK_SET: break;
    case BIO_SEEK_CUR: pos += obj->FilePos; break;
    default:           pos += obj->FLength;
 }
 if((unsigned long)pos > obj->FLength)
 {
    pos = obj->FLength;
    ret = 0;
 }
 obj->FilePos = pos;
 if(__isOutOfBuffer(obj,pos))
 {
    __flush(obj);
    switch(obj->optimize & BIO_OPT_DIRMASK)
    {
      default:
      case BIO_OPT_DB:         break;
      case BIO_OPT_RANDOM:     pos -= obj->MBufSize / 2;
                               break;
      case BIO_OPT_BACKSCAN:   pos -= (obj->MBufSize - 1);
                               break;
      case BIO_OPT_RFORWARD:   pos -= obj->MBufSize / 10;
                               break;
      case BIO_OPT_RBACKSCAN:  pos -= (obj->MBufSize/10)*9;
                               break;
    }
    if(obj->FilePos < obj->FLength)
    {
      rret = __fill(obj,pos);
      if(ret) ret = rret;
    }
    else
    {
      obj->FBufStart = obj->FilePos;
      obj->MBufLen = 0;
    }
 }
 return ret;
}

static int founderr;

static unsigned char __NEAR__ __getc(BFILE  *obj)
{
  int ret;
  unsigned buffpos;
  char *  buff;
  unsigned char ch;
  ret = 1;
  founderr = 0;
  buff = MK_FPTR(obj->MBuffer);
  if(__isOutOfContents(obj,obj->FilePos)) ret = __seek(obj,obj->FilePos,SEEKF_START);
  ch = -1;
  if(obj->MBufLen && ret && obj->FilePos <= obj->FLength)
  {
    buffpos = (unsigned)(obj->FilePos - obj->FBufStart);
    if(buffpos < obj->MBufLen)  ch = buff[buffpos];
    if(obj->FilePos < obj->FLength) obj->FilePos++;
  }
  else { errno = EACCES; ret = 0; }
  if(ret == 0) founderr = 1;
  return ch;
}

static int __NEAR__ __putc(BFILE  *obj,unsigned char ch)
{
  int ret;
  unsigned buffpos;
  char *  buff;
  if(((obj->openmode & (FO_READWRITE | FO_WRITEONLY)) == FO_WRITEONLY) ||
     ((obj->openmode & (FO_READWRITE | FO_WRITEONLY)) == FO_READWRITE)
    )
  {
    ret = 1;
    buff = MK_FPTR(obj->MBuffer);
    if(__isOutOfBuffer(obj,obj->FilePos)) __seek(obj,obj->FilePos,SEEKF_START);
    buffpos = (unsigned)(obj->FilePos - obj->FBufStart);
    buff[buffpos++] = ch;
    obj->updated = 0;
    if(obj->MBufLen < buffpos) if(obj->MBufLen < obj->MBufSize) obj->MBufLen = buffpos;
    if(obj->FLength <= obj->FBufStart + buffpos) obj->FLength = obj->FBufStart + buffpos;
    obj->FilePos++;
  }
  else { errno = EACCES; ret = 0; }
  return ret;
}

static int __NEAR__ __getbuff(BFILE  *obj,char *  buff,unsigned cbBuff)
{
#if 1 /* most optimized, but large method */
  int ret;
  unsigned diffsize;
  unsigned MBufStart,MBufRem;
  int optimize = obj->optimize;
  ret = 1;
  obj->optimize = (obj->optimize & ~BIO_OPT_DIRMASK) | BIO_OPT_DB;
  while(cbBuff)
  {
    MBufStart = (unsigned)(obj->FilePos - obj->FBufStart);
    MBufRem = obj->MBufLen - MBufStart;
    if(!MBufRem || __isOutOfContents(obj,obj->FilePos))
    {
     ret = __seek(obj,obj->FilePos,SEEKF_START);
     if(!ret) break;
     if(obj->FilePos >= obj->FLength) break;
     MBufStart = (unsigned)(obj->FilePos - obj->FBufStart);
     MBufRem = obj->MBufLen - MBufStart;
    }
    if(!MBufRem)
    {
       errno = E2BIG; /* fault: internal error */
       ret = 0;
       break;
    }
    diffsize = min(MBufRem,cbBuff);
    memcpy(buff,&obj->MBuffer[MBufStart],diffsize);
    obj->FilePos += diffsize;
    if(obj->FilePos > obj->FLength) obj->FilePos = obj->FLength;
    cbBuff -= diffsize;
    buff += diffsize;
  }
  obj->optimize = optimize;
  return ret;
#else /* look more safety */
  int ret;
  unsigned i;
  ret = 1;
  for(i = 0;i < cbBuff;i++)
  {
    buff[i] = __getc(obj);
    if(founderr) { ret = 0; break; }
  }
  return ret;
#endif
}

static int __NEAR__ __putbuff(BFILE  *obj,char *  buff,unsigned cbBuff)
{
#if 1 /* most optimized, but large method */
  int ret;
  unsigned diffsize;
  unsigned MBufStart,MBufRem;
  int optimize = obj->optimize;
  obj->optimize = (obj->optimize & ~BIO_OPT_DIRMASK) | BIO_OPT_DB;
  if(((obj->openmode & (FO_READWRITE | FO_WRITEONLY)) == FO_WRITEONLY) ||
     ((obj->openmode & (FO_READWRITE | FO_WRITEONLY)) == FO_READWRITE)
    )
  {
    ret = 1;
    while(cbBuff)
    {
      MBufStart = (unsigned)(obj->FilePos - obj->FBufStart);
      MBufRem = obj->MBufSize - MBufStart;
      if(!MBufRem || __isOutOfBuffer(obj,obj->FilePos))
      {
       ret = __seek(obj,obj->FilePos,SEEKF_START);
       if(!ret) break;
       MBufStart = (unsigned)(obj->FilePos - obj->FBufStart);
       MBufRem = obj->MBufSize - MBufStart;
      }
      if(!MBufRem)
      {
         errno = E2BIG; /* fault: internal error */
         ret = 0;
         break;
      }
      diffsize = min(MBufRem,cbBuff);
      memcpy(&obj->MBuffer[MBufStart],buff,diffsize);
      obj->updated = 0;
      if(obj->MBufLen < MBufStart + diffsize) obj->MBufLen = MBufStart + diffsize;
      if(obj->FLength < obj->FilePos + diffsize) obj->FLength = obj->FilePos + diffsize;
      obj->FilePos += diffsize;
      cbBuff -= diffsize;
      buff += diffsize;
    }
  }
  else { errno = EACCES; ret = 0; }
  obj->optimize = optimize;
  return ret;
#else /* look more safety */
  unsigned i;
  int ret = 0;
  for(i = 0;i < cbBuff;i++)
  {
    ret = __putc(obj,buff[i]);  if(!ret) break;
  }
  return ret;
#endif
}

struct tagBFILE bNull = { 2, -1, 0, 0, 0, 0, 0, 0L, 0L, 0L };

BGLOBAL  bioOpen(const char * fname,unsigned char openmode,unsigned bSize)
{
 BGLOBAL ret = NULL;
 int handle;
 unsigned len;
 ret = PMalloc(sizeof(BFILE));
 if(ret)
 {
   BFILE  * bFile;
   bFile = MK_FPTR(ret);
   bFile->FilePos = 0;
   bFile->FBufStart = 0L;
   bFile->openmode = openmode;
   bFile->updated  = 1;
   bFile->optimize = BIO_OPT_DB;
   if(!(bFile->FileName = PMalloc(strlen(fname)+1)))
   {
     PFREE(bFile);
     return &bNull;
   }
   strcpy(bFile->FileName,fname);
   handle = __OsOpen(fname,openmode);
   if(handle == -1)
   {
     PFREE(bFile->FileName);
     PFREE(ret);
     return &bNull;
   }
   bFile->handle = handle;
   __OsSeek(handle,0L,SEEKF_END);
   bFile->FLength = __OsTell(handle);
   len = bSize;
   if(len)
   {
     if((bFile->MBuffer = PMalloc(len)) != NULL)
     {
       bFile->MBufLen = 0;
       bFile->MBufSize = len;
     }
     else
     {
       PFREE(bFile->FileName);
       __OsClose(bFile->handle);
       PFREE(ret);
       return &bNull;
     }
   }
   __fill(bFile,0L);
 }
 else ret = &bNull;
 return ret;
}

int  bioClose(BGLOBAL handle)
{
  BFILE * bFile;
  bFile = MK_FPTR(handle);
  if(bFile->handle != -1)
  {
    if(((bFile->openmode & FO_READWRITE) == FO_WRITEONLY ||
       (bFile->openmode & FO_READWRITE) == FO_READWRITE))
               __flush(bFile);
    if(bFile->handle != 2) __OsClose(bFile->handle);
  }
  if(bFile->MBuffer) { PFREE(bFile->FileName); PFREE(bFile->MBuffer); }
  PFREE(handle);
  return 1;
}


int   bioSeek(BGLOBAL bioFile,long pos,unsigned char orig)
{
 BFILE  *obj;
 int ret;
 obj = MK_FPTR(bioFile);
 if(obj->MBuffer && !(obj->optimize & BIO_OPT_NOCACHE)) ret = __seek(obj,pos,orig);
 else
 {
   ret = __OsSeek(obj->handle,pos,orig) != 0;
   if(pos == 0L && orig == SEEKF_START) ret = 1;
 }
 return ret;
}

unsigned long  bioTell(BGLOBAL bioFile)
{
  BFILE  *obj;
  unsigned long ret;
  obj = MK_FPTR(bioFile);
  if(obj->MBuffer && !(obj->optimize & BIO_OPT_NOCACHE))
                                            ret = obj->FilePos;
  else                                      ret = __OsTell(obj->handle);
  return ret;
}

unsigned char                 bioReadByte(BGLOBAL bioFile)
{
  unsigned char ret;
  BFILE  *obj;
  obj = MK_FPTR(bioFile);
  if(obj->MBuffer && !(obj->optimize & BIO_OPT_NOCACHE)) ret = __getc(obj);
  else  if(__OsRead(obj->handle,&ret,sizeof(unsigned char)) != sizeof(unsigned char)) ret = -1;
  return ret;
}

unsigned short              bioReadWord(BGLOBAL bioFile)
{
  unsigned short ret;
  BFILE  *obj;
  obj = MK_FPTR(bioFile);
  if(obj->MBuffer && !(obj->optimize & BIO_OPT_NOCACHE))
  {
    if(!__getbuff(obj,(void *)&ret,sizeof(unsigned short))) ret = -1;
  }
  else if(__OsRead(obj->handle,&ret,sizeof(unsigned short)) != sizeof(unsigned short)) ret = -1;
  return ret;
}

unsigned long                bioReadDWord(BGLOBAL bioFile)
{
  unsigned long ret;
  BFILE  *obj;
  obj = MK_FPTR(bioFile);
  if(obj->MBuffer && !(obj->optimize & BIO_OPT_NOCACHE))
  {
    if(!__getbuff(obj,(void *)&ret,sizeof(unsigned long))) ret = -1;
  }
  else if(__OsRead(obj->handle,&ret,sizeof(unsigned long)) != sizeof(unsigned long)) ret = -1;
  return ret;
}

int                 bioReadBuffer(BGLOBAL bioFile,void * buffer,unsigned cbBuffer)
{
  int ret;
  BFILE  *obj;
  obj = MK_FPTR(bioFile);
  if(obj->MBuffer && !(obj->optimize & BIO_OPT_NOCACHE))
  {
    ret = __getbuff(obj,buffer,cbBuffer);
  }
  else ret = __OsRead(obj->handle,buffer,cbBuffer) == cbBuffer;
  return ret;
}

int     bioWriteByte(BGLOBAL bioFile,unsigned char bVal)
{
  int ret;
  BFILE  *obj;
  obj = MK_FPTR(bioFile);
  if(obj->MBuffer && !(obj->optimize & BIO_OPT_NOCACHE))
  {
    ret = __putc(obj,bVal);
  }
  else ret = __OsWrite(obj->handle,&bVal,sizeof(unsigned char)) == sizeof(unsigned char);
  return ret;
}

int  bioWriteWord(BGLOBAL bioFile,unsigned short wVal)
{
  BFILE  *obj;
  int ret;
  obj = MK_FPTR(bioFile);
  if(obj->MBuffer && !(obj->optimize & BIO_OPT_NOCACHE))
                   ret = __putbuff(obj,(void *)&wVal,sizeof(unsigned short));
  else             ret = __OsWrite(obj->handle,&wVal,sizeof(unsigned short)) == sizeof(unsigned short);
  return ret;
}

int     bioWriteDWord(BGLOBAL bioFile,unsigned long dwVal)
{
  BFILE  *obj;
  int ret;
  obj = MK_FPTR(bioFile);
  if(obj->MBuffer && !(obj->optimize & BIO_OPT_NOCACHE))
                   ret = __putbuff(obj,(void *)&dwVal,sizeof(unsigned long));
  else             ret = __OsWrite(obj->handle,&dwVal,sizeof(unsigned long)) == sizeof(unsigned long);
  return ret;
}

int     bioWriteBuffer(BGLOBAL bioFile,const void * buffer,unsigned cbBuffer)
{
  int ret;
  BFILE  *obj;
  obj = MK_FPTR(bioFile);
  if(obj->MBuffer && !(obj->optimize & BIO_OPT_NOCACHE))
  {
    ret = __putbuff(obj,(char *)buffer,cbBuffer);
  }
  else ret = __OsWrite(obj->handle,buffer,cbBuffer) == cbBuffer;
  return ret;
}

int  bioFlush(BGLOBAL bioFile)
{
  BFILE  * obj;
  int ret;
  obj = MK_FPTR(bioFile);
  if(obj->MBuffer && !(obj->optimize & BIO_OPT_NOCACHE))
                   ret = __flush(obj);
  else             ret = 1;
  return ret;
}

int  bioReRead(BGLOBAL bioFile)
{
  BFILE  * obj;
  unsigned long fpos;
  int ret;
  obj = MK_FPTR(bioFile);
  fpos = __OsTell(obj->handle);
  __OsSeek(obj->handle,0L,SEEKF_END);
  obj->FLength = __OsTell(obj->handle);
  __OsSeek(obj->handle,fpos,SEEKF_START);
  ret = __fill(obj,obj->FBufStart);
  return ret;
}

unsigned long   bioFLength(BGLOBAL bioFile)
{
  BFILE  * bFile;
  unsigned long ret;
  bFile = MK_FPTR(bioFile);
  ret = bFile->FLength;
  return ret;
}

int  bioChSize(BGLOBAL bioFile,unsigned long newsize)
{
    unsigned long length, fillsize;
    unsigned  bufsize, numtowrite;
    char * buf;
    BFILE  *obj;
    int ret;
    obj = MK_FPTR(bioFile);

    length = obj->FLength;
    if(length >= newsize)       /* truncate size */
    {
        __seek(obj, newsize, SEEKF_START);
        length = newsize;
        ret = __OsTruncFile(obj->handle, length);
        obj->FLength = newsize;
        if(obj->FBufStart > obj->FLength) { obj->FBufStart = obj->FLength; obj->MBufLen = 0; }
        if(obj->FBufStart + obj->MBufLen > obj->FLength) obj->MBufLen = (unsigned)(obj->FLength - obj->FBufStart);
        return ret ? 0 : 1;
    }
    fillsize=newsize-length;		    /* increase size */
    bufsize = obj->MBuffer ? obj->MBufSize : 8192;
    bufsize = (unsigned) min(fillsize,bufsize);
    buf = PMalloc(bufsize);
    if(buf == NULL)
    {
      return 0;
    }
    memset(buf, 0, bufsize);		    /* write zeros to pad file */
    bioSeek(bioFile,0L,SEEKF_END);
    do
    {
      numtowrite = (unsigned)min(bufsize,fillsize);
      if(!bioWriteBuffer(bioFile, (void * )buf, numtowrite))
      {
         PFREE(buf);
         return 0;
      }
      fillsize-=numtowrite;
    } while(fillsize);
    PFREE(buf);
    return 1;
}

unsigned  bioSetOptimization(BGLOBAL bioFile,unsigned flags)
{
  BFILE  *obj;
  unsigned ret;
  obj = MK_FPTR(bioFile);
  ret = obj->optimize;
  obj->optimize = flags;
  return ret;
}

unsigned  bioGetOptimization(BGLOBAL bioFile)
{
  BFILE  *obj;
  obj = MK_FPTR(bioFile);
  return obj->optimize;
}

int  bioHandle(BGLOBAL bioFile)
{
   BFILE *obj;
   obj = MK_FPTR(bioFile);
   return obj->handle;
}

char * bioFileName(BGLOBAL bioFile)
{
  BFILE *obj;
  obj = MK_FPTR(bioFile);
  return obj->FileName;
}

void * bioBuffer(BGLOBAL bioFile)
{
  BFILE *obj;
  obj = MK_FPTR(bioFile);
  return obj->MBuffer;
}

unsigned bioBuffLen(BGLOBAL bioFile)
{
  BFILE *obj;
  obj = MK_FPTR(bioFile);
  return obj->MBufLen;
}

unsigned bioBuffPos(BGLOBAL bioFile)
{
  BFILE *obj;
  obj = MK_FPTR(bioFile);
  return obj->MBufLen - (unsigned)(obj->FilePos - obj->FBufStart);
}

BGLOBAL bioDup(BGLOBAL bioFile)
{
 BGLOBAL ret = NULL;
 int handle;
 unsigned len;
 ret = PMalloc(sizeof(BFILE));
 if(ret)
 {
   BFILE * bFile,* fromFile;
   bFile = MK_FPTR(ret);
   fromFile = MK_FPTR(bioFile);
   bFile->FilePos = fromFile->FilePos;
   bFile->FBufStart = fromFile->FBufStart;
   bFile->openmode = fromFile->openmode;
   bFile->updated  = fromFile->updated;
   bFile->optimize = fromFile->optimize;
   if(!(bFile->FileName = PMalloc(strlen(fromFile->FileName)+1)))
   {
     PFREE(bFile);
     return &bNull;
   }
   strcpy(bFile->FileName,fromFile->FileName);

   handle = __OsDupHandle(fromFile->handle);
   if(handle == -1)
   {
     PFREE(bFile->FileName);
     PFREE(ret);
     return &bNull;
   }
   bFile->handle = handle;
   bFile->FLength = fromFile->FLength;
   len = fromFile->MBufSize;
   if(len)
   {
     if((bFile->MBuffer = PMalloc(len)) != NULL)
     {
       bFile->MBufLen = fromFile->MBufLen;
       bFile->MBufSize = len;
       memcpy(bFile->MBuffer,fromFile->MBuffer,fromFile->MBufLen);
     }
     else
     {
       PFREE(bFile->FileName);
       __OsClose(bFile->handle);
       PFREE(ret);
       return &bNull;
     }
   }
 }
 else ret = &bNull;
 return ret;
}

/*
retok = mmap(wanted_address_may_NULL,
             need_length,
             protection_read_write_exec,
             flags_shared_fixed_privated,
             filehandle,
             offset_from_begin_file_std_NULL);
*/
