/****************************************************************************\
*   BBIO - this header file is part of Binary vIEW project (BIEW)            *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1995, 2000 Nick Kurshev                            *
*   License:              See below                                          *
*   Author and developer: Nick Kurshev                                       *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   lib/bbio.h                                         *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  This file contain prototypes of BBIO technology functions.                *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/
#ifndef __BBIO_H
#define __BBIO_H

/******************************************************************\
*  Buffered binary file streams input/output section               *
*  Helpful for read/write small size objects from/to file          *
\******************************************************************/

                         /* FORWARD: default (forward scan)
                            reposition of cache as 100% forward
                            from current file position */
#define BIO_OPT_FORWARD  0x0000
#define BIO_OPT_DB       BIO_OPT_FORWARD
                         /* RANDOM: middle scan
                            reposition of cache as 50% forward & 50%
                            backward from current file position */
#define BIO_OPT_RANDOM   0x0001
                         /* BACKWARD: backward scan
                            reposition of cache as 100% backward
                            from current file position */
#define BIO_OPT_BACKSCAN 0x0002
                         /* RANDOM FORWARD: reposition of cache as 90% forward
                            & 10% backward from current file position */
#define BIO_OPT_RFORWARD 0x0003
                         /* RANDOM BACKWARD: reposition of cache as 10% forward
                            & 90% backward from current file position */
#define BIO_OPT_RBACKSCAN 0x0004
                         /* direction mask */
#define BIO_OPT_DIRMASK  0x000F
                         /* NOCACHE:  disable cache */
#define BIO_OPT_NOCACHE  0x8000

#define BIO_SEEK_SET     0
#define BIO_SEEK_CUR     1
#define BIO_SEEK_END     2

typedef void * BGLOBAL;
extern struct tagBFILE bNull; /* pointer to it return bioOpen if fail */

BGLOBAL               bioOpen(const char * fname,unsigned char openmode,unsigned buffSize);
int                   bioChSize(BGLOBAL bioFile,unsigned long newsize);
int                   bioClose(BGLOBAL bioFile);
unsigned long         bioFLength(BGLOBAL bioFile);
int                   bioFlush(BGLOBAL bioFile); /* Flushed buffer onto disk. This operation does automatically when called bioSeek , bioClose functions if openmode == READWRITE || WRITE */
unsigned char         bioReadByte(BGLOBAL bioFile);
unsigned short        bioReadWord(BGLOBAL bioFile);
unsigned long         bioReadDWord(BGLOBAL bioFile);
int                   bioReadBuffer(BGLOBAL bioFile,void * buffer,unsigned cbBuffer);
int                   bioReRead(BGLOBAL bioFile); /*  if file enable write for another process reread current portion of information from disk */
int                   bioSeek(BGLOBAL bioFile,long offset,unsigned char origin);
unsigned              bioGetOptimization(BGLOBAL bioFile);
unsigned              bioSetOptimization(BGLOBAL bioFile,unsigned flags);
unsigned long         bioTell(BGLOBAL bioFile);
int                   bioWriteByte(BGLOBAL bioFile,unsigned char bVal);
int                   bioWriteWord(BGLOBAL bioFile,unsigned short wVal);
int                   bioWriteDWord(BGLOBAL bioFile,unsigned long dwVal);
int                   bioWriteBuffer(BGLOBAL bioFile,const void * buffer,unsigned cbBuffer);
char *                bioFileName(BGLOBAL bioFile);
BGLOBAL               bioDup(BGLOBAL bioFile);

int                   bioHandle(BGLOBAL bioFile);

/* Below describing functions for direct access to cache */
void *                bioBuffer(BGLOBAL bioFile);
unsigned              bioBuffLen(BGLOBAL  bioFile);
unsigned              bioBuffPos(BGLOBAL bioFile);
#endif
