/****************************************************************************\
*   FILE_INI - this header file is part of Binary vIEW project (BIEW)        *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1995, 2000 Nick Kurshev                            *
*   License:              See below                                          *
*   Author and developer: Nick Kurshev                                       *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   lib/file_ini.h                                     *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  This file contain prototypes of .ini file services.                       *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/
/****************************************************************************\
*                         File_ini     Version 1.00                          *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
* File_ini.h - header file for software support work with files have         *
* extension [.ini]; use self syntax for it's files.                          *
*----------------------------------------------------------------------------*
* Preface : This format enabled full initialization of programms and programm*
*           groups. Easy to use and understand like a traditional MS Windows *
*           form, but have expanded syntax and simple C-like preprocessor.   *
*============================================================================*
* Helpful diagramm :                                                         *
*    ; - symbol of commentaries from here and to end of line.                *
*  [ xx ] - Section. Section is global concept, may contain items and        *
*           subsections                                                      *
*  < xx > - SubSection. SubSection is local concept, may contain items only. *
*  xx=yy  - Item. Item is named variable, that contain value that used in    *
*           program.                                                         *
* #xx    -  Preprocessor instructions. Internal command for [.ini] file      *
*           interpretator.                                                   *
\****************************************************************************/

#ifndef __FILE_INI_RUNTIME_SUPPORT_SYSTEM__
#define __FILE_INI_RUNTIME_SUPPORT_SYSTEM__

#include "bbio.h"

/*
    possible errors that are generic
*/

#define __FI_NOERRORS      0
#define __FI_BADFILENAME  -1
#define __FI_BADHANDLER   -2
#define __FI_TOOMANY      -3
#define __FI_READ_AT_END  -4
#define __FI_NOTMEM       -5
#define __FI_OPENCOND     -6
#define __FI_IFNOTFOUND   -7
#define __FI_ELSESTAT     -8
#define __FI_UNRECOGN     -9
#define __FI_BADCOND      -10
#define __FI_BADBRACKET   -11
#define __FI_OPENSECT     -12
#define __FI_BADCHAR      -13
#define __FI_BADVAR       -14
#define __FI_BADVAL       -15
#define __FI_NOVAR        -16
#define __FI_NODEFVAR     -17
#define __FI_ELIFSTAT     -18
#define __FI_OPENVAR      -19
#define __FI_NOTEQU       -20
#define __FI_USER         -21
#define __FI_FIUSER       -22

/*
    possible answers to the errors
*/

#define __FI_IGNORE   0
#define __FI_EXITPROC 1

/*
    return constants for FiSearch
*/

#define __FI_NOTFOUND   0
#define __FI_SECTION    1
#define __FI_SUBSECTION 2
#define __FI_ITEM       3

#ifdef __cplusplus
extern "C" {
#endif

typedef BGLOBAL FiHandler;

typedef struct tagIniInfo
{
  const char * section;
  const char * subsection;
  const char * item;
  const char * value;
}IniInfo;

typedef void      (*FiUserFunc)(IniInfo * info);

/******************************************************\
* You can exchange all this pointers to self routines  *
*    and release virtual access to the methods         *
*                                                      *
* Chematic disgramm :                                  *
*                                                      *
*   FiFileProcessor   -- call --> FiStringProcessor    *
*   FiStringProcessor -- call --> FiCommandProcessor   *
*   FiCommandProcessor - call --> FiFileProcessor &    *
*                                 FiStringProcessor    *
\******************************************************/

extern  int    (*FiError)(int nError,int row);
extern  void   (*FiFileProcessor)(const char *fname);
extern  void   (*FiStringProcessor)(char * curr_str);
extern  char   (*FiCommandProcessor)(const char * cmd);
extern  int    (*FiGetCondition)(const char * cond);


char *               FiDecodeError(int nError);
extern unsigned int  *FinCurrString;
extern char **       FiFileNames;
extern unsigned int  FiFilePtr;
extern unsigned int  FiMaxStrLen;

extern char          FiAllWantInput ; /* All input exclude commentaries, i.e. carriage return and line feed and space characters will be returned */

extern char *        FiUserMessage;

extern char FiOpenComment; /* comment always start with FiOpenComment symbol and to end of line */

void         FiAError(int nError,int row);

/********************************************************************\
* High Level procedure                                               *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
* Use only this function.                                            *
* This function calls UserFunc as CALLBACK routine.                  *
* All internal variables will be expanded, and command processor     *
* will be done. Unknown instruction and hotkeys should be ignored.   *
\********************************************************************/

void          FiProgress(const char * filename,FiUserFunc);

/******************************************************************\
* Low level routines                                               *
\******************************************************************/

FiHandler     FiOpen( const char * filename );
void          FiClose( FiHandler h );
int           FiSearch( FiHandler h, const char * Name );
void          FiSeekTo( FiHandler h, int nSection, int nSubSection, int nItem );
char *        FiGetNextString( FiHandler h, char * store, unsigned int len );
unsigned int  FiGetNumberOfSections( FiHandler h);
unsigned int  FiGetTotalNumberOfSubSections( FiHandler h );
unsigned int  FiGetLocalNumberOfSubSections( FiHandler h, int nSection );
unsigned int  FiGetTotalNumberOfItems( FiHandler h);
unsigned int  FiGetLocalNumberOfItems( FiHandler h,int nSection , int nSubSection);

int           FiisSection( const char * str );
int           FiisSubSection( const char * str );
int           FiisItem( const char * str);
int           FiisCommand( const char * str);

unsigned int  FiGetLengthSection( const char * src );
unsigned int  FiGetLengthSubSection( const char * src );
unsigned int  FiGetLengthItem( const char * src );
unsigned int  FiGetLengthValue( const char * src );
unsigned int  FiGetLengthCommandString( const char * src );

char *        FiGetSectionName(const char * src,char * store);
char *        FiGetSubSectionName(const char * src, char * store);
char *        FiGetItemName(const char * src, char * store);
char *        FiGetValueOfItem(const char * src, char * store);
char *        FiGetCommandString(const char * src, char * store);

void FiFileProcessorStd( const char * filename );
void FiStringProcessorStd( char * string );
char FiCommandProcessorStd( const char * cmd );

/*
    WORD processor
*/

typedef struct tagSTRING
{
  char * str;
  unsigned int iptr;
}STRING;

char *        __FiCMaxStr( void );
char *        __FiCNWord( STRING * str , const char * illegal_set);
char *        __FiCBString( const char * src );
char *        __FiCItem( const char * src );
char *        __FiCValue( const char * src );
char *        __FiCSection( const char * src );
char *        __FiCSubSection( const char * src );
char *        __FiCCmd( const char * src );

unsigned int  FiGetLengthBracketString( const char * src );
char *        FiGetBracketString(const char * str, char * store);
int           FiisLegal(const char * illegal,char c);
unsigned int  FiGetLengthNextWord( STRING * str, const char * illegal_symbols);
char *        FiGetNextWord( STRING * str,const char * illegal_symbols,char * store );
char *        FiSkipWhiteSpace( STRING * str, const char * illegal_symbols );

/*
    variables set
*/

typedef struct tagVar
{
  char * variables;
  char * associate;
  struct tagVar * next;
  struct tagVar * prev;
}Var;

typedef Var * pVar;

pVar            FiConstructVar(const char *v,const char *a);
void            FiDeleteVar(pVar pp);
void            FiDeleteAllVar( void );
const char    * FiExpandVariables(const char * var);
void            FiExpandAllVar(const char * value,char * store);
void            FiAddVariables(const char * var,const char * associate);
void            FiRemoveVariables(const char * var);

int             FiGetConditionStd( const char *condstr);

/*
    High level routines (similar to MS WIN SDK)
*/

typedef struct tag_iniProfile
{
   FiHandler  handler;
   char *fname;
}hIniProfile;

/* For internal purposes */
extern void hlFiProgress(hIniProfile *ini,FiUserFunc usrproc);

/* For public use */

extern hIniProfile*    iniOpenFile(const char *fname);
extern void            iniCloseFile(hIniProfile *ini);

extern unsigned iniReadProfileString(hIniProfile *ini,
                                     const char *section,
                                     const char *subsection,
                                     const char *_item,
                                     const char *def_value,
                                     char *buffer,
                                     unsigned cbBuffer);

extern int iniWriteProfileString(hIniProfile *ini,
                                 const char *section,
                                 const char *subsection,
                                 const char *item,
                                 const char *value);

#ifdef __cplusplus
}
#endif

#endif
