/****************************************************************************\
*   FINIHIGH - this source file is part of Binary vIEW project (BIEW)        *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1995, 2000 Nick Kurshev                            *
*   License:              See below                                          *
*   Author and developer: Nick Kurshev                                       *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   lib/finihigh.c                                     *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  This file contain highe level implementation of .ini files services.      *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/
#include "file_ini.h"
#include "__os_dep.h"
#include "_sys_dep.h"
#include "pmalloc.h"

#include <string.h>
#include <stdio.h>

static unsigned ret;
static unsigned buf_len;
static char *buf_ptr;
static const char *sect,* subsect,* item;

static int __NEAR__ make_temp(const char *path,char *name_ptr)
{
  unsigned i,len;
  int handle;
  char *fullname, *nptr;
  char tbuf[10];
  fullname = PMalloc(strlen(path+1)*2);
  if(!fullname) return -1;
  strcpy(fullname,path);
  len = strlen(fullname);
  if(fullname[len-4] == '.') nptr = &fullname[len-4];
  else                       nptr = &fullname[len];
  handle = -1;
  for(i = 0;i < 100;i++)
  {
  /*
     in this loop we are change only extension:
     file.tmp file.t1 ... file.t99
     (it's only for compatibilities netween diferent OS file systems)
  */
    strcpy(nptr,".T");
    sprintf(tbuf,"%i",i);
    strcat(nptr,tbuf);
    handle = __OsOpen(fullname,FO_READONLY | SO_DENYNONE);
    if(handle == -1) handle = __OsOpen(fullname,FO_READONLY | SO_COMPAT);
    if(handle == -1)
    {
      handle = __OsCreate(fullname);
      if(handle != -1)
      {
        strcpy(name_ptr,fullname);
        goto bye;
      }
    }
  }
  bye:
  return handle;
}

static void MyCallback(IniInfo * ini)
{
  int cond;
  cond = 0;
  if(sect) cond = strcmp(ini->section,sect) == 0;
  if(subsect) cond = strcmp(ini->subsection,subsect) == 0;
  if(item) cond = strcmp(ini->item,item) == 0;
  if(cond)
  {
    ret = min(strlen(ini->value),buf_len);
    strncpy(buf_ptr,ini->value,buf_len);
  }
}

static int __NEAR__ out_sect(FILE * handle,const char *section)
{
   fprintf(handle,"[ %s ]\n",section);
   return 2;
}

static int __NEAR__ out_subsect(FILE * handle,const char *subsection)
{
   fprintf(handle,"  < %s >\n",subsection);
   return 2;
}

extern char ifSmarting;
static void __NEAR__ out_item(FILE * handle,unsigned nled,const char *item,const char *value)
{
  unsigned i;
  char *sm_char;
  sm_char = strchr(value,'%');
  if(sm_char && ifSmarting) fprintf(handle,"#nosmart\n");
  for(i = 0;i < nled;i++) fprintf(handle," ");
  fprintf(handle,"%s = ",item);
  fwrite(value,strlen(value),1,handle);
  fprintf(handle,"\n");
  if(sm_char && ifSmarting) fprintf(handle,"#smart\n");
}

hIniProfile * iniOpenFile(const char *fname)
{
  hIniProfile *ret;
  unsigned fname_size;
  ret = PMalloc(sizeof(hIniProfile));
  if(ret)
  {
    fname_size = strlen(fname)+1;
    ret->fname = PMalloc(fname_size);
    if(ret->fname) strcpy(ret->fname,fname);
    else
    {
      PFree(ret);
      return NULL;
    }
    if(__IsFileExists(ret->fname)) ret->handler = FiOpen(fname);
    else                           ret->handler = &bNull;
  }
  return ret;
}

void iniCloseFile(hIniProfile *ini)
{
  if(ini)
  {
    if(ini->handler != &bNull) FiClose(ini->handler);
    PFree(ini->fname);
    PFree(ini);
  }
}

unsigned iniReadProfileString(hIniProfile *ini,const char *section,const char *subsection,
                           const char *_item,const char *def_value,char *buffer,
                           unsigned cbBuffer)
{
   if(!buffer) return 0;
   ret = 1;
   buf_len = cbBuffer;
   buf_ptr = buffer;
   sect = section;
   subsect = subsection;
   item = _item;
   if(cbBuffer) strncpy(buffer,def_value,cbBuffer);
   else         buffer[0] = 0;
   if(ini) if(ini->handler != &bNull) hlFiProgress(ini,MyCallback);
   return ret;
}

extern char *GETS(char *str,unsigned num,BGLOBAL h);

extern FiHandler ActiveFile;

int iniWriteProfileString(hIniProfile *ini, const char *section, const char *subsection,
                          const char *item, const char *value)
{
   int hsrc,s_ok,ss_ok,i_ok,done,sb_ok,ssb_ok,written,ret,Cond,if_on;
   int need_write;
   unsigned nled,prev_val_size;
   FILE * tmphandle;
   char *tmpname, *prev_val;
   char * workstr, *wstr2;
   /* test for no change of value */
   prev_val_size = strlen(value)+2;
   prev_val = PMalloc(prev_val_size);
   need_write = 1;
   if(!ini) return 0;
   if(prev_val && ini->handler != &bNull)
   {
     char *def_val;
     if(strcmp(value,"y") == 0) def_val = "n";
     else                       def_val = "y";
     iniReadProfileString(ini,section,subsection,item,def_val,prev_val,prev_val_size);
     if(strcmp(prev_val,value) == 0) need_write = 0;
     PFree(prev_val);
   }
   if(!need_write) return 1;
   tmpname = PMalloc((strlen(ini->fname)+1)*2);
   if(!tmpname) return 0;
   if(ini->handler == &bNull) /* if file does not exist make it. */
   {
     FILE *hout;
     hout = fopen(ini->fname,"wt");
     if(hout == NULL) { ret = 0; goto Exit_WS; }
     nled = 0;
     if(section)
     {
       nled += out_sect(hout,section);
     }
     if(subsection)
     {
       nled += out_subsect(hout,subsection);
     }
     if(item)
     {
       out_item(hout,nled,item,value);
     }
     fclose(hout);
     ini->handler = FiOpen(ini->fname);
     ret = 1;
     goto Exit_WS;
   }
   bioSeek(ini->handler,0L,BIO_SEEK_SET);
   ActiveFile = ini->handler;
   hsrc = make_temp(ini->fname,tmpname);
   if(hsrc == -1) { ret = 0; goto Exit_WS; }
   __OsClose(hsrc);
   tmphandle = fopen(tmpname,"wt");
   if(tmphandle == NULL) { ret = 0; goto Exit_WS; }
   workstr = __FiCMaxStr();
   wstr2 = __FiCMaxStr();
   if(!workstr || !wstr2)
   {
     fclose(tmphandle);
     __OsDelete(tmpname);
     if(workstr) PFREE(workstr);
     if(wstr2) PFREE(wstr2);
     ret = 0;
     goto Exit_WS;
   }
   nled = sb_ok = ssb_ok = s_ok = ss_ok = i_ok = done = 0;
   if(!section) { s_ok = sb_ok = 1; if(!subsection) ss_ok = ssb_ok = 1; }
   wstr2[0] = 1;
   Cond = 1;
   if_on = 0;
   while(1)
   {
     char *ch;
     written = 0;
     GETS(workstr,FiMaxStrLen,ini->handler);
     if(workstr[0] == 0) break;
     while(1)
     {
       ch = &workstr[strlen(workstr)-1];
       if(*ch == '\r' || *ch == '\n') *ch = 0;
       else break;
     }
     strcat(workstr,"\n");
     if(FiisCommand(workstr))
     {
       char *cstr;
       cstr = __FiCCmd(workstr);
       if(strncmp(cstr,"if",2) == 0 ||
          strncmp(cstr,"elif",4) == 0 ||
          strncmp(cstr,"else",4) == 0 )
       if_on = 1;
       if(strncmp(cstr,"endif",5) == 0) if_on = 0;
       PFREE(cstr);
     }
     if(Cond)
     {
       if(FiisSection(workstr))
       {
          if(section)
          {
            FiGetSectionName(workstr,wstr2);
            if(strcmp(section,wstr2) == 0)
            {
              s_ok = 1;
              sb_ok = 1;
              nled = 2;
              ss_ok = ssb_ok = !subsection;
            }
            else s_ok = 0;
          }
          else s_ok = 0;
       }
       else
       if(FiisSubSection(workstr))
       {
          if(subsection)
          {
            FiGetSubSectionName(workstr,wstr2);
            if(strcmp(subsection,wstr2) == 0)
            {
              ss_ok = 1;
              ssb_ok = 1;
              nled = 4;
            }
            else ss_ok = 0;
          }
          else ss_ok = 0;
       }
       else
       if(FiisItem(workstr))
       {
          FiGetItemName(workstr,wstr2);
          if(strcmp(item,wstr2) == 0) i_ok = 1;
          if(i_ok && s_ok && ss_ok)
          {
            if(!done || if_on)
            {
               out_item(tmphandle,nled,item,value);
               written = 1;
               done = 1;
            }
          }
          else i_ok = 0;
      }
      /**********************************************/
      if(sb_ok && !s_ok)
      {
          if(!done)
          {
            if(!ssb_ok && !ss_ok && subsection) { nled += 2; out_subsect(tmphandle,subsection); }
            out_item(tmphandle,nled,item,value);
            written = 0;
            done = 1;
          }
      }
      if(s_ok && ssb_ok && !ss_ok)
      {
         if(!done)
         {
             out_item(tmphandle,nled,item,value);
             written = 0;
             done = 1;
         }
      }
    }
    if(!written) fprintf(tmphandle,workstr);
   }
   /************ this is end of loop *************/
   if(!done)
   {
     if(!sb_ok && section)
     {
        nled = out_sect(tmphandle,section);
        if(subsection)
        {
          ssb_ok = 1;
          nled+=out_subsect(tmphandle,subsection);
        }
     }
     if(!ssb_ok && subsection)
     {
        nled+=out_subsect(tmphandle,subsection);
     }
     out_item(tmphandle,nled,item,value);
     done = 1;
   }
   fclose(tmphandle);
   FiClose(ini->handler);
   __OsDelete(ini->fname);
   __OsRename(tmpname,ini->fname);
   ini->handler = FiOpen(ini->fname);
   PFREE(workstr);
   PFREE(wstr2);
   ret = 1;
   Exit_WS:
   PFREE(tmpname);
   return ret;
}


