/****************************************************************************\
*   _SYS_DEP - this header file is part of Binary vIEW project (BIEW)        *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1995, 2000 Nick Kurshev                            *
*   License:              See below                                          *
*   Author and developer: Nick Kurshev                                       *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   lib/sysdep/_sys_dep.h                              *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  This file contain all development system depended part of BIEW project.   *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/

#ifndef __SYS_DEP_H
#define __SYS_DEP_H

#ifdef __cplusplus
extern "C" {
#endif

/* Try to determine target system and processor */
#ifndef __BITNESS__
#ifdef __TSC__
#define __BITNESS__ 16
#else
#define __BITNESS__ 32
#endif
#endif

#if __BITNESS__ == 32
#define __NEAR__
#define __FAR__
#define __HUGE__
#define __INTERRUPT__
#define halloc malloc
#define hrealloc realloc
#define hfree free
#define HMemCpy memcpy
#else
#define __NEAR__  near
#define __FAR__   far
#define __HUGE__  huge
#define __INTERRUPT__ interrupt
/*
   We needs for own version of function hmemcpy because of some 16-bits
   compilers like TopSpeed have last argument as unsigned instead long.
*/
extern void __HUGE__ * HMemCpy(void __HUGE__ *_dest, const void __HUGE__ *_source, unsigned long n);
#endif

#ifndef max
#define max(a,b)    (((a) > (b)) ? (a) : (b))
#endif
#ifndef min
#define min(a,b)    (((a) < (b)) ? (a) : (b))
#endif

extern void memupr(char *,unsigned);
extern void memlwr(char *,unsigned);
#if defined(__GLIBC__) || defined (__UNIX__)
#define strupr(s) memupr(s,strlen(s))
#define strlwr(s) memlwr(s,strlen(s))
#define stricmp strcasecmp
#endif

extern unsigned long ByteSwapL(unsigned long);
extern unsigned short ByteSwapS(unsigned short);

#ifdef __BYTE_ORDER__
  #if __BYTE_ORDER__ == __BIG_ENDIAN__
    #define BYTE_SWAP_L(longval)      (longval)
    #define BYTE_SWAP_S(shortval)     (shortval)
    #define BIG_BYTE_SWAP_L(lonval)   ByteSwapL(longval)
    #define BIG_BYTE_SWAP_S(shortval) ByteSwapS(shortval)
  #else
    #define BYTE_SWAP_L(lonval)       ByteSwapL(longval)
    #define BYTE_SWAP_S(shortval)     ByteSwapS(shortval)
    #define BIG_BYTE_SWAP_L(longval)  (longval)
    #define BIG_BYTE_SWAP_S(shortval) (shortval)
  #endif
#endif

#ifdef __GNUC__

#define atoul atoi /* weal alias */

extern char *        ltoa(long _value, char *_s, int _radix);
extern char *        ultoa(unsigned long _value, char *_s, int _radix);

#endif

#ifdef __cplusplus
}
#endif

#endif
