/****************************************************************************\
*   VIO - this source file is part of Binary vIEW project (BIEW)             *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1999-2000 Konstantin Boldyshev                     *
*   License:              See below                                          *
*   Author and developer: Konstantin Boldyshev                               *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   lib/sysdep/generic/linux/vio.c                     *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  general implementation of video i/o functions for linux.                  *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/

#include "__os_dep.h"
#include "_sys_dep.h"
#include "biewlib.h"
#include "console.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/vt.h>

#define VIO_FILE "/dev/vcsa0"

#define VMAX_X __TVIO_MAXSCREENWIDTH
#define VMAX_Y	0x400
#define VTMP_LEN 0x80

int console_flags = 0, on_console = 0;
unsigned short tvioWidth, tvioHeight, tvioNumColors = 16;

const static int __ansi_color[8] = {0,4,2,6,1,5,3,7};
static int saveX, saveY, firstX, firstY, cursor_type = CUR_NORM;
static int viohandle, out_fd, initialized, output_7 = 0;

static unsigned char *viomem;
static char *vtmp;

#define	calc_vioaddr(x,y)	(unsigned char *) (viomem + ((x) + (y) * tvioWidth) * 2)
#define twrite(x)		write(out_fd, (x), strlen(x))

#define _bg(x) ((x) >> 4)
#define _fg(x) ((x) & 0x0f)
#define _2color(x) (char)__ansi_color[((x) & 7)]

static char *_2ansi(unsigned char attr)
{
    int bg = _bg(attr);
    int bc = _2color(bg);

    if ((bg == 0 || bg == 1) && transparent) bc = 9;

    sprintf(vtmp,
	"\033[%d;3%d;4%d%sm",
	_fg(attr) > 7,
	_2color(_fg(attr) & 7),
	bc,
	bg > 7 ? ";5" : ""
    );
    return vtmp;
}

inline static int printable(unsigned char c)
{
    return !(c < 0x20 || c == 0x7f || c == 0x9b);
}


static void gotoxy(int x, int y)
{
    sprintf(vtmp,"\033[%d;%dH", y + 1, x + 1);
    twrite(vtmp);
}

/*

*/

int __vioGetCursorType( void )
{
    return cursor_type;
}

void __vioSetCursorType(int type)
{
    cursor_type = type;
    if	(terminal == TERM_LINUX && (cursor_type == CUR_SOLID || type == CUR_SOLID)) {
	sprintf(vtmp,"\033[?%dc", type == CUR_SOLID ? 7 : 0);
	twrite(vtmp);
    }
    sprintf(vtmp,"\033[?25%c", type ? 'h' : 'l');
    twrite(vtmp);
}

void __viodGetCursorPos(int *x, int *y)
{
    *x = saveX;
    *y = saveY;
}

void __vioSetCursorPos(int x, int y)
{
    saveX = x;
    saveY = y;
    gotoxy(x, y);
}

void __vioReadBuff(int x, int y, void *buff, unsigned len)
{
    memcpy(buff, calc_vioaddr(x, y), len);
}

void __vioWriteBuff(int x, int y, void *buff, unsigned len)
{
    unsigned char *addr = calc_vioaddr(x, y);

    memcpy(addr, buff, len);

    if (on_console) {
	off_t offset = addr - viomem + 4;
	int i;

	for (i = 0; i < len; i++) {
	    unsigned char *p = (unsigned char *) buff + i;
	    if (*p == 0) *p = ' ';
	}
	lseek(viohandle, offset, SEEK_SET);
	write(viohandle, buff, len);
    } else {
	unsigned int i;
	unsigned char c, ca, *p;
	char *pb = malloc(len * 20);

	if (pb == NULL) {
	    printm("Memory allocation failed: %s\nExiting..", strerror(errno));
	    exit(errno);
	}

	*pb = 0;

	gotoxy(x, y);

	for (i = 0; i < len; i += 2) {
	    p = (unsigned char *) buff + i; c = *p; ca = *(p + 1);

	    if (output_7) c &= 0x7f;
	    if (!c) c = ' '; else if (!printable(c)) c = '.';

	    if ((ca != *(p - 1) && i >= 2) || i == (len - 2) || !i)
		strcat(pb, _2ansi(ca));
	    strncat(pb, &c, 1);
	}
	twrite(pb);
	free(pb);
    }
}

void __term_vio(void)
{
    if (!initialized) return;

    __vioSetCursorType(CUR_NORM);
    if (!on_console) {
	if (terminal == TERM_XTERM) twrite("\033[?1000l");
	twrite("\033[3l\033[0m\033[2J");
    }
    __vioSetCursorPos(firstX, firstY);

    if (on_console) close(viohandle);

    free(vtmp);
    free(viomem);

    close(out_fd);
    initialized = 0;
}

void __init_vio(unsigned long flags)
{
    struct vt_mode vtmode;

    console_flags = flags;

    out_fd = open(ttyname(STDOUT_FILENO), O_WRONLY);
    if (out_fd < 0) out_fd = STDOUT_FILENO;

    if (TESTFLAG(console_flags, __TVIO_FLG_DIRECT_CONSOLE_ACCESS))
	on_console = ioctl(out_fd, VT_GETMODE, &vtmode) >= 0;

    if (on_console && (viohandle = open(VIO_FILE, O_RDWR)) < 0) {
	    printm("Can't open %s: %s\nDirect console access disabled..\n", VIO_FILE, strerror(errno));
	    on_console = 0;
    }

    if (on_console) {
	unsigned char b[4];
	read(viohandle, &b, 4);
	tvioHeight = b[0]; tvioWidth = b[1];
	firstX = b[2]; firstY = b[3];
    } else {
	struct winsize w;
	ioctl(out_fd, TIOCGWINSZ, &w);
	tvioWidth = w.ws_col;
	tvioHeight = w.ws_row;
    }
    output_7 = TESTFLAG(console_flags, __TVIO_FLG_USE_7BIT);
    if (tvioWidth <= 0) tvioWidth = 80;
    if (tvioHeight <= 0) tvioHeight = 25;
    if (tvioWidth > VMAX_X) tvioWidth = VMAX_X;
    if (tvioHeight > VMAX_Y) tvioHeight = VMAX_Y;
    saveX = firstX;
    saveY = firstY;

    vtmp = malloc(VTMP_LEN);
    viomem = malloc(tvioWidth * tvioHeight * 2);
    if (vtmp == NULL || viomem == NULL) {
	printm("Can't allocate memory for output: %s\nExiting..\n", strerror(errno));
	exit(errno);
    }
    memset(vtmp, 0, VTMP_LEN);

    if (on_console)
    	read(viohandle, viomem, tvioWidth * tvioHeight * 2);
    else {
	memset(viomem, 0, tvioWidth * tvioHeight * 2);
	twrite("\033[0m\033[3h");
	if (terminal == TERM_XTERM) {
	    twrite("\033[?1000h\033]0;BIEW: Binary vIEWer\007");
	    output_7 = 1;
	}
    }
    initialized = 1;
}
