/****************************************************************************\
*   VIO - this source file is part of Binary vIEW project (BIEW)             *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1999-2000 Konstantin Boldyshev                     *
*   License:              See below                                          *
*   Author and developer: Konstantin Boldyshev                               *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   lib/sysdep/generic/unix/vio.c                      *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
* curses/slang/vt100 implementation of screen functions                      *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/

#include "__os_dep.h"
#include "_sys_dep.h"
#include "biewlib.h"

#include "console.h"

#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>

#define VMAX_X __TVIO_MAXSCREENWIDTH
#define VMAX_Y	0x400

#ifdef	_SLANG_
#include <slang.h>
#define _bg(x) ((x) >> 4)
#define _fg(x) ((x) & 0x0f)

static char *_color[16] = {
"black",
"blue",
"green",
"cyan",
"red",
"magenta",
"brown",
"lightgray",
"gray",
"brightblue",
"brightgreen",
"brightcyan",
"brightred",
"brightmagenta",
"yellow",
"white"
};

const static unsigned char _pg[] = {
SLSMG_CKBRD_CHAR, SLSMG_CKBRD_CHAR,	SLSMG_CKBRD_CHAR, SLSMG_VLINE_CHAR,
SLSMG_RTEE_CHAR, SLSMG_RTEE_CHAR,	SLSMG_RTEE_CHAR, SLSMG_URCORN_CHAR,
SLSMG_URCORN_CHAR, SLSMG_RTEE_CHAR,	SLSMG_VLINE_CHAR, SLSMG_URCORN_CHAR,
SLSMG_LRCORN_CHAR, SLSMG_LRCORN_CHAR,	SLSMG_LRCORN_CHAR, SLSMG_URCORN_CHAR,
SLSMG_LLCORN_CHAR, SLSMG_DTEE_CHAR,	SLSMG_UTEE_CHAR, SLSMG_LTEE_CHAR,
SLSMG_HLINE_CHAR, SLSMG_PLUS_CHAR,	SLSMG_LTEE_CHAR, SLSMG_LTEE_CHAR,
SLSMG_LLCORN_CHAR, SLSMG_ULCORN_CHAR,	SLSMG_DTEE_CHAR, SLSMG_UTEE_CHAR,
SLSMG_LTEE_CHAR, SLSMG_HLINE_CHAR,	SLSMG_PLUS_CHAR, SLSMG_DTEE_CHAR,
SLSMG_DTEE_CHAR, SLSMG_UTEE_CHAR,	SLSMG_UTEE_CHAR, SLSMG_LLCORN_CHAR,
SLSMG_LLCORN_CHAR, SLSMG_ULCORN_CHAR,	SLSMG_ULCORN_CHAR, SLSMG_PLUS_CHAR,
SLSMG_PLUS_CHAR, SLSMG_LRCORN_CHAR,	SLSMG_ULCORN_CHAR, SLSMG_BLOCK_CHAR,
SLSMG_BLOCK_CHAR, SLSMG_BLOCK_CHAR,	SLSMG_BLOCK_CHAR, SLSMG_BLOCK_CHAR
};
#define	_2ps(c) (((c) >=_PSMIN && (c) <= _PSMAX) ? _pg[(c) - _PSMIN] : 0)
#endif

#ifdef	_CURSES_
#include <curses.h>
#define _bg(x)	(((x) >> 4) & 7)
#define _fg(x)	((x) & 7)
#define _2cpair(x) (_bg(x) * 010 + _fg(x))
const static int _color[8] = {
    COLOR_BLACK,
    COLOR_BLUE,
    COLOR_GREEN,
    COLOR_CYAN,
    COLOR_RED,
    COLOR_MAGENTA,
    COLOR_YELLOW,
    COLOR_WHITE
};

unsigned char _2ps(unsigned char c)
{
#define	ret(x) return(x)
    switch (c) {
	case BAR0:	ret(ACS_CKBOARD);
	case BAR1:	ret(ACS_CKBOARD);
	case BAR2:	ret(ACS_CKBOARD);
	case VSLINE:	ret(ACS_VLINE);
	case RSSTEE:	ret(ACS_RTEE);
	case RSDTEE:	ret(ACS_RTEE);
	case RDSTEE:	ret(ACS_RTEE);
        case TRSDCR:	ret(ACS_URCORNER);
        case TRDSCR:	ret(ACS_URCORNER);
        case RDDTEE:	ret(ACS_RTEE);
        case VDLINE:	ret(ACS_VLINE);
        case TRDDCR:	ret(ACS_URCORNER);
        case LRDDCR:	ret(ACS_LRCORNER);
        case LRSDCR:	ret(ACS_LRCORNER);
        case LRDSCR:	ret(ACS_LRCORNER);
        case TRSSCR:	ret(ACS_URCORNER);
        case LLSSCR:	ret(ACS_LLCORNER);
        case BSSTEE:	ret(ACS_BTEE);
        case TSSTEE:	ret(ACS_TTEE);
        case LSSTEE:	ret(ACS_LTEE);
        case HSLINE:	ret(ACS_HLINE);
        case SSPLUS:	ret(ACS_PLUS);
        case LSDTEE:	ret(ACS_LTEE);
        case LDSTEE:	ret(ACS_LTEE);
        case LLDDCR:	ret(ACS_LLCORNER);
        case TLDDCR:	ret(ACS_ULCORNER);
        case BDDTEE:	ret(ACS_BTEE);
        case TDDTEE:	ret(ACS_TTEE);
        case LDDTEE:	ret(ACS_LTEE);
        case HDLINE:	ret(ACS_HLINE);
        case DDPLUS:	ret(ACS_PLUS);
        case BDSTEE:	ret(ACS_BTEE);
        case BSDTEE:	ret(ACS_BTEE);
        case TDSTEE:	ret(ACS_TTEE);
        case TSDTEE:	ret(ACS_TTEE);
        case BLDSCR:	ret(ACS_LLCORNER);
        case BLSDCR:	ret(ACS_LLCORNER);
        case TLSDCR:	ret(ACS_ULCORNER);
        case TLDSCR:	ret(ACS_ULCORNER);
        case SDPLUS:	ret(ACS_PLUS);
        case DSPLUS:	ret(ACS_PLUS);
        case BRSSCR:	ret(ACS_LRCORNER);
        case TLSSCR:	ret(ACS_ULCORNER);
        case DBLK:	ret(ACS_BLOCK);
        case BBLK:	ret(ACS_BLOCK);
        case LBLK:	ret(ACS_BLOCK);
        case RBLK:	ret(ACS_BLOCK);
        case TBLK:	ret(ACS_BLOCK);
    }
    ret(0);
#undef ret
}
#endif

#ifdef	_VT100_

#include <errno.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/ioctl.h>

#define _bg(x) ((x) >> 4)
#define _fg(x) ((x) & 0x0f)

#ifndef STDOUT_FILENO
#define STDOUT_FILENO 1
#endif

#define VTMP_LEN 100

static char *vtmp;
static int out_fd;
const static int _color[8] = {0,4,2,6,1,5,3,7};

#define twrite(x)	write(out_fd,(x),strlen(x))
#define	_2ps(c) (0)

/*
    convert vga attrubute to ansi escape sequence
*/

static char *_2ansi(unsigned char attr)
{
    int bg = _bg(attr);
    int bc = _color[bg & 7];

    if ((bg == 0 || bg == 1) && transparent) bc = 9;

    sprintf(vtmp,
	"\033[%d;3%d;4%d%sm",
	_fg(attr) > 7,
	_color[_fg(attr) & 7],
	bc,
	bg > 7 ? ";5" : ""
    );
    return vtmp;
}

#endif

/*

*/

#define	calc_vioaddr(x,y) (unsigned char *) (viomem + ((x) + (y) * tvioWidth) * 2)

int console_flags = 0, on_console = 0, output_7 = 0, output_raw = 0;
unsigned short tvioWidth = 80, tvioHeight = 25, tvioNumColors = 16;
static int initialized = 0,cursor_status = CUR_NORM;
static int saveX, saveY, firstX = 0, firstY = 0;
static unsigned char *viomem;

/*

*/

inline static int printable(unsigned char c)
{
    return !(c < 0x20 || c == 0x7f || c == 0x9b);
}

int __vioGetCursorType(void)
{
    return cursor_status;
}

void __vioSetCursorType(int type)
{
    cursor_status = type;
#ifdef	_SLANG_
    SLtt_set_cursor_visibility(type);
#endif
#ifdef	_CURSES_
    curs_set(type);
#endif
#ifdef	_VT100_
    if	(terminal == TERM_LINUX && (cursor_status == CUR_SOLID || type == CUR_SOLID)) {
	sprintf(vtmp,"\033[?%dc", type == CUR_SOLID ? 7 : 0);
	twrite(vtmp);
    }
    sprintf(vtmp, "\033[?25%c", type == CUR_OFF ? 'l' : 'h');
    twrite(vtmp);
#endif
}


void __vioGetCursorPos(int *x, int *y)
{
    *x = saveX;
    *y = saveY;
}

void __vioSetCursorPos(int x, int y)
{
    saveX = x;
    saveY = y;
#ifdef	_CURSES_
    move(y, x);
#endif
#ifdef	_SLANG_
    SLsmg_gotorc(y, x);
#endif
#ifdef	_VT100_
    sprintf(vtmp,"\033[%d;%dH", y + 1, x + 1);
    twrite(vtmp);
#endif
}

void __vioReadBuff(int x, int y, void *buff, unsigned len)
{
    memcpy(buff, calc_vioaddr(x, y), len);
}

void __vioWriteBuff(int x, int y, void *buff, unsigned len)
{
    unsigned char c, ca, *p;
    unsigned int i;

#ifdef	_SLANG_
    int xx, yy;
    __vioGetCursorPos(&xx, &yy);
#endif
#ifdef	_CURSES_
    chtype ch = 0;
#endif
#ifdef	_VT100_
    unsigned char *pb = malloc(len * 20);
    if (pb == NULL) {
	printm("Memory allocation failed: %s\nExiting..", strerror(errno));
	exit(errno);
    }
    *pb = 0;
#endif

    memcpy(calc_vioaddr(x, y), buff, len);

    __vioSetCursorPos(x, y);

    for (i = 0; i < len; i += 2) {
	int pseudo = 0;

	p = buff + i; c = *p; ca = *(p + 1);

	if (c >= _PSMIN && c <=	_PSMAX) pseudo = 1;

	if (c >= 0x80 && !output_raw) {
	    unsigned char cc = _2ps(c);
	    if (output_7) cc = _2ps7(c);
	    if (cc) {
		c = cc;
		pseudo = 1;
	    }
	    if (c >= 0x80 && (terminal == TERM_LINUX || !pseudo))
		c = alt2koi[c - 0x80];
	}
	if (output_7) c &= 0x7f;
	if (!c) c = ' '; else if (!printable(c)) c = '.';

#ifdef	_SLANG_
	SLsmg_set_color(ca);
	SLsmg_set_char_set(!output_7 && pseudo);
	SLsmg_write_char(c);
#endif
#ifdef	_CURSES_
	ch = COLOR_PAIR(_2cpair(ca));
	if ((ca & 0xf) > 7)	ch |= A_BOLD;
	if ((ca & 0xf0) > 0x70)	ch |= A_BLINK;
	addch(c | ch);
#endif
#ifdef	_VT100_
	if ((ca != *(p - 1) && i >= 2) || i == (len - 2) || !i)
	strcat(pb, _2ansi(ca));
	strncat(pb, &c, 1);
#endif
    }
#ifdef	_SLANG_
    __vioSetCursorPos(xx, yy);
    SLsmg_refresh();
#endif
#ifdef	_CURSES_
    refresh();
#endif
#ifdef	_VT100_
    twrite(pb);
    free(pb);
#endif
}

void __init_vio(unsigned long flags)
{
#ifdef	_VT100_
    struct winsize w;
#else
    int i;
#endif

    console_flags = flags;
    output_7 = TESTFLAG(console_flags, __TVIO_FLG_USE_7BIT);

#ifdef	_SLANG_
    SLtt_get_terminfo();
    SLkp_init();
    SLang_init_tty(0xff, 1, 1);
    for (i = 1; i < 0xff; i++) {
	char *bc = _color[_bg(i)];
	if ((_bg(i) == 1 || _bg(i) == 0) && transparent) bc = "default";
	SLtt_set_color(i, NULL, _color[_fg(i)], bc);
    }
    SLsmg_init_smg();
    SLsmg_Display_Eight_Bit = 0x80;
    tvioWidth = SLtt_Screen_Cols;
    tvioHeight = SLtt_Screen_Rows;
    tvioNumColors = SLtt_Use_Ansi_Colors ? 16 : 2;
    if (!SLtt_Has_Alt_Charset) output_7 = 1;
#endif
#ifdef	_CURSES_
    initscr();
    if (has_colors()) start_color();
    else tvioNumColors = 2;
    tvioWidth = COLS;
    tvioHeight = LINES;
    for (i = 0; i < 0xff; i++)
	init_pair(_2cpair(i), _color[_fg(i)], _color[_bg(i)]);
#endif
#ifdef	_VT100_
    if ((vtmp = malloc(VTMP_LEN)) == NULL) {
	printm("Can't allocate memory for output: %s\nExiting..", strerror(errno));
	exit(errno);
    }
    memset(vtmp, 0, VTMP_LEN);

    out_fd = open(ttyname(STDOUT_FILENO), O_WRONLY);
    if (out_fd < 0) out_fd = STDOUT_FILENO;

    ioctl(out_fd, TIOCGWINSZ, &w);
    tvioWidth = w.ws_col;
    tvioHeight = w.ws_row;
#endif

    if (tvioWidth <= 0) tvioWidth = 80;
    if (tvioHeight <= 0) tvioHeight = 25;
    if (tvioWidth > VMAX_X) tvioWidth = VMAX_X;
    if (tvioHeight > VMAX_Y) tvioHeight = VMAX_Y;
    saveX = firstX;
    saveY = firstY;

    if ((viomem = malloc(tvioWidth * tvioHeight * 2)) == NULL) {
	printm("Can't allocate memory for output: %s\nExiting..", strerror(errno));
	exit(errno);
    }
    memset(viomem, 0, tvioWidth * tvioHeight * 2);

#ifdef	_VT100_
    twrite("\033[0m\033[3h");
    if (terminal == TERM_XTERM) {
	twrite("\033[?1001s\033[?1000h\033]0;BIEW: Binary vIEWer\007");
	output_7 = 1;
    }
#endif

    initialized = 1;
}

void __term_vio(void)
{
    if (!initialized) return;

    __vioSetCursorPos(firstX, firstY);
    __vioSetCursorType(CUR_NORM);

#ifdef	_SLANG_
    SLsmg_reset_smg();
    SLang_reset_tty();
#endif
#ifdef	_CURSES_
    endwin();
#endif
#ifdef	_VT100_
    twrite("\033[?1001r\033[?1000l\033[3l\033[0m\033[2J");
    close(out_fd);
    free(vtmp);
#endif
    free(viomem);
    initialized = 0;
}
