/****************************************************************************\
*   FILEIO - this source file is part of Binary vIEW project (BIEW)          *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1995, 2000 Nick Kurshev                            *
*   License:              See below                                          *
*   Author and developer: Nick Kurshev                                       *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   lib/sysdep/i386/os2/fileio.c                       *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  This file contain implementation of file i/o routines for OS/2-32.        *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/
#include <os2.h>
#include <string.h>
#include <errno.h>

#ifdef __EMX__
extern void _sys_set_errno(unsigned long rc);
#endif

static void fnUnix2Dos(char *fn)
{
  char *ptr;
  ptr = fn;
  while((ptr = strchr(ptr,'/')) != NULL) *ptr = '\\';
}

void __OsClose(int handle)
{
  unsigned long rc;
  rc = DosClose(handle);
#ifdef __EMX__
     _sys_set_errno(rc);
#else
     errno = rc;
#endif
}

int __OsDelete(const char *name)
{
  unsigned long rc;
  fnUnix2Dos(name);
  rc = DosDelete((PSZ)name);
#ifdef __EMX__
     _sys_set_errno(rc);
#else
     errno = rc;
#endif
  return rc;
}

int __OsRename(const char *oldname,const char *newname)
{
  unsigned long rc;
  fnUnix2Dos(oldname);
  fnUnix2Dos(newname);
  rc = DosMove((PSZ)oldname,(PSZ)newname);
#ifdef __EMX__
     _sys_set_errno(rc);
#else
     errno = rc;
#endif
  return rc;
}

int  __OsDupHandle(int handle)
{
  HFILE ret;
  unsigned long rc;
  ret = -1; /* need by OS/2 SDK */
  rc = DosDupHandle(handle,&ret);
  if(rc) ret = -1;
#ifdef __EMX__
     _sys_set_errno(rc);
#else
     errno = rc;
#endif
  return ret;
}

int __OsChSize(int handle, long size)
{
  unsigned long rc;
  rc = DosSetFileSize(handle,size);
#ifdef __EMX__
     _sys_set_errno(rc);
#else
     errno = rc;
#endif
  return rc ? -1 : 0;
}

int  __OsCreate(char *fname)
{
   HFILE fhandle;
   ULONG action,rc;
   fnUnix2Dos(fname);
   rc = DosOpen((PSZ)fname,&fhandle,&action,0L,0,
                OPEN_ACTION_CREATE_IF_NEW | OPEN_ACTION_FAIL_IF_EXISTS,
                OPEN_ACCESS_READWRITE | OPEN_SHARE_DENYREADWRITE,0L);
   if(rc)
   {
     fhandle = -1;
#ifdef __EMX__
     _sys_set_errno(rc);
#else
     errno = rc;
#endif
   }
   return fhandle;
}

int  __OsOpen(char *fname,int mode)
{
   HFILE fhandle;
   ULONG action,rc;
   fnUnix2Dos(fname);
   mode |= OPEN_FLAGS_FAIL_ON_ERROR;
          /* report media i/o error trough retcode but not
             through critical error */
   rc = DosOpen((PSZ)fname,&fhandle,&action,0L,0,OPEN_ACTION_OPEN_IF_EXISTS,mode,0L);
   if(rc)
   {
     fhandle = -1;
#ifdef __EMX__
     _sys_set_errno(rc);
#else
     errno = rc;
#endif
   }
   return fhandle;
}

long __OsSeek( int handle, long offset, int origin)
{
  unsigned long ret,rc;
  rc = DosSetFilePtr(handle,offset,origin,&ret);
  if(rc) ret = -1;
#ifdef __EMX__
     _sys_set_errno(rc);
#else
     errno = rc;
#endif
  return ret;
}

int __OsTruncFile( int handle, unsigned long size)
{
  unsigned long rc;
  rc = DosSetFileSize(handle,size);
#ifdef __EMX__
     _sys_set_errno(rc);
#else
     errno = rc;
#endif
  return rc;
}

int __OsRead(int handle, void *buff,unsigned size)
{
  ULONG ret;
  unsigned long rc;
  rc = DosRead(handle,buff,size,&ret);
#ifdef __EMX__
     _sys_set_errno(rc);
#else
     errno = rc;
#endif
  return ret;
}

int __OsWrite(int handle, void *buff,unsigned size)
{
  ULONG ret;
  unsigned long rc;
  rc = DosWrite(handle,buff,size,&ret);
#ifdef __EMX__
     _sys_set_errno(rc);
#else
     errno = rc;
#endif
  return ret;
}
