/****************************************************************************\
*   FILEIO - this source file is part of Binary vIEW project (BIEW)          *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1995, 2000 Nick Kurshev                            *
*   License:              See below                                          *
*   Author and developer: Nick Kurshev                                       *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   lib/sysdep/i386/win32/fileio.c                     *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  This file contain implementation of file i/o routines for Win32s.         *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/
#include <windows.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include "__os_dep.h"
#include "_sys_dep.h"

static void fnUnix2Dos(char *fn)
{
  char *ptr;
  ptr = fn;
  while((ptr = strchr(ptr,'/')) != NULL) *ptr = '\\';
}

int __OsCreate(const char *name)
{
  HANDLE handle;
  fnUnix2Dos(name);
  handle = CreateFile(name,
                      GENERIC_WRITE | GENERIC_READ,
                      0,
                      NULL,
                      CREATE_NEW,
                      FILE_ATTRIBUTE_NORMAL,
                      NULL);
  if(handle == INVALID_HANDLE_VALUE)
  {
    errno = GetLastError();
    (int)handle = -1;
  }
  return (int)handle;
}

int __OsDelete(const char *name)
{
  fnUnix2Dos(name);
  if(!DeleteFile(name))
  {
    errno = GetLastError();
    return -1;
  }
  else return 0;
}

int __OsRename(const char *oldname,const char *newname)
{
  fnUnix2Dos(oldname);
  fnUnix2Dos(newname);
  if(!MoveFile(oldname,newname))
  {
    errno = GetLastError();
    return -1;
  }
  else return 0;
}

void __OsClose(int handle)
{
  CloseHandle((HANDLE)handle);
}

int __OsDupHandle(int handle)
{
  HANDLE hret;
  if(DuplicateHandle(GetCurrentProcess(),
                     (HANDLE)handle,
                     GetCurrentProcess(),
                     &hret,
                     FILE_MAP_ALL_ACCESS,
                     TRUE,
                     DUPLICATE_SAME_ACCESS) == 0)
  {
    errno = GetLastError();
    (int)hret = -1;
  }
  return (int)hret;
}

int  __OsOpen(const char *fname,int mode)
{
  HANDLE handle;
  int access,smode;
  access = smode = 0;
  access = GENERIC_READ;
  fnUnix2Dos(fname);
  if(mode & FO_WRITEONLY) access |= GENERIC_WRITE;
  if(mode & FO_READWRITE) access |= GENERIC_WRITE | GENERIC_READ;
  if(mode & SO_DENYWRITE) smode |= FILE_SHARE_READ;
  if(mode & SO_DENYREAD) smode |= FILE_SHARE_WRITE;
  if(mode & SO_DENYNONE) smode |= FILE_SHARE_READ | FILE_SHARE_WRITE;
  handle = CreateFile(fname,
                      access,
                      smode,
                      NULL,
                      OPEN_EXISTING,
                      FILE_ATTRIBUTE_NORMAL,
                      NULL);
  if(handle == INVALID_HANDLE_VALUE)
  {
    errno = GetLastError();
    (int)handle = -1;
  }
  return (int)handle;
}

long __OsSeek(int handle,long offset,int origin)
{
  return SetFilePointer((HANDLE)handle,offset,NULL,origin);
}

int __OsTruncFile( int handle, unsigned long newsize)
{
  __OsSeek(handle,newsize,SEEKF_START);
  SetEndOfFile((HANDLE)handle);
  return newsize;
}

int  __OsRead(int handle, void *buff, unsigned size)
{
  unsigned ret = size;
  if(ReadFile((HANDLE)handle,buff,size,&ret,NULL) == 0)
  {
    errno = GetLastError();
    ret = 0;
  }
  return ret;
}

int  __OsWrite(int handle,const void *buff, unsigned size)
{
  unsigned ret = size;
  if(WriteFile((HANDLE)handle,buff,size,&ret,NULL) == 0)
  {
    errno = GetLastError();
    ret = 0;
  }
  return ret;
}

int __OsChSize(int handle, long size)
{
  return __OsTruncFile(handle,size);
}
