/****************************************************************************\
*   KEYBOARD - this source file is part of Binary vIEW project (BIEW)        *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1995, 2000 Nick Kurshev                            *
*   License:              See below                                          *
*   Author and developer: Nick Kurshev                                       *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   lib/sysdep/i386/win32/keyboard.c                   *
*----------------------------------------------------------------------------*
*   28.03.2000 Shoorick <alex@eunet.lt> fixing code for workaround of        *
*                                       console input bug under Win95        *
*                                       Tested on Win95 PE, OSR2, Win98.     *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  This file contain implementation of keyboard handles for Win32s.          *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/
#include "windows.h"
#ifdef _MSC_VER
#include <limits.h>
#endif
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include "__os_dep.h"
#include "kbd_code.h"

static int KB_Buff[64];
static unsigned char KB_freq = 0;
static int shiftkeys = 0;

HANDLE hIn;
extern OSVERSIONINFO win32_verinfo;
static int is_win9x;


static int __mou_nbtns;

int is_VKCtrl(int code)
{
  return code == VK_MENU ||
         code == VK_CONTROL ||
         code == VK_SHIFT;
}

int is_VKMouse(int code)
{
  return code == VK_LBUTTON ||
         code == VK_RBUTTON ||
         code == VK_MBUTTON;
}

extern int win32_mx,win32_my,win32_mbuttons;

static int __normalize_keycode(int keycode)
{
  int ret;
  switch(keycode)
  {
    case 0x50E0: ret = 0x5000; break;
    case 0x48E0: ret = 0x4800; break;
    case 0x4BE0: ret = 0x4B00; break;
    case 0x4DE0: ret = 0x4D00; break;
    case 0x47E0: ret = 0x4700; break;
    case 0x4FE0: ret = 0x4F00; break;
    case 0x49E0: ret = 0x4900; break;
    case 0x51E0: ret = 0x5100; break;
    case 0x52E0: ret = 0x5200; break;
    case 0x53E0: ret = 0x5300; break;
    default: ret = keycode;
  }
  return ret;
}

void win32_readNextMessage( void )
{
  INPUT_RECORD ir;
  DWORD nread;
  int vkeycode,keycode,i;
  GetNumberOfConsoleInputEvents(hIn,&nread);
  for(i = 0;i < nread;i++)
  {
    int is_read;
    is_read = 0;
    if(is_win9x) PeekConsoleInput(hIn,&ir,1,&nread);
    else         ReadConsoleInput(hIn,&ir,1,&nread);
    switch(ir.EventType)
    {
      case MOUSE_EVENT:
      {
         if(ir.Event.MouseEvent.dwEventFlags == MOUSE_MOVED)
         {
            win32_mx = ir.Event.MouseEvent.dwMousePosition.X;
            win32_my = ir.Event.MouseEvent.dwMousePosition.Y;
         }
         else  win32_mbuttons = ir.Event.MouseEvent.dwButtonState;
         if(KB_freq < sizeof(KB_Buff)/sizeof(int)) KB_Buff[KB_freq++] = KE_MOUSE;
      }
      break;
      case KEY_EVENT:
      {
        shiftkeys = ir.Event.KeyEvent.dwControlKeyState;
        vkeycode = ir.Event.KeyEvent.wVirtualKeyCode;
        if(is_VKMouse(vkeycode))
        {
          if(vkeycode == VK_LBUTTON)
          {
            if(ir.Event.KeyEvent.bKeyDown == TRUE) win32_mbuttons |= MS_LEFTPRESS;
            else                                   win32_mbuttons &= ~MS_LEFTPRESS;
          }
          if(vkeycode == VK_RBUTTON)
          {
            if(ir.Event.KeyEvent.bKeyDown == TRUE) win32_mbuttons |= MS_RIGHTPRESS;
            else                                   win32_mbuttons &= ~MS_RIGHTPRESS;
          }
          if(vkeycode == VK_MBUTTON)
          {
            if(ir.Event.KeyEvent.bKeyDown == TRUE) win32_mbuttons |= MS_MIDDLEPRESS;
            else                                   win32_mbuttons &= ~MS_MIDDLEPRESS;
          }
         if(KB_freq < sizeof(KB_Buff)/sizeof(int)) KB_Buff[KB_freq++] = KE_MOUSE;
        }
        if(ir.Event.KeyEvent.bKeyDown == TRUE)
        {
          /*
             We must do this stupid work cause of in win95 - OSR2
             when CAPSLOCK is ON and ESCAPE key is PRESSED then
             ...uChar.AsciiChar == 0 for it. I do not know any more such bugs.
          */
          if((unsigned char)ir.Event.KeyEvent.uChar.AsciiChar == 0 &&
             (unsigned char)ir.Event.KeyEvent.wVirtualKeyCode == 27 &&
             (unsigned char)ir.Event.KeyEvent.wVirtualScanCode == 1 &&
             is_win9x)
                   ir.Event.KeyEvent.uChar.AsciiChar = 27;
          keycode = ((ir.Event.KeyEvent.wVirtualScanCode << 8) & 0xFF00) |
                    (ir.Event.KeyEvent.uChar.AsciiChar & 0xFF);
          if(is_VKCtrl(vkeycode))
          {
            if(KB_freq < sizeof(KB_Buff)/sizeof(int)) KB_Buff[KB_freq++] = KE_SHIFTKEYS;
          }
          else
          if(KB_freq < sizeof(KB_Buff)/sizeof(int))
          {
            if(is_win9x)
            {
              if(shiftkeys & SHIFT_PRESSED) keycode |= ADD_SHIFT;
              else
                if((shiftkeys & LEFT_CTRL_PRESSED) ||
                   (shiftkeys & RIGHT_CTRL_PRESSED)) keycode |= ADD_CONTROL;
                else
                  if((shiftkeys & LEFT_ALT_PRESSED) ||
                     (shiftkeys & RIGHT_ALT_PRESSED))  keycode |= ADD_ALT;
 /*
    ReadConsoleInput in win9x don't return correct keycode
    for national language letters, so use ReadConsole (it works)
    this method can't be enabled in winnt, because it has its own bugs.
 */
              if((unsigned char)ir.Event.KeyEvent.uChar.AsciiChar>=0x20
                 && !(keycode & ADD_CONTROL) && !(keycode & ADD_ALT))
              {
                char key;
                ReadConsole(hIn,&key,1,&nread,NULL);
                keycode = __normalize_keycode(key);
                if(shiftkeys & SHIFT_PRESSED) keycode |= KS_SHIFT;
                is_read = 1;
              }
            }
            else
            {
              int sk;
              sk = __kbdGetShiftsKey();
              keycode = __normalize_keycode(keycode);
              if(sk & KS_SHIFT) keycode |= ADD_SHIFT;
              else
                if(sk & KS_CTRL) keycode |= ADD_CONTROL;
                else
                  if(sk & KS_ALT) keycode |= ADD_ALT;
            }
            if(KB_freq < sizeof(KB_Buff)/sizeof(int)) KB_Buff[KB_freq++] = keycode;
          }
        }
        else
         if(ir.Event.KeyEvent.bKeyDown == FALSE)
         {
           if(is_VKCtrl(vkeycode))
           {
             if(KB_freq < sizeof(KB_Buff)/sizeof(int)) KB_Buff[KB_freq++] = KE_SHIFTKEYS;
           }
         }
      }
      break;
      default: break;
    }
    if(is_win9x && !is_read) ReadConsoleInput(hIn,&ir,1,&nread);
  }
}

void __init_keyboard( void )
{
  hIn = GetStdHandle(STD_INPUT_HANDLE);
  is_win9x = win32_verinfo.dwMajorVersion == 4 &&
             win32_verinfo.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS;
  if((__mou_nbtns = __init_mouse()) != INT_MAX)
  {
    __MsSetState(1);
  }
}

void __term_keyboard( void )
{
  if(__mou_nbtns != INT_MAX)
  {
    __MsSetState(0);
    __term_mouse();
  }
}

int __kbdTestKey( unsigned long flg )
{
  if(__MsGetBtns() && flg == KBD_NONSTOP_ON_MOUSE_PRESS) return KE_MOUSE;
  win32_readNextMessage();
  return KB_freq;
}

int __kbdGetShiftsKey( void )
{
  int ret;
  ret = 0;
  win32_readNextMessage();
  if(shiftkeys & SHIFT_PRESSED) ret |= KS_SHIFT;
  if((shiftkeys & LEFT_ALT_PRESSED) || (shiftkeys & RIGHT_ALT_PRESSED)) ret |= KS_ALT;
  if((shiftkeys & LEFT_CTRL_PRESSED) || (shiftkeys & RIGHT_CTRL_PRESSED)) ret |= KS_CTRL;
  if(shiftkeys & CAPSLOCK_ON) ret |= KS_CAPSLOCK;
  if(shiftkeys & NUMLOCK_ON) ret |= KS_NUMLOCK;
  if(shiftkeys & SCROLLLOCK_ON) ret |= KS_SCRLOCK;
  return ret;
}

int __kbdGetKey ( unsigned long flg )
{
  int ret;
  if(__MsGetBtns() && flg == KBD_NONSTOP_ON_MOUSE_PRESS) return KE_MOUSE;
  while(!KB_freq) { __OsYield(); win32_readNextMessage(); }
  ret = KB_Buff[0];
  --KB_freq;
  if(KB_freq) memmove(KB_Buff,&KB_Buff[1],KB_freq-1);
  return ret;
}

