/****************************************************************************\
*   TIMER - this source file is part of Binary vIEW project (BIEW)           *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1995, 2000 Nick Kurshev                            *
*   License:              See below                                          *
*   Author and developer: Nick Kurshev                                       *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   lib/sysdep/i86/dos/timer.c                         *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  This file contain implementation of timer depended part for DOS.          *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/
#include <windows.h>
#include <stddef.h>
#include <stdlib.h>
#include "__os_dep.h"

#ifdef __GNUC__
/****************************************************************\
* Cygnus GNU C/C++ v0.20b does not have 'mmsystem.h' header file *
\****************************************************************/
#ifdef __cplusplus
extern "C" {
#endif

/* timer device capabilities data structure */
typedef struct timecaps_tag {
    UINT    wPeriodMin;     /* minimum period supported  */
    UINT    wPeriodMax;     /* maximum period supported  */
    } TIMECAPS;
typedef TIMECAPS       *PTIMECAPS;
typedef TIMECAPS       *LPTIMECAPS;

UINT WINAPI timeGetDevCaps(TIMECAPS FAR* lpTimeCaps, UINT uSize);

/* timer data types */
typedef void (CALLBACK TIMECALLBACK) (UINT uTimerID,
                                      UINT uMessage,
                                      DWORD dwUser,
                                      DWORD dw1,
                                      DWORD dw2);

typedef TIMECALLBACK FAR *LPTIMECALLBACK;

/* flags for wFlags parameter of timeSetEvent() function */
#define TIME_ONESHOT    0   /* program timer for single event */
#define TIME_PERIODIC   1   /* program for continuous periodic event */

extern UINT WINAPI timeSetEvent( UINT uDelay,
                                 UINT uResolution,
                                 LPTIMECALLBACK lpTimeProc,
                                 DWORD dwUser,
                                 UINT fuEvent );

extern UINT WINAPI timeKillEvent( UINT uTimerID );
extern UINT WINAPI timeBeginPeriod(UINT uPeriod);
extern UINT WINAPI timeEndPeriod(UINT uPeriod);

#ifdef __cplusplus
}
#endif
#endif

static UINT uTimerID = 0;
static UINT uPeriod = 0;
static timer_callback *user_callback = NULL;

static VOID CALLBACK my_callback( UINT uTimerID, UINT uMessage, DWORD dwUser, DWORD dw1, DWORD dw2 )
{
  if(user_callback) (*user_callback)();
}

unsigned              __OsSetTimerCallBack(unsigned ms,timer_callback func)
{
  TIMECAPS tc;
  UINT cur_period;
   if(!uTimerID)
   {
     user_callback = func;
     if(timeGetDevCaps(&tc,sizeof(TIMECAPS)) == 0)
     {
       cur_period = tc.wPeriodMin;
       uPeriod = ms / cur_period;
       if(uPeriod < ms) uPeriod += cur_period-1;
       if(uPeriod > tc.wPeriodMax) uPeriod = tc.wPeriodMax-1;
       if(timeBeginPeriod(uPeriod) == 0)
       {
         uTimerID = timeSetEvent(uPeriod,0,my_callback,0L,TIME_PERIODIC);
         return uPeriod;
       }
     }
   }
   return 0;
}
                             /* Restore time callback function to original
                                state */
void                   __OsRestoreTimer(void)
{
  if(uTimerID)
  {
    timeEndPeriod(uPeriod);
    timeKillEvent(uTimerID);
    uTimerID = 0;
  }
}
