/****************************************************************************\
*   VIO - this source file is part of Binary vIEW project (BIEW)             *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1995, 2000 Nick Kurshev                            *
*   License:              See below                                          *
*   Author and developer: Nick Kurshev                                       *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   lib/sysdep/i386/win32/vio.c                        *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  This file contain implementation of video subsystem handles for Win32s.   *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/
#include "windows.h"
#include "__os_dep.h"
#include "_sys_dep.h"
#include <stdlib.h>
#include <string.h>

unsigned short tvioWidth,tvioHeight,tvioNumColors;
static CONSOLE_SCREEN_BUFFER_INFO csbinfo;
static HANDLE hOut;
static unsigned long vio_flags;
int    win32_use_ansi;
static int is_winnt;
extern OSVERSIONINFO win32_verinfo;

void __init_vio( unsigned long flg )
{
  hOut = GetStdHandle(STD_OUTPUT_HANDLE);
  is_winnt = win32_verinfo.dwPlatformId == VER_PLATFORM_WIN32_NT;
  vio_flags = flg;
  __vioRereadState();
}

void __term_vio( void )
{
}

void __vioRereadState( void )
{
  win32_use_ansi = 0;
  switch(GetConsoleOutputCP())
  {
     case 1251:  win32_use_ansi = 1; break;
     default:    break;
  }
  GetConsoleScreenBufferInfo(hOut,&csbinfo);
  if((vio_flags & __TVIO_FLG_DIRECT_CONSOLE_ACCESS) == __TVIO_FLG_DIRECT_CONSOLE_ACCESS)
  {
    tvioWidth  = csbinfo.dwSize.X;
    tvioHeight = csbinfo.dwSize.Y;
  }
  else
  {
    tvioWidth  = csbinfo.srWindow.Right-csbinfo.srWindow.Left+1;
    tvioHeight = csbinfo.srWindow.Bottom-csbinfo.srWindow.Top+1;
  }
  tvioNumColors = 16;
}

int __vioGetCursorType( void )
{
  CONSOLE_CURSOR_INFO cci;
  GetConsoleCursorInfo(hOut,&cci);
  if(cci.bVisible == FALSE) return CUR_OFF;
  if(cci.dwSize == 100) return CUR_SOLID;
  return CUR_NORM;
}

/* type: 0 - off  1 - normal  2 - solid */
void __vioSetCursorType(int type)
{
  CONSOLE_CURSOR_INFO cci;
  GetConsoleCursorInfo(hOut,&cci);
  cci.bVisible = TRUE;
  switch(type)
  {
    case CUR_OFF: cci.bVisible = FALSE;
                  break;
    case CUR_SOLID:
                  cci.dwSize = 100;
                  break;
    default:
                  cci.dwSize = 20;
  }
  SetConsoleCursorInfo(hOut,&cci);
}

void __vioGetCursorPos(int *x,int *y)
{
  CONSOLE_SCREEN_BUFFER_INFO csbinfo;
  GetConsoleScreenBufferInfo(hOut,&csbinfo);
  *x = csbinfo.dwCursorPosition.X;
  *y = csbinfo.dwCursorPosition.Y;
}

void __vioSetCursorPos(int x,int y)
{
  COORD cc;
  cc.X = x;
  cc.Y = y;
  SetConsoleCursorPosition(hOut,cc);
}

/*
 * Here we support two versions of console i/o.
 * Because of some Win95 releases have bug in
 * - (Read)WriteConsoleOutput
 * But Win2000 (without SP) has bug in
 * - (Read)WriteConsoleOutputCharacter
 * - (Read)WriteConsoleOutputAttribute
 * functions. Sorry! Life ain't easy with MS.
*/

void __vioWriteBuff(int x,int y,void *buff,unsigned len)
{
  unsigned j,i,rlen;
  rlen = min(len/2,__TVIO_MAXSCREENWIDTH);
  if(is_winnt)
  {
    CHAR_INFO obuff[__TVIO_MAXSCREENWIDTH];
    COORD pos,size;
    SMALL_RECT sr;
    for(j = i = 0;i < rlen;i++,j+=2)
    {
      unsigned char attr;
      obuff[i].Char.AsciiChar = ((unsigned char *)buff)[j];
      attr = ((unsigned char *)buff)[j+1];
      obuff[i].Attributes = 0;
      if(attr & 0x01) obuff[i].Attributes |= FOREGROUND_BLUE;
      if(attr & 0x02) obuff[i].Attributes |= FOREGROUND_GREEN;
      if(attr & 0x04) obuff[i].Attributes |= FOREGROUND_RED;
      if(attr & 0x08) obuff[i].Attributes |= FOREGROUND_INTENSITY;
      if(attr & 0x10) obuff[i].Attributes |= BACKGROUND_BLUE;
      if(attr & 0x20) obuff[i].Attributes |= BACKGROUND_GREEN;
      if(attr & 0x40) obuff[i].Attributes |= BACKGROUND_RED;
      if(attr & 0x80) obuff[i].Attributes |= BACKGROUND_INTENSITY;
    }
    pos.X = pos.Y = 0;
    size.X = rlen;
    size.Y = 1;
    sr.Left = x;
    sr.Top = y;
    sr.Right = x + rlen;
    sr.Bottom = y;
    WriteConsoleOutput(hOut,obuff,size,pos,&sr);
  }
  else
  {
    unsigned char chars[__TVIO_MAXSCREENWIDTH];
    unsigned short attr[__TVIO_MAXSCREENWIDTH];
    unsigned long w;
    COORD cc;
    cc.X = x;
    cc.Y = y;
    for(j = i = 0;i < rlen;i++,j+=2)
    {
      chars[i] = ((unsigned char *)buff)[j];
      attr[i] = ((unsigned char *)buff)[j+1];
    }
    WriteConsoleOutputCharacter(hOut,chars,rlen,cc,(LPDWORD)&w);
    WriteConsoleOutputAttribute(hOut,attr,rlen,cc,(LPDWORD)&w);
  }
}

void __vioReadBuff(int x,int y,void *buff,unsigned len)
{
  unsigned j,i,rlen;
  rlen = min(len/2,__TVIO_MAXSCREENWIDTH);
  if(is_winnt)
  {
    CHAR_INFO obuff[__TVIO_MAXSCREENWIDTH];
    COORD pos,size;
    SMALL_RECT sr;
    pos.X = pos.Y = 0;
    size.X = rlen;
    size.Y = 1;
    sr.Left = x;
    sr.Top = y;
    sr.Right = x + rlen;
    sr.Bottom = y;
    ReadConsoleOutput(hOut,obuff,size,pos,&sr);
    for(j = i = 0;i < rlen;i++,j+=2)
    {
      unsigned char attr;
      ((unsigned char *)buff)[j] = obuff[i].Char.AsciiChar;
      attr = obuff[i].Attributes;
      ((unsigned char *)buff)[j+1] = 0;
      if((attr & FOREGROUND_BLUE) == FOREGROUND_BLUE) ((unsigned char *)buff)[j+1] |= 0x01;
      if((attr & FOREGROUND_GREEN) == FOREGROUND_GREEN) ((unsigned char *)buff)[j+1] |= 0x02;
      if((attr & FOREGROUND_RED) == FOREGROUND_RED) ((unsigned char *)buff)[j+1] |= 0x04;
      if((attr & FOREGROUND_INTENSITY) == FOREGROUND_INTENSITY) ((unsigned char *)buff)[j+1] |= 0x08;
      if((attr & BACKGROUND_BLUE) == BACKGROUND_BLUE) ((unsigned char *)buff)[j+1] |= 0x10;
      if((attr & BACKGROUND_GREEN) == BACKGROUND_GREEN) ((unsigned char *)buff)[j+1] |= 0x20;
      if((attr & BACKGROUND_RED) == BACKGROUND_RED) ((unsigned char *)buff)[j+1] |= 0x40;
      if((attr & BACKGROUND_INTENSITY) == BACKGROUND_INTENSITY) ((unsigned char *)buff)[j+1] |= 0x80;
    }
  }
  else
  {
    unsigned char chars[__TVIO_MAXSCREENWIDTH];
    unsigned short attr[__TVIO_MAXSCREENWIDTH];
    unsigned long r;
    COORD cc;
    cc.X = x;
    cc.Y = y;
    ReadConsoleOutputCharacter(hOut,chars,rlen,cc,(LPDWORD)&r);
    ReadConsoleOutputAttribute(hOut,attr,rlen,cc,(LPDWORD)&r);
    for(j = i = 0;i < rlen;i++,j+=2)
    {
      ((unsigned char *)buff)[j] = chars[i];
      ((unsigned char *)buff)[j+1] = attr[i];
    }
  }
}

