/****************************************************************************\
*   FILEIO - this source file is part of Binary vIEW project (BIEW)          *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1995, 2000 Nick Kurshev                            *
*   License:              See below                                          *
*   Author and developer: Nick Kurshev                                       *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   lib/sysdep/i86/dos/fileio.c                        *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  This file contain implementation of file i/o routines for DOS.            *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/
#include <dos.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include "__os_dep.h"
#include "_sys_dep.h"

static void __NEAR__ fnUnix2Dos(char *fn)
{
  char *ptr;
  ptr = fn;
  while((ptr = strchr(ptr,'/')) != NULL) *ptr = '\\';
}

int __OsCreate(const char *name)
{
  union REGS reg;
  struct SREGS sreg;
  memset(&reg,0,sizeof(union REGS));
  memset(&sreg,0,sizeof(struct SREGS));
  fnUnix2Dos(name);
  reg.h.ah = 0x3C;
  reg.x.dx = FP_OFF(name);
  sreg.ds = FP_SEG(name);
  int86x(0x21,&reg,&reg,&sreg);
  if(reg.x.flags & 0x01) { errno = reg.x.ax; reg.x.ax = -1; }
  return reg.x.ax;
}

int __OsDelete(const char *name)
{
  union REGS reg;
  struct SREGS sreg;
  memset(&reg,0,sizeof(union REGS));
  memset(&sreg,0,sizeof(struct SREGS));
  fnUnix2Dos(name);
  reg.h.ah = 0x41;
  reg.x.dx = FP_OFF(name);
  sreg.ds = FP_SEG(name);
  int86x(0x21,&reg,&reg,&sreg);
  if(reg.x.flags & 0x01) { errno = reg.x.ax; reg.x.ax = -1; }
  else reg.x.ax = 0;
  return reg.x.ax;
}

int __OsRename(const char *oldname,const char *newname)
{
  union REGS reg;
  struct SREGS sreg;
  memset(&reg,0,sizeof(union REGS));
  memset(&sreg,0,sizeof(struct SREGS));
  fnUnix2Dos(oldname);
  fnUnix2Dos(newname);
  reg.h.ah = 0x56;
  reg.x.dx = FP_OFF(oldname);
  reg.x.di = FP_OFF(newname);
  sreg.ds = FP_SEG(oldname);
  sreg.es = FP_SEG(newname);
  int86x(0x21,&reg,&reg,&sreg);
  if(reg.x.flags & 0x01) { errno = reg.x.ax; reg.x.ax = -1; }
  else reg.x.ax = 0;
  return reg.x.ax;
}

void __OsClose(int handle)
{
  union REGS reg;
  memset(&reg,0,sizeof(union REGS));
  reg.h.ah = 0x3E;
  reg.x.bx = handle;
  int86(0x21,&reg,&reg);
  if(reg.x.flags & 0x01) errno = reg.x.ax;
}

int __OsDupHandle(int handle)
{
  union REGS reg;
  memset(&reg,0,sizeof(union REGS));
  reg.h.ah = 0x45;
  reg.x.bx = handle;
  int86(0x21,&reg,&reg);
  if(reg.x.flags & 0x01) { errno = reg.x.ax; reg.x.ax = -1; }
  return reg.x.ax;
}

int  __OsOpen(const char *fname,int mode)
{
  union REGS reg;
  struct SREGS sreg;
  memset(&reg,0,sizeof(union REGS));
  memset(&sreg,0,sizeof(struct SREGS));
  fnUnix2Dos(fname);
  reg.x.ax = 0x6C00;
  reg.h.bl = mode;
  reg.h.bh = 0x20; /* internal possible optimization: not commit on every write
                      return media i/o error through errno but not 24h
                      FAT32 size <= 2GB  */
  reg.h.dl = 0x01; /* fail if not exists, open only if exists */
  sreg.ds = FP_SEG(fname);
  reg.x.si = FP_OFF(fname);
  int86x(0x21,&reg,&reg,&sreg);
  if(reg.x.flags & 0x01)
  {
    errno = reg.x.ax;
    reg.x.ax = -1;
  }
  return reg.x.ax;
}

long __OsSeek(int handle,long offset,int origin)
{
  union REGS reg;
  long ret;
  memset(&reg,0,sizeof(union REGS));
  reg.h.ah = 0x42;
  reg.h.al = origin;
  reg.x.bx = handle;
  reg.x.cx = (unsigned short)((offset >> 16) & 0xFFFF);
  reg.x.dx = (unsigned short)(offset & 0xFFFF);
  int86(0x21,&reg,&reg);
  if(reg.x.flags & 0x01)
  {
    errno = reg.x.ax;
    ret = -1;
  }
  else ret = ((long)reg.x.dx << 16) + reg.x.ax;
  return ret;
}

int __OsTruncFile( int handle, unsigned long newsize)
{
  __OsSeek(handle,newsize,0);
  return __OsWrite(handle,NULL,0);
}

int  __OsRead(int handle, void *buff, unsigned size)
{
  union REGS reg;
  struct SREGS sreg;
  memset(&reg,0,sizeof(union REGS));
  memset(&sreg,0,sizeof(struct SREGS));
  reg.h.ah = 0x3F;
  reg.x.bx = handle;
  reg.x.cx = size;
  sreg.ds = FP_SEG(buff);
  reg.x.dx = FP_OFF(buff);
  int86x(0x21,&reg,&reg,&sreg);
  if(reg.x.flags & 0x01)
  {
    errno = reg.x.ax;
    reg.x.ax = 0;
  }
  return reg.x.ax;
}

int  __OsWrite(int handle,const void *buff, unsigned size)
{
  union REGS reg;
  struct SREGS sreg;
  memset(&reg,0,sizeof(union REGS));
  memset(&sreg,0,sizeof(struct SREGS));
  reg.h.ah = 0x40;
  reg.x.bx = handle;
  reg.x.cx = size;
  sreg.ds = FP_SEG(buff);
  reg.x.dx = FP_OFF(buff);
  int86x(0x21,&reg,&reg,&sreg);
  if(reg.x.flags & 0x01)
  {
    errno = reg.x.ax;
    reg.x.ax = 0;
  }
  return reg.x.ax;
}

#define BLKSIZE 32767

int __OsChSize(int handle, long size)

{
    long length, fillsize;
    int  bufsize, numtowrite;
    char *buf;

    length=__OsSeek(handle, 0L, SEEKF_END);	   /* get current length */
    if(length == -1 || size < 0)           return -1;
    if(length >= size)			   /* truncate size */
    {
	return __OsTruncFile(handle,size);
    }
    fillsize=size-length;		    /* increase size **/
    bufsize= (int) min(fillsize, BLKSIZE);
    buf=malloc(bufsize);
    if(buf == NULL) return -1;
    memset(buf, 0, bufsize);		    /* write zeros to pad file */
    do
    {
	numtowrite= (int) min(fillsize, (long) bufsize);
	if(__OsWrite(handle, buf, numtowrite) != numtowrite)
	{
	    free(buf);
	    return -1;
	}
	fillsize-=numtowrite;

    } while(fillsize);
    free(buf);
    return 0;
}
