/****************************************************************************\
*   KEYBOARD - this source file is part of Binary vIEW project (BIEW)        *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1995, 2000 Nick Kurshev                            *
*   License:              See below                                          *
*   Author and developer: Nick Kurshev                                       *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   lib/sysdep/i86/dos/keyboard.c                      *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  This file contain implementation of keyboard handles for DOS.             *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/
#include <dos.h>
#include <string.h>
#include <limits.h>
#include "__os_dep.h"
#include "_sys_dep.h"
#include "kbd_code.h"

static int __ms_nbtns;

void __init_keyboard( void )
{
   if((__ms_nbtns = __init_mouse()) != INT_MAX)
   {
     __MsSetState(1);
   }
}

void __term_keyboard( void )
{
  if(__ms_nbtns != INT_MAX)
  {
    __MsSetState(0);
    __term_mouse();
  }
}

static int __NEAR__ __normalize_code(int code)
{
  switch(code)
  {
    case 0x50E0: return 0x5000; /* down arrow right(gray) */
    case 0x48E0: return 0x4800; /* up  arrow right(gray) */
    case 0x4B30: return 0x4B00; /* letf arrow right(gray) */
    case 0x4D30: return 0x4D00; /* right arrow right(gray) */
    default: return code;
  }
}

static int _mou_btns = 0;
static unsigned _mou_x = 0,_mou_y = 0;

static int __NEAR__ isMouseEventPresent( unsigned long flg, int flush_queue )
{
  int _btns;
  unsigned x,y;
  if(__ms_nbtns == INT_MAX) return 0;
  _btns = __MsGetBtns();
  __MsGetPos(&x,&y);
  if(_btns != _mou_btns ||
     x != _mou_x ||
     y != _mou_y ||
     (((flg & KBD_NONSTOP_ON_MOUSE_PRESS) == KBD_NONSTOP_ON_MOUSE_PRESS) && _btns))
  {
    if(flush_queue)
    {
     _mou_btns = _btns;
     _mou_x = x;
     _mou_y = y;
    }
    return 1;
  }
  return 0;
}

static int __old_ks = 0;

static int __NEAR__ isShiftKeysChange( int flush_queue )
{
  int __ks;
  __ks = __kbdGetShiftsKey();
  if(__ks != __old_ks)
  {
    if(flush_queue) __old_ks = __ks;
    return 1;
  }
  return 0;
}

static int __NEAR__ __test_key( unsigned long flg, int flush_queue )
{
  union REGS reg;
  int ret;
  memset(&reg,0,sizeof(union REGS));
  reg.h.ah = 1;
  int86(0x16,&reg,&reg);
  if((reg.x.flags & 0x40) == 0x40) /* If Zero Flags */
  {
    ret = 0;
    if(isMouseEventPresent(flg,flush_queue)) return KE_MOUSE;
    if(isShiftKeysChange(flush_queue))   return KE_SHIFTKEYS;
  }
  else ret = reg.x.ax;
  return __normalize_code(ret);
}


int __kbdTestKey( unsigned long flg )
{
  return __test_key(flg,0);
}

int __kbdGetShiftsKey( void )
{
  union REGS reg;
  int ret;
  memset(&reg,0,sizeof(union REGS));
  reg.h.ah = 2;
  int86(0x16,&reg,&reg);
  ret = reg.x.ax;
  if(ret & 0x01) ret |= 3; /* return right shift as letf */
  return ret;
}

int __kbdGetKey ( unsigned long flg )
{
  int key;
  while(1)
  {
    key = __test_key(flg,1);
    if(key) break;
    __OsYield();
  }
  if(!(key == KE_MOUSE || key == KE_SHIFTKEYS))
  {
    union REGS reg;
    memset(&reg,0,sizeof(reg));
    reg.h.ah = 0;
    int86(0x16,&reg,&reg);
    key = __normalize_code(reg.x.ax);
  }
  return key;
}
