/****************************************************************************\
*   TWIN - this header file is part of Binary vIEW project (BIEW)            *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1995, 2000 Nick Kurshev                            *
*   License:              See below                                          *
*   Author and developer: Nick Kurshev                                       *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   lib/twin.h                                         *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  This file contain prototypes of Text Window manager functions.            *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/

#ifndef __TWIN_H
#define __TWIN_H 1

#ifdef __cplusplus
extern "C" {
#endif

typedef enum
{
  Black = 0,
  Blue,
  Green,
  Cyan,
  Red,
  Magenta,
  Brown,
  LightGray,
  Gray,
  LightBlue,
  LightGreen,
  LightCyan,
  LightRed,
  LightMagenta,
  Yellow,
  White
}Color;

/* Flags of window */

#define TWS_NONE           0x0000
#define TWS_FRAMEABLE      0x0001
#define TWS_VISIBLE        0x0002
#define TWS_CURSORABLE     0x0004
#define TWS_NLS866         0x0100

/* Flags of title/footer modes */

#define TW_TMODE_LEFT         0x0000
#define TW_TMODE_CENTER       0x0001
#define TW_TMODE_RIGHT        0x0002

#define TW_SINGLE_FRAME "Ŀ"
#define TW_DOUBLE_FRAME "ͻͼ"

#define TW_THIN_FRAME   TW_SINGLE_FRAME
#define TW_MEDIUM_FRAME ""
#define TW_THICK_FRAME  ""

#define TW_UP3D_FRAME   (char *)0
#define TW_DN3D_FRAME   (char *)1

typedef unsigned char ColorAttr;

typedef struct tagDefColor
{
  ColorAttr user;
  ColorAttr system;
}DefColor;

typedef struct tagTWindow
{
  int      X1,Y1,X2,Y2; /* coordinates of window on the screen */
  int      cur_x,cur_y; /* coordinates of cursor position inside the window */
  long     iflags;      /* contain internal flags of window state */
  void *   usrData;     /* user data pointer can contain in windows */
  char *   body;        /* Buffer, containing the image of window frame */
  char *   saved;       /* Buffer, containing the saved image under window */
  int      wsize;       /* Size of buffers in bytes */
  int      wwidth;      /* width of window */
  int      wheight;     /* height of window */
  char *   Title;       /* Caption of window */
  char     TitleMode;   /* Mode displaying of title */
  char *   Footer;      /* Footer of window */
  char     FooterMode;  /* Mode displaying of footer */
  char     Frame[8];    /* Buffer, contaning frame component */
                        /* Frame encoding:   1---2---3 */
                        /*                   |       | */
                        /*                   4       5 */
                        /*                   |       | */
                        /*                   6---7---8 */
  short    flags;       /* Window flags */
  DefColor text;        /* default color of text */
  DefColor frame;       /* default color of frame */
  DefColor title;       /* default color of title */
  DefColor footer;      /* default color of footer text */
  struct   tagTWindow *next;
}TWindow;

#define TW_FRAME_UP        0x00
#define TW_FRAME_DOWN      0x01
#define TW_FRAME_3DMASK    0x0F
#define TW_FRAME_THIN      0x00
#define TW_FRAME_MEDIUM    0x10
#define TW_FRAME_THICK     0x20
#define TW_FRAME_HEIGHTMASK 0xF0
void twinDraw3DFrame(int x1, int y1, int x2, int y2,int flags);


/* Call this function before any other */

#define TWIF_FORCEMONO   0x00000001L

void              twInit( unsigned long vio_flags, unsigned long twin_flags );
/* Call this function after all other */
void              twDestroy( void );

TWindow *         twCreateWin(int x1, int y1, int width, int height, int flags);
void              twDestroyWin(TWindow *);

void              twHideWin(TWindow *);
void              twShowWin(TWindow *);
void              twShowWinBeneath(TWindow *win,TWindow *prev);
void              twShowWinOnTop(TWindow *win);
void              twSnapShot(TWindow *win);
TWindow *         twUseWin(TWindow *);
TWindow *         twUsedWin( void );
void              twMoveWin(TWindow *,int dx, int dy);
void              twResizeWin(TWindow *,int width, int height);
void              twCentredWin(TWindow *it,TWindow *parent); /* parent may be NULL */
void              twFreezeWin(TWindow *);
void              twRefreshWin(TWindow *);
void              twRefreshFullWin(TWindow *);

void *            twGetUsrData(TWindow *);
void *            twSetUsrData(TWindow *,void *);

void              twGetWinPos(TWindow *win,int *x1,int *y1,int *x2,int *y2);
TWindow *         twGetWinAtPos(int x, int y);

#define TW_DEF_FILLER 0x20     /* default filler char */

void              twClearWin(void);
void              twClearWinEx(unsigned char filler);
void              twClrEOL(void);
void              twClrEOLEx(unsigned char filler);

ColorAttr         twSetColor(Color fore,Color back);
ColorAttr         twGetColor( void );
Color             twTextColor( Color col );
Color             twTextBkGnd( Color col );
void              twGetFrame(TWindow *win,char *frame,Color* fore,Color* back);
int               twGetTitle(TWindow *win,char* title,unsigned cb_title,Color* fore,Color* back);
int               twGetFooter(TWindow *win,char* footer,unsigned cb_footer,Color* fore,Color* back);
void              twSetFrame(TWindow *win,char *frame,Color fore,Color back);
void              twSetTitle(TWindow *win,char* title,int titlemode,Color fore,Color back);
void              twSetFooter(TWindow *win,char* footer,int footermode,Color fore,Color back);

void              twGotoXY(int x,int y);
int               twWhereX( void );
int               twWhereY( void );
void              twPutChar(char ch);
char              twGetChar( void );
int               twPutS(const char *);
int               twDirectWrite(int x,int y,const void *,unsigned len);
int               twPrintF(const char *fmt,...);
void              twWriteBuffer(TWindow *win,int x,int y,const void *,unsigned len);
void              twReadBuffer(TWindow *win,int x,int y,void *,unsigned len);

#ifdef __cplusplus
}
#endif

#endif


