/****************************************************************************\
*   MAINLOOP - this source file is part of Binary vIEW project (BIEW)        *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1995, 2000 Nick Kurshev                            *
*   License:              See below                                          *
*   Author and developer: Nick Kurshev                                       *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   mainloop.c                                         *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  This file is analog of message loop routine.                              *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/
#include <string.h>

#include "kbd_code.h"
#include "bmfile.h"
#include "tstrings.h"
#include "reg_form.h"
#include "biewutil.h"
#include "biewhelp.h"
#include "bconsole.h"
#include "codeguid.h"
#include "__os_dep.h"
#include "setup.h"

extern unsigned long LastOffset;

extern char *pnp;
extern void ShowPnP( void );
extern unsigned long WhereAMI(unsigned long fpos);
extern void Evalutor( void );
extern unsigned int drawText( unsigned int shift , int keycode );
extern unsigned int textmaxlen;
extern unsigned int  strmaxlen;
extern void ShowASCII( void );
extern void AsmRef( void );
extern int FStore( void );
extern int FRestore( void );
extern void DigConv( void );
extern void ShowImpProcLXLE( void );
extern int NewSource( void );
extern int FileUtils( void );

unsigned long lastbyte;
unsigned long OldCurrFilePos; /* means previous File position */
unsigned long CurrStrLen = 0;
unsigned long PrevStrLen = 2;
unsigned long CurrPageSize = 0;
unsigned long PrevPageSize = 0;

unsigned long FoundTextSt = 0,FoundTextEnd = 0;

int textshift = 0;

extern void ShowModRef( void );
extern unsigned long ShowResNam( void );
extern unsigned long ShowNResNm( void );
extern unsigned long ShowSegDef( void );
extern unsigned long ShowEntries( void );
extern unsigned long ShowResources( void );
extern unsigned long ShowObjects( void );
extern unsigned long ShowMapTableLX( void );
extern unsigned long ShowMapTableLE( void );
extern unsigned long Search( int mod );
extern unsigned long ShowARCHHeader( void );
extern unsigned long ShowSecHdrElf(void);
extern void          MakeExp( void );

extern void EditBin( void );
extern void EditHex( void );
extern void EditAsm( void );
extern void History( void );

extern void chattr(char x,char y,char len,char attr);

int isHOnLine(unsigned long cp,int width)
{
  if(FoundTextSt == FoundTextEnd) return 0;
  return (FoundTextSt >= cp && FoundTextSt < cp + width)
          || (FoundTextEnd > cp && FoundTextEnd < cp + width)
          || (FoundTextSt <= cp && FoundTextEnd >= cp + width);
}

void HiLightSearch(TWindow *out,unsigned long cfp,char minx,char maxx,char y,char *buff,unsigned flags)
{
 unsigned char i,j;
 char width,attr;
 unsigned __len;
 int x;
 char rbuff[161];
 if((flags & HLS_USE_DOUBLE_WIDTH) == HLS_USE_DOUBLE_WIDTH)
                                                 width = maxx*2;
 else                                            width = maxx - minx;
 attr = LightCyan + (Blue << 4);
 if((flags & HLS_USE_BUFFER_AS_VIDEO) == HLS_USE_BUFFER_AS_VIDEO)
           memcpy(rbuff,buff,width*2);
 else
 {
   memset(rbuff,attr,width*2);
   for(i = 0,j = 0;j < width;i+=2,j++) rbuff[i] = buff[j];
 }
 x = (int)(FoundTextSt - cfp);
 if((flags & HLS_USE_DOUBLE_WIDTH) == HLS_USE_DOUBLE_WIDTH) x *= 2;
 __len = (unsigned)(FoundTextEnd - FoundTextSt);
 if((flags & HLS_USE_DOUBLE_WIDTH) == HLS_USE_DOUBLE_WIDTH) __len *= 2;
 if(__len > width - x) __len = width - x;
 if(x < 0) { __len += x; x = 0; }
 if(__len && x + __len <= width)
 {
   unsigned char end,st;
   st = x*2;
   end = (__len + x)*2;
   attr = Black + (LightGray << 4);
   for(i = st;i < end;i+=2)
                           rbuff[i + 1] = attr;
 }
 twWriteBuffer(out,minx + 1,y + 1,rbuff,width);
}

static void __NEAR__ drawTitle( void )
{
  unsigned percent;
  unsigned long flen;
  flen = BMGetFLength();
  if(flen) percent = (unsigned)(( lastbyte*100 )/ flen );
  else percent = 100;
  if(percent > 100) percent = 100;
  twUseWin(TitleWnd);
  twGotoXY(Width-4,1);
  twPrintF("%u%%",percent);
  twClrEOL();
}

char legalchars[] = "+-0123456789ABCDEFabcdef";

void MainLoop( void )
{
 int ch;
 unsigned long savep = 0,cfp,flen;
 long nfp;
 unsigned long lwidth;
 BMSeek(LastOffset,BM_SEEK_SET);
 drawPrompt();
 twUseWin(MainWnd);
 textshift = activeMode->paint(KE_SUPERKEY,textshift);
 BMSeek(LastOffset,BM_SEEK_SET);
 drawTitle();
 while(1)
 {
  unsigned che;
  ch = GetEvent(drawPrompt,NULL);
  nfp = cfp = OldCurrFilePos = BMGetCurrFilePos();
  flen = BMGetFLength();
  lwidth = activeMode->CurLineWidth();
  che = ch & 0x00FF;
  if((che >= '0' && che <= '9') ||
     (che >= 'A' && che <= 'Z') ||
     (che >= 'a' && che <= 'z') ||
     ch == KE_BKSPACE)
     {
       nfp = GidGetGoAddress(ch);
       goto GO;
     }
  switch(ch)
  {
    case KE_CTL_F(1):
    case KE_CTL_F(2):
    case KE_CTL_F(3):
    case KE_CTL_F(4):
    case KE_CTL_F(5):
    case KE_CTL_F(6):
    case KE_CTL_F(7):
    case KE_CTL_F(8):
    case KE_CTL_F(9):
    case KE_CTL_F(10):
                    {
                       unsigned i;
                       i = (ch - KE_CTL_F(1)) >> 8;
                       if(activeMode->action[i])
                       {
                         if(activeMode->action[i]()) { ch = KE_SUPERKEY; drawPrompt(); }
                       }
                     }
                     break;
    case KE_ALT_F(1):
    case KE_ALT_F(2):
    case KE_ALT_F(3):
    case KE_ALT_F(4):
    case KE_ALT_F(5):
    case KE_ALT_F(6):
    case KE_ALT_F(7):
    case KE_ALT_F(8):
    case KE_ALT_F(9):
    case KE_ALT_F(10):
                      {
                        unsigned i;
                        i = (ch - KE_ALT_F(1)) >> 8;
                        if(detectedFormat->action[i]) nfp = detectedFormat->action[i]();
                      }
                      break;
    case KE_SUPERKEY: goto DRAW;
    case KE_F(1) : About();  continue;
    default : continue;
    case KE_SHIFT_F(1): activeMode->help();
                  break;
    case KE_F(10):
    case KE_ESCAPE : return;
    case KE_ENTER:
                  QuickSelectMode();
                  drawPrompt();
                  ch = KE_SUPERKEY;
                  break;
    case KE_F(2):
                  if(SelectMode()) ch = KE_SUPERKEY;
                  break;
    case KE_F(3):
                  if(NewSource())
                  {
                    ch = KE_SUPERKEY;
                    FoundTextSt = FoundTextEnd; ch = KE_SUPERKEY;
                    PaintTitle();
                  }
                  break;
    case KE_F(4):
                  if(activeMode->misc_action)
                  {
                     unsigned long sfp;
                     sfp = BMGetCurrFilePos();
                     activeMode->misc_action();
                     ch = KE_SUPERKEY;
                     PaintTitle();
                     drawPrompt();
                     BMSeek(sfp,BM_SEEK_SET);
                  }
                  break;
    case KE_F(5):
           {
             static unsigned long shift = 0,flags = GJDLG_ABSOLUTE;
             if(GetJumpDlg(&shift,&flags))
             {
               switch(flags)
               {
                 default:
                 case GJDLG_ABSOLUTE: nfp = shift;
                                      break;
                 case GJDLG_RELATIVE: nfp += (long)shift;
                                      break;
                 case GJDLG_VIRTUAL:
                                      if(detectedFormat->va2pa)
                                      {
                                        unsigned long temp_fp;
                                        temp_fp = detectedFormat->va2pa(shift);
                                        if(!temp_fp) ErrMessageBox(NOT_ENTRY,NULL);
                                        else nfp = temp_fp;
                                      }
                                      else nfp = shift;
                                      break;
               }
               GidAddBackAddress();
               ch = KE_SUPERKEY;
             }
           }
           break;
    case KE_SHIFT_F(5): nfp = WhereAMI(nfp);
                        break;
    case KE_F(6): BMReRead();
             FoundTextSt = FoundTextEnd; ch = KE_SUPERKEY;
             PaintTitle();
             break;
    case KE_SHIFT_F(6): SelectSysInfo(); break;
    case KE_F(7): nfp = Search(0); ch = KE_JUSTFIND; break;
    case KE_SHIFT_F(7) : nfp = Search(1); ch = KE_JUSTFIND; break;
    case KE_F(8):
           {
             if(detectedFormat->showHdr) nfp = detectedFormat->showHdr();
             else if(FType > NOT_EXE)
             {
               if(FType == DOS_SYS) nfp = ShowSysHeader();
               else                 nfp = ShowMZHeader();
             }
           }
           break;
    case KE_SHIFT_F(8): SelectTool(); break;
    case KE_F(9): Setup(); break;
    case KE_SHIFT_F(10): if(FileUtils())
                   {
                      FoundTextSt = FoundTextEnd; ch = KE_SUPERKEY;
                      PaintTitle();
                   }
                   break;
    case KE_HOME: textshift = 0; break;
    case KE_END:  textshift = strmaxlen - tvioWidth + 2; break;
    case KE_UPARROW:
                   nfp = cfp - activeMode->PrevLineWidth();
                   break;
    case KE_DOWNARROW:
                     nfp = cfp + activeMode->CurLineWidth();
                     break;
    case KE_RIGHTARROW:
                     if(!activeMode->is_text) nfp = cfp + activeMode->get_symbol_size();
                     else                      textshift+=activeMode->get_symbol_size();
                     break;
    case KE_LEFTARROW:
                     if(!activeMode->is_text) nfp = cfp - activeMode->get_symbol_size();
                     else                     textshift-=activeMode->get_symbol_size();
                     if(textshift < 0) textshift = 0;
                     break;
    case KE_CTL_RIGHTARROW:
                     if(!activeMode->is_text) nfp = cfp + 8*activeMode->get_symbol_size();
                     else                      textshift+=8*activeMode->get_symbol_size();
                     break;
    case KE_CTL_LEFTARROW:
                     if(!activeMode->is_text) nfp = cfp - 8*activeMode->get_symbol_size();
                     else                     textshift-=8*activeMode->get_symbol_size();;
                     if(textshift < 0) textshift = 0;
                     break;
    case KE_PGUP:
                    nfp = cfp - activeMode->PrevPageSize();
                    break;
    case KE_PGDN:
                      nfp = cfp + activeMode->CurPageSize();
                      break;
    case KE_CTL_PGUP: nfp = 0;
                   break;
    case KE_CTL_PGDN:
                   nfp = flen;
                   break;
    case KE_CTL_(O): /* User screen */
                   {
                     twHideWin(MainWnd);
                     twHideWin(TitleWnd);
                     while(GetEvent(drawEmptyPrompt,NULL) != KE_CTL_(O));
                     twShowWin(MainWnd);
                     twShowWin(TitleWnd);
                   }
                   continue;

  }
  GO:
  if(cfp != nfp)
  {
    unsigned long twidth = ( activeMode->is_text ) ? 1 : lwidth;
    unsigned long p = flen - twidth;
    if(nfp > 0) if(nfp > p) nfp = p;
    if(nfp < 0) nfp = 0;
    BMSeek(nfp,BM_SEEK_SET);
  }
  DRAW:
  twUseWin(MainWnd);
  if(!activeMode->is_text) savep = BMGetCurrFilePos();
  textshift = activeMode->paint(ch,textshift);
  if(!activeMode->is_text) BMSeek(savep,BM_SEEK_SET);
  drawTitle();
 }
}


