##############################################################################
#   MAKEFILE.INC - this file is part of Binary vIEW project (BIEW)           #
##############################################################################
#   Copyrights:           1998, 2000 Nick Kurshev                            #
#   License:              See below                                          #
#   Author and developer: Nick Kurshev                                       #
#   Requirement:          GNU make                                           #
#   Original file name:   makefile                                           #
####################### [ D e s c r i p t i o n ] ############################
#  This include file is script for make utility of GNU development system.   #
########################### [ L i c e n c e ] ################################
# The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 #
# All rights reserved. This software is redistributable under the licence    #
# given in the file "Licence.en" ("Licence.ru" in russian translation)       #
# distributed in the BIEW archive.                                           #
##############################################################################

##############################################################################
# ATTENTION!!! I use manual method configuring of BIEW project rather than   #
# 'configure' script technology, because 'configure' script does not work    #
# propertly on non *NIX systems (like Cygwin, EMX,...).                      #
##############################################################################

##############################################################################
# HOST determines host operating system                                      #
##############################################################################
# Please write name of any desired operating system if this name is present  #
# in lib/sysdep/$(MACHINE) as subfolder, or port BIEW for your OS manually   #
# and please bring to author your version (with changes list)                #
# NOTE: I use file name model: 8.3 + all character in lowercase +            #
#       no symlinks (project must be compatible with DOS compilers)          #
##############################################################################
ifdef CONFIG_DOS
HOST=dos
CDEFOS=-D__MSDOS__ -D__OS_NAME__='"DOS"'
OSLDEF=--force-exe-suffix
OS_LIBS=
endif
ifdef CONFIG_OS2
HOST=os2
CDEFOS=-D__OS2__ -D__OS_NAME__='"OS/2"' -Zomf
# here I want have static link of glibc instead emx.dll
OSLDEF=-Zexe -Zsmall-conv -Zsys -Zomf -Zstack 1024
OS_LIBS=
endif
ifdef CONFIG_WIN32
HOST=win32
# here I link crtdll.dll instead cygwin1.dll (it's safe)
CDEFOS=-D__WIN32__ -D__OS_NAME__='"Win32"' -mno-cygwin
OSLDEF= -mno-cygwin --force-exe-suffix
OS_LIBS= -lwinmm
endif

ifdef CONFIG_LINUX

HOST=linux
CDEFOS=-D__LINUX__ -D__OS_NAME__='"Linux"' -D_VT100_
OSLDEF=
OS_LIBS=

ifdef USE_MOUSE
CDEFOS += -DHAVE_MOUSE
OS_LIBS += -lgpm
endif

endif

ifdef CONFIG_UNIX

#
# screen library to use: SLANG/CURSES/VT100
#

USE_SLANG = y
#USE_CURSES = y
#USE_VT100 = y

#

USE_SELECT = y	# system has select() call?
#USE_MOUSE = y	# try to use mouse (doesnot work yet)


HOST=unix
CDEFOS = -D__UNIX__
OSLDEF=

ifdef USE_SLANG
CDEFOS += -D_SLANG_
OS_LIBS += -lslang
OSNAME=Unix/Slang
endif
ifdef USE_CURSES
CDEFOS += -D_CURSES_
OS_LIBS += -lncurses
OSNAME=Unix/Curses
endif
ifdef USE_VT100
CDEFOS += -D_VT100_
OSNAME=Unix/VT100
endif

CDEFOS += -D__OS_NAME__='"$(OSNAME)"'

endif	# CONFIG_UNIX


ifdef USE_SELECT
CDEFOS += -DHAVE_SELECT
endif
ifdef USE_MOUSE
CDEFOS += -DHAVE_MOUSE
endif

##############################################################################
# MACHINE determines hardware platform                                       #
##############################################################################
# WARNING!!!                                                                 #
# BIEW is designed for i86/i386 platforms and was not tested on BIG ENDIAN   #
# platforms. Project also was not tested on non 16/32 bits systems.          #
# If you want you may port project by hand and please bring to author your   #
# version (with changes list)                                                #
##############################################################################
ifdef CONFIG_I386
MACHINE=i386
CDEFSYS=-D__BITNESS__=32 -D__PLATFORM__='"i386"' -D__BYTE_ORDER__=__LITTLE_ENDIAN__ -D__DEFAULT_DISASM=0 -m386
endif
ifdef CONFIG_I486
MACHINE=i386
CDEFSYS=-D__BITNESS__=32 -D__PLATFORM__='"i486"' -D__BYTE_ORDER__=__LITTLE_ENDIAN__ -D__DEFAULT_DISASM=0 -m486
endif
ifdef CONFIG_I586
MACHINE=i386
CDEFSYS=-D__BITNESS__=32 -D__PLATFORM__='"i586"' -D__BYTE_ORDER__=__LITTLE_ENDIAN__ -D__DEFAULT_DISASM=0 -march=pentium -mcpu=pentium -fschedule-insns2
endif
ifdef CONFIG_I686
MACHINE=i386
CDEFSYS=-D__BITNESS__=32 -D__PLATFORM__='"i686"' -D__BYTE_ORDER__=__LITTLE_ENDIAN__ -D__DEFAULT_DISASM=0 -march=pentiumpro -mcpu=pentiumpro -fschedule-insns2
endif
ifdef CONFIG_AMD_K6
MACHINE=i386
CDEFSYS=-D__BITNESS__=32 -D__PLATFORM__='"amdk6"' -D__BYTE_ORDER__=__LITTLE_ENDIAN__ -D__DEFAULT_DISASM=0 -march=k6 -mcpu=k6 -fschedule-insns2
endif
ifdef CONFIG_I86
MACHINE=i86
CDEFSYS=-D__BITNESS__=16 -D__PLATFORM__='"i86"' -D__BYTE_ORDER__=__LITTLE_ENDIAN__ -D__DEFAULT_DISASM=0
endif
ifdef CONFIG_GENERIC
MACHINE=generic
CDEFSYS=-D__BITNESS__=32 -D__PLATFORM__='"generic"' -D__BYTE_ORDER__=__BIG_ENDIAN__ -D__DEFAULT_DISASM=1
endif
##############################################################################
# If you do not have GNU development system, modify RM CC LD variables       #
##############################################################################
RM=rm -f
CC=gcc
LD=gcc
###########################################################################
#  CFLAGS set it for optimization                                         #
#  LDFLAGS set it for link method                                         #
###########################################################################
CFLAGS = -Wall -O2 -fomit-frame-pointer -fcaller-saves\
 -DNDEBUG=1 $(CDEFOS) $(CDEFSYS)
#-ffunction-sections
LDFLAGS = -s $(OSLDEF)

