/****************************************************************************\
*   COFF386 - this source file is part of Binary vIEW project (BIEW)         *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1995, 2000 Nick Kurshev                            *
*   License:              See below                                          *
*   Author and developer: Nick Kurshev                                       *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   plugins/bin/coff386.c                              *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  This file contain implementation of COFF-i386 file format decoder.        *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/
#include <stddef.h>
#include <stdio.h>

#include "reg_form.h"
#include "bconsole.h"
#include "bin_util.h"
#include "bmfile.h"
#include "kbd_code.h"
#include "biewutil.h"
#include "biewhelp.h"
#include "plugins/bin/coff386.h"
#include "__os_dep.h"

static char * __NEAR__ coff386_encode_hdr(unsigned short info)
{
   switch(info)
   {
     case I386MAGIC: return " i386 coff ";
     case I386PTXMAGIC: return " i386 PTX coff ";
     case I386AIXMAGIC: return " i386 AIX coff ";
     case LYNXCOFFMAGIC: return " Lynx coff ";
     default:    return " Unknow coff i386 format ";
   }
}

unsigned long ShowCoff386Header( void )
{
  struct external_filehdr coff386;
  unsigned long fpos;
  unsigned keycode;
  TWindow * w;
  fpos = BMGetCurrFilePos();
  BMReadBufferEx(&coff386,sizeof(struct external_filehdr),0,SEEKF_START);
  w = CrtHlpWnd(coff386_encode_hdr(*((unsigned short *)&coff386.f_magic)),54,11);
  twGotoXY(1,1);
  twPrintF("Number of sections          = %04XH\n"
           "Time & data stamp           = %08lXH\n"
           "File pointer to symtab      = %08lXH\n"
           "Number of symtab entries    = %08lXH\n"
           "Size of optional header     = %04XH\n"
           "COFF flags                  = %04XH\n"
           "  [%c] - Rel. info stripped\n"
           "  [%c] - File is executable (no unresolv. refs.)\n"
           "  [%c] - Line numbers stripped\n"
           "  [%c] - Local symbols stripped"
           ,*((unsigned short *)&coff386.f_nscns)
           ,*((unsigned long *)&coff386.f_timdat)
           ,*((unsigned long *)&coff386.f_symptr)
           ,*((unsigned long *)&coff386.f_nsyms)
           ,*((unsigned short *)&coff386.f_opthdr)
           ,*((unsigned short *)&coff386.f_flags)
           ,GetBool((*((unsigned long *)&coff386.f_flags) & F_RELFLG) == F_RELFLG)
           ,GetBool((*((unsigned long *)&coff386.f_flags) & F_EXEC) == F_EXEC)
           ,GetBool((*((unsigned long *)&coff386.f_flags) & F_LNNO) == F_LNNO)
           ,GetBool((*((unsigned long *)&coff386.f_flags) & F_LSYMS) == F_LSYMS));
  while(1)
  {
    keycode = GetEvent(drawEmptyPrompt,w);
    if(keycode == KE_ENTER) { /* fpos = *((unsigned long *)&aout.e_entry); */; break; }
    else
      if(keycode == KE_ESCAPE) break;
  }
  CloseWnd(w);
  BMSeek(fpos,BM_SEEK_SET);
  return fpos;
}


int coff386_check_fmt( void )
{
  unsigned short id;
  id = BMReadWordEx(0,SEEKF_START);
  return !(I386BADMAG(id));
}

void coff386_init_fmt( void ) {}
void coff386_destroy_fmt() {}

int coff386_bitness(unsigned long off) { return DAB_USE32; }

void coff386_AddrResolv(char *addr,unsigned long cfpos)
{
  if(cfpos < sizeof(struct external_filehdr))
  {
    sprintf(addr,"coffih:%s",Get2Digit(cfpos));
  }
}

unsigned long coff386Help( void )
{
  hlpDisplay(10002);
  return BMGetCurrFilePos();
}

REGISTRY_BIN coff386Table =
{
  "coff-i386 Common Object File Format)",
  { "CofHlp", NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL },
  { coff386Help, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL },
  coff386_check_fmt,
  coff386_init_fmt,
  coff386_destroy_fmt,
  ShowCoff386Header,
  NULL,
  NULL,
  coff386_bitness,
  coff386_AddrResolv,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL
};

