/****************************************************************************\
*   ELF386 - this source file is part of Binary vIEW project (BIEW)          *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1995, 2000 Nick Kurshev                            *
*   License:              See below                                          *
*   Author and developer: Nick Kurshev                                       *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   plugins/bin/elf386.c                               *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  This file contain implementation of ELF-32 file format decoder.           *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/
#include <limits.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "plugins/bin/elf386.h"
#include "bin_util.h"
#include "pmalloc.h"
#include "kbd_code.h"
#include "codeguid.h"
#include "bmfile.h"
#include "biewhelp.h"
#include "tstrings.h"
#include "biewutil.h"
#include "bconsole.h"
#include "reg_form.h"
#include "__os_dep.h"

static Elf386_External_Ehdr elf;
static unsigned active_shtbl = 0;
static unsigned long elf_min_va = ULONG_MAX;

struct tag_elfVAMap
{
  unsigned long va;
  unsigned long size;
  unsigned long foff;
};

struct tag_ELFPubName
{
   unsigned long  pa;
   unsigned long  nameoff;
   unsigned short active_shtbl;
   unsigned short attr;
};

static linearArray *elfPubNames = NULL;

static int __NEAR__ FindPubName(char *buff,unsigned cb_buff,unsigned long pa);
static void __NEAR__ elf_ReadPubNameList(BGLOBAL handle,int interact);

static BGLOBAL namecache = &bNull;
static BGLOBAL namecache2 = &bNull;
static BGLOBAL elfcache = &bNull;

static linearArray *va_map;

static int vamap_comp(const void __HUGE__ *v1,const void __HUGE__ *v2)
{
  struct tag_elfVAMap __HUGE__ *pnam1,__HUGE__ *pnam2;
  pnam1 = (struct tag_elfVAMap __HUGE__ *)v1;
  pnam2 = (struct tag_elfVAMap __HUGE__ *)v2;
  if(pnam1->va < pnam2->va) return -1;
  else
    if(pnam1->va > pnam2->va) return 1;
    else                      return 0;
}

unsigned long elfVA2PA(unsigned long va)
{
  unsigned long i;
  for(i = 0; i < va_map->nItems;i++)
  {
    struct tag_elfVAMap __HUGE__ *evm;
    evm = &((struct tag_elfVAMap __HUGE__ *)va_map->data)[i];
    if(va >= evm->va && va < evm->va + evm->size) return evm->foff + (va - evm->va);
  }
  return 0L;
}

unsigned long elfPA2VA(unsigned long pa)
{
  unsigned long i,ret;
  ret = 0L;
  for(i = 0; i < va_map->nItems;i++)
  {
    struct tag_elfVAMap __HUGE__ *evm;
    evm = &((struct tag_elfVAMap __HUGE__ *)va_map->data)[i];
    if(pa >= evm->foff && pa < evm->foff + evm->size)
    {
      ret = evm->va + (pa - evm->foff);
      break;
    }
  }
  return ret;
}

int elfLowMemFunc( unsigned long need_mem )
{
  if(!fmtActiveState)
  {
    if(elfPubNames)
    {
       la_Destroy(elfPubNames);
       elfPubNames = NULL;
       return 1;
    }
  }
  return 0;
}

static void __NEAR__ elf386_readnametable(unsigned long off,char *buf,unsigned blen)
{
  unsigned long foff;
  Elf386_External_Shdr sh;
  unsigned char ch;
  unsigned freq;
  BGLOBAL b_cache,b_cache2;
  b_cache = namecache;
  b_cache2 = namecache2;
  foff = *((unsigned long *)&elf.e_shoff)+(*((unsigned short *)&elf.e_shstrndx)*sizeof(Elf386_External_Shdr));
  bioSeek(b_cache2,foff,SEEKF_START);
  bioReadBuffer(b_cache2,&sh,sizeof(Elf386_External_Shdr));
  foff = *((unsigned long *)&sh.sh_offset) + off;
  freq = 0;
  while(1)
  {
     bioSeek(b_cache,foff++,SEEKF_START);
     ch = bioReadByte(b_cache);
     buf[freq++] = ch;
     if(!ch || freq >= blen) break;
  }
}

static void __NEAR__ elf386_readnametableex(unsigned long off,char *buf,unsigned blen)
{
  unsigned long foff;
  Elf386_External_Shdr sh;
  unsigned char ch;
  unsigned freq;
  BGLOBAL b_cache,b_cache2;
  b_cache = namecache;
  b_cache2 = namecache2;
  foff = *((unsigned long *)&elf.e_shoff)+active_shtbl*sizeof(Elf386_External_Shdr);
  bioSeek(b_cache2,foff,SEEKF_START);
  bioReadBuffer(b_cache2,&sh,sizeof(Elf386_External_Shdr));
  foff = *((unsigned long *)&sh.sh_offset) + off;
  freq = 0;
  while(1)
  {
     bioSeek(b_cache,foff++,SEEKF_START);
     ch = bioReadByte(b_cache);
     buf[freq++] = ch;
     if(!ch || freq >= blen) break;
  }
}

static char * __NEAR__ elf_class(char id)
{
  switch(id)
  {
    default: return "Unk.obj/";
    case ELFCLASSNONE: return "Err.obj/";
    case ELFCLASS32: return "32bit/";
  }
}

static char * __NEAR__ elf_data(char id)
{
  switch(id)
  {
    default: return "Unk.data/";
    case ELFDATANONE: return "Err.data/";
    case ELFDATA2LSB: return "little endian/";
    case ELFDATA2MSB: return "big endian/";
  }
}

static char *__NEAR__ elf_otype(unsigned short id)
{
  switch(id)
  {
    case ET_NONE: return "none";
    case ET_REL: return "relocatable";
    case ET_EXEC: return "executable";
    case ET_DYN: return "shared object";
    case ET_CORE: return "core";
    case ET_NUM: return "number of defined types";
    case ET_LOPROC: return "low processor-specific";
    case ET_HIPROC: return "high processor-specific";
    default: return "Unknown";
  }
}

static char *__NEAR__ elf_machine(unsigned short id)
{
  switch(id)
  {
    case EM_NONE: return "none";
    case EM_M32: return "AT&T WE32100";
    case EM_SPARC: return "Sun Sparc";
    case EM_386: return "Intel 386";
    case EM_68K: return "Motorola m68k family";
    case EM_88K: return "Motorola m88k family";
    case EM_860: return "Intel 80860";
    case EM_MIPS: return "MIPS R3000";
    case EM_MIPS_RS4_BE: return "MIPS R4000";
    case EM_SPARC64: return "Sparc v9 64-bit";
    case EM_PARISC: return "HPPA";
    case EM_SPARC32PLUS: return "Sun's \"v8plus\"";
    case EM_PPC: return "Power PC";
    case EM_SH: return "Hitachi SH";
    default: return "Unknown";
  }
}

static char *__NEAR__ elf_ver(long id)
{
  switch(id)
  {
    case EV_NONE:    return "Invalid ELF version";
    case EV_CURRENT: return "Current version";
    default:         return "Unknown";
  }
}

unsigned long ShowELFHeader( void )
{
  unsigned long fpos;
  TWindow *w;
  char hdr[81];
  char vs[2];
  unsigned keycode;
  unsigned long entrya;
  entrya = elfVA2PA(*((unsigned long *)&elf.e_entry));
  fpos = BMGetCurrFilePos();
  strcpy(hdr," Executable and Linkable File (ELF) ");
  strcat(hdr,elf_class(elf.e_ident[4]));
  strcat(hdr,elf_data(elf.e_ident[5]));
  strcat(hdr,"ver.");
  vs[1] = 0;
  vs[0] = elf.e_ident[6]+'0';
  strcat(hdr,vs);
  strcat(hdr," ");
  w = CrtHlpWnd(hdr,74,13);
  twGotoXY(1,1);
  twPrintF("Object file type                  = %s\n"
           "Machine architecture              = %s\n"
           "Object file version               = %s\n"
           ,elf_otype(*((unsigned short*)(&elf.e_type)))
           ,elf_machine(*((unsigned short*)(&elf.e_machine)))
           ,elf_ver(*((unsigned long*)(&elf.e_version))));
  twSetColor(White,Black);
  twPrintF("Entry point VA                    = %08lXH (Offset: %08lXH)"
           ,*((unsigned long*)(&elf.e_entry)),entrya);
  twClrEOL(); twPrintF("\n");
  twSetColor(Black,LightGray);
  twPrintF("Program header table offset       = %08lXH\n"
           "Section header table offset       = %08lXH\n"
           "Processor specific flag           = %08lXH\n"
           "ELF header size (bytes)           = %04XH\n"
           "Program header table entry size   = %04XH\n"
           "Program header table entry count  = %04XH\n"
           "Section header table entry size   = %04XH\n"
           "Section header table entry count  = %04XH\n"
           "Section header string table index = %04XH"
           ,*((unsigned long*)(&elf.e_phoff))
           ,*((unsigned long*)(&elf.e_shoff))
           ,*((unsigned long*)(&elf.e_flags))
           ,*((unsigned short*)(&elf.e_ehsize))
           ,*((unsigned short*)(&elf.e_phentsize))
           ,*((unsigned short*)(&elf.e_phnum))
           ,*((unsigned short*)(&elf.e_shentsize))
           ,*((unsigned short*)(&elf.e_shnum))
           ,*((unsigned short*)(&elf.e_shstrndx)));
  while(1)
  {
    keycode = GetEvent(drawEmptyPrompt,w);
    if(keycode == KE_ENTER) { fpos = entrya; break; }
    else
      if(keycode == KE_ESCAPE) break;
  }
  CloseWnd(w);
  return fpos;
}

static char * __NEAR__ elf_encode_p_type(long p_type)
{
   switch(p_type)
   {
      case PT_NULL: return "Unusable";
      case PT_LOAD: return "Loadable";
      case PT_DYNAMIC: return "Dynalinking";
      case PT_INTERP: return "Interpreter";
      case PT_NOTE:  return "Auxiliary";
      case PT_SHLIB: return "Unspecified";
      case PT_PHDR:  return "header itself";
      case PT_NUM: return "Number of types";
      case PT_LOPROC: return "Low processor";
      case PT_HIPROC: return "High processor";
      default:  return "Unknown";
   }
}

static int __elfReadPrgHdr(BGLOBAL handle,memArray *obj,unsigned nnames)
{
 int i;
  bioSeek(handle,*((unsigned long *)&elf.e_phoff),SEEKF_START);
  for(i = 0;i < nnames;i++)
  {
   unsigned long fp;
   char stmp[80];
   Elf386_External_Phdr phdr;
   if(IsKbdTerminate()) break;
   fp = bioTell(handle);
   bioReadBuffer(handle,&phdr,sizeof(Elf386_External_Phdr));
   bioSeek(handle,fp+(*((unsigned short *)&elf.e_phentsize)),SEEKF_START);
   sprintf(stmp,"%-15s %08lX %08lX %08lX %08lX %08lX %c%c%c %08lX",
                elf_encode_p_type(*((long *)&phdr.p_type)),
                *((long *)&phdr.p_offset),
                *((long *)&phdr.p_vaddr),
                *((long *)&phdr.p_paddr),
                *((long *)&phdr.p_filesz),
                *((long *)&phdr.p_memsz),
                (*((long *)&phdr.p_flags) & PF_X) == PF_X ? 'X' : ' ',
                (*((long *)&phdr.p_flags) & PF_W) == PF_W ? 'W' : ' ',
                (*((long *)&phdr.p_flags) & PF_R) == PF_R ? 'R' : ' ',
                *((long *)&phdr.p_align)
          );
   if(!ma_AddString(obj,stmp,1)) break;
  }
  return 1;
}

static char * __NEAR__ elf_encode_sh_type(long p_type)
{
   switch(p_type)
   {
      case SHT_NULL: return "NULL";
      case SHT_PROGBITS: return "PRGBTS";
      case SHT_SYMTAB: return "SYMTAB";
      case SHT_STRTAB: return "STRTAB";
      case SHT_RELA:  return "RELA";
      case SHT_HASH: return "HSHTAB";
      case SHT_DYNAMIC:  return "DYNLNK";
      case SHT_NOTE: return "NOTES";
      case SHT_NOBITS: return "NOBITS";
      case SHT_REL:  return "REL";
      case SHT_SHLIB: return "UNSPEC";
      case SHT_DYNSYM: return "DYNSYM";
      case SHT_NUM: return "NTYPES";
      case SHT_GNU_verdef: return "VERDEF";
      case SHT_GNU_verneed: return "VERNED";
      case SHT_GNU_versym: return "SYMVER";
      case SHT_LOPROC: return "LOPROC";
      case SHT_HIPROC: return "HIPROC";
      case SHT_LOUSER: return "LOUSER";
      case SHT_HIUSER: return "HIUSER";
      default:  return "UNK";
   }
}

static int __elfReadSecHdr(BGLOBAL handle,memArray *obj,unsigned nnames)
{
 int i;
  bioSeek(handle,*((unsigned long *)&elf.e_shoff),SEEKF_START);
  for(i = 0;i < nnames;i++)
  {
   Elf386_External_Shdr shdr;
   char tmp[17];
   unsigned long fp;
   char stmp[80];
   if(IsKbdTerminate()) break;
   fp = bioTell(handle);
   bioReadBuffer(handle,&shdr,sizeof(Elf386_External_Shdr));
   elf386_readnametable(*((long *)&shdr.sh_name),tmp,sizeof(tmp));
   bioSeek(handle,fp+(*((unsigned short *)&elf.e_shentsize)),SEEKF_START);
   tmp[16] = 0;
   sprintf(stmp,"%-16s %-6s %c%c%c %08lX %08lX %08lX %04hX %04hX %04hX %04hX",
                tmp,
                elf_encode_sh_type(*((long *)&shdr.sh_type)),
                (*((long *)&shdr.sh_flags) & SHF_WRITE) == SHF_WRITE ? 'W' : ' ',
                (*((long *)&shdr.sh_flags) & SHF_ALLOC) == SHF_ALLOC ? 'A' : ' ',
                (*((long *)&shdr.sh_flags) & SHF_EXECINSTR) == SHF_EXECINSTR ? 'X' : ' ',
                *((long *)&shdr.sh_addr),
                *((long *)&shdr.sh_offset),
                *((long *)&shdr.sh_size),
                (short)*((long *)&shdr.sh_link),
                (short)*((long *)&shdr.sh_info),
                (short)*((long *)&shdr.sh_addralign),
                (short)*((long *)&shdr.sh_entsize)
          );
    if(!ma_AddString(obj,stmp,1)) break;
  }
  return 1;
}

static char * __NEAR__ elf_SymTabType(char type)
{
  switch(ELF_ST_TYPE(type))
  {
    case STT_NOTYPE:  return "NoType";
    case STT_OBJECT:  return "Object";
    case STT_FUNC:    return "Func. ";
    case STT_SECTION: return "Sect. ";
    case STT_FILE:    return "File  ";
    case STT_NUM:     return "Number";
    case STT_LOPROC:  return "LoProc";
    case STT_HIPROC:  return "HiProc";
    default: return "Unknwn";
  }
}

static char * __NEAR__ elf_SymTabBind(char type)
{
  switch(ELF_ST_BIND(type))
  {
    case STB_LOCAL:  return "Local ";
    case STB_GLOBAL: return "Global";
    case STB_WEAK:   return "Weak  ";
    case STB_NUM:    return "Number";
    case STB_LOPROC: return "LoProc";
    case STB_HIPROC: return "HiProc";
    default: return "Unknwn";
  }
}

static char * __NEAR__ elf_SymTabShNdx(unsigned idx)
{
  static char ret[10];
  switch(idx)
  {
    case SHN_UNDEF:  return "Undef ";
    case SHN_LOPROC: return "LoProc";
    case SHN_HIPROC: return "HiProc";
    case SHN_ABS:    return "Abs.  ";
    case SHN_COMMON: return "Common";
    case SHN_HIRESERVE: return "HiRes.";
    default: sprintf(ret,"%04XH ",idx); return ret;
  }
}

static unsigned __elfNumSymTab = 0;
static unsigned long __elfEntSize = 0;
static unsigned long __elfSymPtr = 0;

static unsigned __elfGetNumSymTab( BGLOBAL handle )
{
  return __elfNumSymTab;
}

static int __elfReadSymTab(BGLOBAL handle,memArray *obj,unsigned nsym)
{
 int i;
 char text[31];
  bioSeek(handle,__elfSymPtr,SEEK_SET);
  for(i = 0;i < nsym;i++)
  {
   unsigned long fp;
   char stmp[80];
   Elf386_External_Sym sym;
   if(IsKbdTerminate()) break;
   fp = bioTell(handle);
   bioReadBuffer(handle,&sym,sizeof(Elf386_External_Sym));
   bioSeek(handle,fp+__elfEntSize,SEEKF_START);
   elf386_readnametableex(*((long *)&sym.st_name),text,sizeof(text));
   text[sizeof(text)-1] = 0;
   sprintf(stmp,"%-31s %08lX %08lX %04hX %s %s %s"
               ,text
               ,*((long *)&sym.st_value)
               ,*((long *)&sym.st_size)
               ,*((short *)&sym.st_other)
               ,elf_SymTabType(sym.st_info[0])
               ,elf_SymTabBind(sym.st_info[0])
               ,elf_SymTabShNdx(*((unsigned short *)&sym.st_shndx))
               );
   if(!ma_AddString(obj,stmp,1)) break;
  }
  return 1;
}

static int __NEAR__ __elfReadDynTab(BGLOBAL handle,memArray *obj, unsigned ntbl,unsigned long entsize)
{
 int i,j;
 char sout[80];
 unsigned len,rlen;
  for(i = 0;i < ntbl;i++)
  {
   unsigned long fp;
   char stmp[80];
   Elf386_External_Dyn pdyn;
   fp = bioTell(handle);
   bioReadBuffer(handle,&pdyn,sizeof(Elf386_External_Dyn));
   bioSeek(handle,fp+entsize,SEEKF_START);
   fp = bioTell(handle);
   elf386_readnametableex(*((unsigned long *)&pdyn.d_tag),sout,sizeof(sout));
   len = strlen(sout);
   if(len > 56) len = 53;
   if(IsKbdTerminate()) break;
   strncpy(stmp,sout,len);
   stmp[len] = 0;
   if(len > 56) strcat(stmp,"...");
   rlen = strlen(stmp);
   for(j = rlen;j < 60;j++) strcat(stmp," ");
   sprintf(&stmp[strlen(stmp)]," vma=%08lXH",*((unsigned long *)&pdyn.d_un.d_val));
   if(!ma_AddString(obj,stmp,1)) break;
   bioSeek(handle,fp,SEEKF_START);
  }
  return 1;
}

static unsigned Elf386PrgHdrNumItems(BGLOBAL handle) { return *((unsigned short *)&elf.e_phnum); }

unsigned long ShowPrgHdrElf(void)
{
  unsigned long fpos;
  int ret;
  fpos = BMGetCurrFilePos();
  ret = fmtShowList(Elf386PrgHdrNumItems,__elfReadPrgHdr,
                    " type            fileoffs virtaddr physaddr filesize memsize  flg align   ",
                    LB_SELECTIVE,NULL);
  if(ret != -1)
  {
    Elf386_External_Phdr it;
    BMSeek(*((unsigned long *)&elf.e_phoff)+sizeof(Elf386_External_Phdr)*ret,SEEKF_START);
    BMReadBuffer(&it,sizeof(Elf386_External_Phdr));
    fpos = *((unsigned long *)&it.p_offset);
  }
  return fpos;
}

static unsigned Elf386SecHdrNumItems(BGLOBAL handle) { return *((unsigned short *)&elf.e_shnum); }

unsigned long ShowSecHdrElf(void)
{
  unsigned long fpos;
  int ret;
  fpos = BMGetCurrFilePos();
  ret = fmtShowList(Elf386SecHdrNumItems,__elfReadSecHdr,
                    " name             type   flg virtaddr fileoffs   size   link info algn esiz",
                   LB_SELECTIVE,NULL);
  if(ret != -1)
  {
    Elf386_External_Shdr it;
    BMSeek(*((unsigned long *)&elf.e_shoff)+sizeof(Elf386_External_Shdr)*ret,SEEKF_START);
    BMReadBuffer(&it,sizeof(Elf386_External_Shdr));
    fpos = *((unsigned long *)&it.sh_offset);
  }
  return fpos;
}

static unsigned long __NEAR__ displayELFsymtab( void )
{
  unsigned long fpos;
  int ret;
  fpos = BMGetCurrFilePos();
  ret = fmtShowList(__elfGetNumSymTab,__elfReadSymTab,
                    " Name                            Value    Size     Oth. Type   Bind   Sec# ",
                    LB_SELECTIVE,NULL);
  if(ret != -1)
  {
         Elf386_External_Sym it;
         BMSeek(__elfSymPtr+__elfEntSize*ret,SEEKF_START);
         BMReadBuffer(&it,sizeof(Elf386_External_Sym));
         fpos = elfVA2PA(*((unsigned long *)&it.st_value));
  }
  return fpos;
}

static unsigned long __NEAR__ displayELFdyntab(unsigned long dynptr,unsigned long nitem,long entsize)
{
  unsigned long fpos;
  BGLOBAL handle;
  fpos = BMGetCurrFilePos();
  {
    memArray *obj;
    unsigned ndyn;
    ndyn = (unsigned)nitem;
    if(!(obj = ma_Build(ndyn,1))) return fpos;
    handle = elfcache;
    bioSeek(handle,dynptr,SEEK_SET);
    if(__elfReadDynTab(handle,obj,ndyn,entsize))
    {
      int ret;
      ret = ma_Display(obj," Dynamic section ",LB_SELECTIVE | LB_SORTABLE,-1);
      if(ret != -1)
      {
         char *addr;
         addr = strstr(obj->data[ret],"vma=");
         if(addr)
         {
           unsigned long addr_probe;
           addr_probe = strtoul(&addr[4],NULL,16);
           if(addr_probe && addr_probe >= elf_min_va)
           {
             addr_probe = elfVA2PA(addr_probe);
             if(addr_probe < BMGetFLength()) fpos = addr_probe;
           }
           else goto not_entry;
         }
         else
         {
           not_entry:
           ErrMessageBox(NOT_ENTRY,NULL);
         }
      }
    }
    ma_Destroy(obj);
  }
  return fpos;
}

unsigned long ShowELFSymTab( void )
{
  unsigned long fpos,symptr,number;
  Elf386_External_Shdr shdr;
  unsigned i;
  fpos = BMGetCurrFilePos();
  symptr = 0;
  number = 0;
  for(i = 0;i < *((unsigned short *)&elf.e_shnum);i++)
  {
   BMReadBufferEx(&shdr,sizeof(Elf386_External_Shdr),*((unsigned long *)&elf.e_shoff) + i*(*((unsigned short *)&elf.e_shentsize)),SEEKF_START);
   if(*((unsigned long *)&shdr.sh_type) == SHT_SYMTAB)
   {
     symptr = *((unsigned long *)&shdr.sh_offset);
     number = (*((unsigned long *)&shdr.sh_size)/(*((unsigned long *)&shdr.sh_entsize)));
     active_shtbl = *((short *)&shdr.sh_link);
     break;
   }
  }
  BMSeek(fpos,SEEKF_START);
  if(!symptr) { NotifyBox(NOT_ENTRY," ELF symbol table "); return fpos; }
  __elfSymPtr = symptr;
  __elfNumSymTab = (unsigned)number;
  __elfEntSize = *((long *)&shdr.sh_entsize);
  return displayELFsymtab();
}

unsigned long ShowELFDynSec( void )
{
  unsigned long fpos,dynptr,number;
  Elf386_External_Shdr shdr;
  unsigned i;
  fpos = BMGetCurrFilePos();
  dynptr = 0;
  number = 0;
  for(i = 0;i < *((unsigned short *)&elf.e_shnum);i++)
  {
   BMReadBufferEx(&shdr,sizeof(Elf386_External_Shdr),*((unsigned long *)&elf.e_shoff) + i*(*((unsigned short *)&elf.e_shentsize)),SEEKF_START);
   if(*((unsigned long *)&shdr.sh_type) == SHT_DYNSYM)
   {
     dynptr = *((unsigned long *)&shdr.sh_offset);
     number = (*((unsigned long *)&shdr.sh_size)/(*((unsigned long *)&shdr.sh_entsize)));
     active_shtbl = *((short *)&shdr.sh_link);
     break;
   }
  }
  BMSeek(fpos,SEEKF_START);
  if(!dynptr) { NotifyBox(NOT_ENTRY," ELF dynamic symbol table "); return fpos; }
  return displayELFdyntab(dynptr,number,*((long *)&shdr.sh_entsize));
}

#define S_INTERPRETER "Interpreter : "

static void __NEAR__ displayELFdyninfo(unsigned long f_off,unsigned nitems)
{
  Elf386_External_Dyn dyntab;
  Elf386_External_Phdr phdr;
  unsigned long foff,curroff,stroff;
  unsigned i,is_add;
  memArray * obj;
  char stmp[80];
  foff = BMGetCurrFilePos();
  curroff = f_off;
  stroff = 0;
  BMSeek(curroff,SEEK_SET);
  for(i = 0;i < nitems;i++)
  {
    BMReadBuffer(&dyntab,sizeof(dyntab));
    if(*((unsigned long *)&dyntab.d_tag) == DT_STRTAB)
    {
       stroff = *((unsigned long *)&dyntab.d_un.d_ptr);
       break;
    }
  }
  if(!stroff) { NotifyBox(" String information not found!",NULL); return; }
  stroff = elfVA2PA(stroff);
  f_off = curroff;
  BMSeek(f_off,SEEKF_START);
  if(!(obj = ma_Build(0,1))) return;
  strcpy(stmp,S_INTERPRETER);
  for(i = 0;i < *((unsigned short *)&elf.e_phnum);i++)
  {
   BMReadBufferEx(&phdr,sizeof(Elf386_External_Phdr),*((unsigned long *)&elf.e_phoff) + i*(*((unsigned short *)&elf.e_phentsize)),SEEKF_START);
   if(*((unsigned long *)&phdr.p_type) == PT_INTERP)
   {
     BMReadBufferEx(&stmp[sizeof(S_INTERPRETER) - 1],sizeof(stmp)-sizeof(S_INTERPRETER)-1,
                    *((unsigned long *)&phdr.p_offset),SEEKF_START);
     break;
   }
  }
  if(!ma_AddString(obj,stmp,1)) goto dyn_end;
  BMSeek(f_off,SEEKF_START);
  for(i = 0;i < nitems;i++)
  {
    BMReadBufferEx(&dyntab,sizeof(dyntab),f_off,SEEKF_START);
    f_off += sizeof(dyntab);
    is_add = 1;
    switch(*((unsigned long *)&dyntab.d_tag))
    {
      case DT_NULL: goto dyn_end;
      case DT_NEEDED:
                    {
                      strcpy(stmp,"Needed : ");
                      BMReadBufferEx(&stmp[strlen(stmp)],70,*((unsigned long *)&dyntab.d_un.d_ptr) + stroff,SEEKF_START);
                    }
                    break;
      case DT_SONAME:
                    {
                      strcpy(stmp,"SO name: ");
                      BMReadBufferEx(&stmp[strlen(stmp)],70,*((unsigned long *)&dyntab.d_un.d_ptr) + stroff,SEEKF_START);
                    }
                    break;
      case DT_RPATH:
                    {
                      strcpy(stmp,"LibPath: ");
                      BMReadBufferEx(&stmp[strlen(stmp)],70,*((unsigned long *)&dyntab.d_un.d_ptr) + stroff,SEEKF_START);
                    }
                    break;
       default:     is_add = 0; break;
    }
    if(is_add) if(!ma_AddString(obj,stmp,1)) break;
  }
  dyn_end:
  ma_Display(obj," Dynamic linking information ",LB_SORTABLE,-1);
  ma_Destroy(obj);
  BMSeek(foff,SEEKF_START);
}

unsigned long ShowELFDynInfo( void )
{
  unsigned long fpos,dynptr,number;
  Elf386_External_Phdr phdr;
  unsigned i;
  fpos = BMGetCurrFilePos();
  dynptr = 0;
  number = 0;
  for(i = 0;i < *((unsigned short *)&elf.e_phnum);i++)
  {
   BMReadBufferEx(&phdr,sizeof(Elf386_External_Phdr),*((unsigned long *)&elf.e_phoff) + i*(*((unsigned short *)&elf.e_phentsize)),SEEKF_START);
   if(*((unsigned long *)&phdr.p_type) == PT_DYNAMIC)
   {
     dynptr = *((unsigned long *)&phdr.p_offset);
     number = (*((unsigned long *)&phdr.p_filesz)/sizeof(Elf386_External_Dyn));
     break;
   }
  }
  BMSeek(fpos,SEEKF_START);
  if(!dynptr) { NotifyBox(NOT_ENTRY," ELF dynamic linking information "); return fpos; }
  displayELFdyninfo(dynptr,(unsigned)number);
  BMSeek(fpos,SEEKF_START);
  return fpos;
}

static int AppendELFRef(char *str,unsigned long ulShift,int flags,int codelen,unsigned long r_sh)
{
  char buff[400];
  int ret;
  ret = 0;
  memset(buff,-1,sizeof(buff));
  if(flags & APREF_TRY_LABEL)
  {
     if(!elfPubNames) elf_ReadPubNameList(BMbioHandle(),1);
     if(FindPubName(buff,sizeof(buff),r_sh))
     {
       strcat(str,buff);
       if(!DumpMode && !EditMode) GidAddGoAddress(str,r_sh);
       ret = 1;
     }
  }
  return ret;
}

int IsELF32( void )
{
  char id[4];
  BMReadBufferEx(id,sizeof(id),0,SEEKF_START);
  return id[0] == 0x7F && id[1] == 'E' && id[2] == 'L' && id[3] == 'F'
         &&  BMReadByteEx(4,SEEKF_START) <= ELFCLASS32;
}

void ELFinit( void )
{
 Elf386_External_Phdr phdr;
 BGLOBAL main_handle;
 unsigned va_map_count;
 int i;
   PMRegLowMemCallBack(elfLowMemFunc);
   BMReadBufferEx(&elf,sizeof(Elf386_External_Ehdr),0,SEEKF_START);
   for(i = 0;i < *(unsigned short *)&elf.e_phnum;i++)
   {
     BMReadBufferEx(&phdr,sizeof(Elf386_External_Phdr),*((unsigned long *)&elf.e_phoff),SEEKF_START);
     if(*(unsigned long *)&phdr.p_type == PT_LOAD) break;
   }
   va_map_count = *((unsigned short *)&elf.e_phnum);
   if(!(va_map = la_Build(va_map_count,sizeof(struct tag_elfVAMap),1)))
   {
     exit(EXIT_FAILURE);
   }
   BMSeek(*((unsigned long *)&elf.e_phoff),SEEKF_START);
   for(i = 0;i < va_map_count;i++)
   {
     unsigned long fp;
     Elf386_External_Phdr phdr;
     struct tag_elfVAMap vamap;
     fp = BMGetCurrFilePos();
     BMReadBuffer(&phdr,sizeof(Elf386_External_Phdr));
     BMSeek(fp+(*((unsigned short *)&elf.e_phentsize)),SEEKF_START);
     vamap.va = *(unsigned long *)&phdr.p_vaddr;
     vamap.size = *(unsigned long *)&phdr.p_filesz;
     vamap.foff = *(unsigned long *)&phdr.p_offset;
     if(!elfVA2PA(vamap.va))
       if(!la_AddData(va_map,&vamap,1))
       {
         exit(EXIT_FAILURE);
       }
     /* We must sort va_map after adding of each element because ELF program
        header has unsorted and nested elements */
     la_Sort(va_map,vamap_comp);
   }
   /* Find min value of virtual address */
   for(i = 0; i < va_map->nItems;i++)
   {
     struct tag_elfVAMap __HUGE__ *evm;
     evm = &((struct tag_elfVAMap __HUGE__ *)va_map->data)[i];
     if(evm->va < elf_min_va) elf_min_va = evm->va;
   }
   main_handle = BMbioHandle();
   if((namecache = bioDup(main_handle)) == &bNull) namecache = main_handle;
   if((namecache2 = bioDup(main_handle)) == &bNull) namecache2 = main_handle;
   if((elfcache = bioDup(main_handle)) == &bNull) elfcache = main_handle;
}

void ELFdestroy( void )
{
   BGLOBAL main_handle;
   main_handle = BMbioHandle();
   if(namecache != &bNull && namecache != main_handle) bioClose(namecache);
   if(namecache2 != &bNull && namecache2 != main_handle) bioClose(namecache2);
   if(elfcache != &bNull && elfcache != main_handle) bioClose(elfcache);
   if(elfPubNames) { la_Destroy(elfPubNames); elfPubNames = 0; }
   PMUnregLowMemCallBack(elfLowMemFunc);
   la_Destroy(va_map);
}

int ELFbitness(unsigned long off) { return DAB_USE32; }

unsigned long ELFHelp( void )
{
  hlpDisplay(10003);
  return BMGetCurrFilePos();
}

void ELFAddrResolv(char *addr,unsigned long cfpos)
{
  if(cfpos < sizeof(Elf386_External_Ehdr))
  {
    sprintf(addr,"ELFhdr:%s",Get2Digit(cfpos));
  }
}

static int compare_pubnames(const void __HUGE__ *v1,const void __HUGE__ *v2)
{
  struct tag_ELFPubName __HUGE__ *pnam1,__HUGE__ *pnam2;
  pnam1 = (struct tag_ELFPubName __HUGE__ *)v1;
  pnam2 = (struct tag_ELFPubName __HUGE__ *)v2;
  if(pnam1->pa < pnam2->pa) return -1;
  else
    if(pnam1->pa > pnam2->pa) return 1;
    else                      return 0;
}

static int __NEAR__ FindPubName(char *buff,unsigned cb_buff,unsigned long pa)
{
  struct tag_ELFPubName *ret,key;
  key.pa = pa;
  ret = la_Find(elfPubNames,&key,compare_pubnames);
  if(ret)
  {
    unsigned long fpos;
    active_shtbl = ret->active_shtbl;
    fpos = BMGetCurrFilePos();
    elf386_readnametableex(ret->nameoff,buff,cb_buff);
    BMSeek(fpos,SEEK_SET);
    buff[cb_buff-1] = 0;
    return 1;
  }
  return 0;
}

static void __NEAR__ elf_ReadPubNameList(BGLOBAL handle,int interact)
{
  unsigned long i,tableptr,number,fpos;
  unsigned short pubname_shtbl;
  Elf386_External_Shdr shdr;
  struct tag_ELFPubName epn;
  BGLOBAL b_cache;
  b_cache = handle;
  fpos = bioTell(b_cache);
  tableptr = 0;
  number = 0;
  pubname_shtbl = 0;
  for(i = 0;i < *((unsigned short *)&elf.e_shnum);i++)
  {
   bioSeek(b_cache,*((unsigned long *)&elf.e_shoff) + i*(*((unsigned short *)&elf.e_shentsize)),SEEKF_START);
   bioReadBuffer(b_cache,&shdr,sizeof(Elf386_External_Shdr));
   if(*((unsigned long *)&shdr.sh_type) == SHT_DYNSYM)
   {
     tableptr = *((unsigned long *)&shdr.sh_offset);
     number = (*((unsigned long *)&shdr.sh_size)/(*((unsigned long *)&shdr.sh_entsize)));
     pubname_shtbl = *((short *)&shdr.sh_link);
     break;
   }
  }
  if(!(elfPubNames = la_Build(0,sizeof(struct tag_ELFPubName),interact))) return;
  if(tableptr)
  {
    bioSeek(b_cache,tableptr,SEEK_SET);
    for(i = 0;i < number;i++)
    {
     unsigned long fp;
     Elf386_External_Dyn pdyn;
     fp = bioTell(b_cache);
     bioReadBuffer(b_cache,&pdyn,sizeof(Elf386_External_Dyn));
     bioSeek(b_cache,fp+*((long *)&shdr.sh_entsize),SEEKF_START);
     epn.nameoff = *((unsigned long *)&pdyn.d_tag);
     epn.pa = elfVA2PA(*((unsigned long *)&pdyn.d_un.d_val));
     epn.active_shtbl = pubname_shtbl;
     epn.attr = ELF_ST_INFO(STB_GLOBAL,STT_NOTYPE);
     if(!la_AddData(elfPubNames,&epn,interact)) break;
    }
  }
  /* If present symbolic information we must read it */

  for(i = 0;i < *((unsigned short *)&elf.e_shnum);i++)
  {
   BMReadBufferEx(&shdr,sizeof(Elf386_External_Shdr),*((unsigned long *)&elf.e_shoff) + i*(*((unsigned short *)&elf.e_shentsize)),SEEKF_START);
   if(*((unsigned long *)&shdr.sh_type) == SHT_SYMTAB)
   {
     tableptr = *((unsigned long *)&shdr.sh_offset);
     number = (*((unsigned long *)&shdr.sh_size)/(*((unsigned long *)&shdr.sh_entsize)));
     pubname_shtbl = *((short *)&shdr.sh_link);
     break;
   }
  }
  if(tableptr)
  {
    bioSeek(handle,tableptr,SEEK_SET);
    for(i = 0;i < number;i++)
    {
      unsigned long fp;
      Elf386_External_Sym sym;
      if(IsKbdTerminate()) break;
      fp = bioTell(handle);
      bioReadBuffer(handle,&sym,sizeof(Elf386_External_Sym));
      bioSeek(handle,fp+*((long *)&shdr.sh_entsize),SEEKF_START);
      if((ELF_ST_TYPE(sym.st_info[0]) == STT_OBJECT ||
         ELF_ST_TYPE(sym.st_info[0]) == STT_FUNC) &&
         ELF_ST_BIND(sym.st_info[0]) == STB_LOCAL)
      {
        epn.nameoff = *((long *)&sym.st_name);
        epn.pa = elfVA2PA(*((long *)&sym.st_value));
        epn.active_shtbl = pubname_shtbl;
        epn.attr = sym.st_info[0];
        if(!la_AddData(elfPubNames,&epn,1)) break;
      }
    }
  }
  la_Sort(elfPubNames,compare_pubnames);
  bioSeek(b_cache,fpos,SEEK_SET);
}

unsigned long elfGetPubSym(char *str,unsigned cb_str,unsigned *func_class,
                           unsigned long pa,int as_prev)
{
  unsigned long cfpos,ret_addr;
  unsigned i,idx;
  struct tag_ELFPubName key;
  BGLOBAL b_cache;
  b_cache = elfcache;
  cfpos = BMGetCurrFilePos();
  if(!elfPubNames) elf_ReadPubNameList(b_cache,0);
  if(!elfPubNames->nItems) return 0;
  ret_addr = 0L;
  idx = UINT_MAX;
  *func_class = SC_GLOBAL;
  key.pa = pa;
  i = (unsigned)la_FindNearest(elfPubNames,&key,compare_pubnames);
  if(as_prev) idx = i;
  else
  {
    while(((struct tag_ELFPubName __HUGE__ *)elfPubNames->data)[i].pa <= pa) i++;
    idx = i;
  }
  if(idx < elfPubNames->nItems)
  {
    unsigned attr;
    ret_addr = ((struct tag_ELFPubName __HUGE__ *)elfPubNames->data)[idx].pa;
    attr = ((struct tag_ELFPubName __HUGE__ *)elfPubNames->data)[idx].attr;
    switch(ELF_ST_BIND(attr))
    {
      case       STB_LOCAL: *func_class = SC_LOCAL; break;
      default:
      case       STB_GLOBAL: *func_class = SC_GLOBAL; break;
    }
    if(!idx && pa < ret_addr && as_prev)
    {
      ret_addr = 0;
    }
    else
    {
      active_shtbl = ((struct tag_ELFPubName __HUGE__ *)elfPubNames->data)[idx].active_shtbl;
      elf386_readnametableex(((struct tag_ELFPubName __HUGE__ *)elfPubNames->data)[idx].nameoff,str,cb_str);
    }
  }
  BMSeek(cfpos,SEEK_SET);
  return ret_addr;
}

unsigned elfGetObjAttr(unsigned long pa,char *name,unsigned cb_name,
                       unsigned long *start,unsigned long *end,int *class,int *bitness)
{
  unsigned i,nitems,ret;
  unsigned long cfpos,fp;
  *start = 0;
  *end = BMGetFLength();
  *class = OC_NOOBJECT;
  *bitness = ELFbitness(pa);
  name[0] = 0;
  nitems = *((unsigned short *)&elf.e_shnum);
  ret = 0;
  cfpos = BMGetCurrFilePos();
  BMSeek(*((unsigned long *)&elf.e_shoff),SEEK_SET);
  for(i = 0;i < nitems;i++)
  {
    Elf386_External_Shdr sh;
    if(IsKbdTerminate()) break;
    fp = BMGetCurrFilePos();
    BMReadBuffer(&sh,sizeof(Elf386_External_Shdr));
    BMSeek(fp+(*((unsigned short *)&elf.e_shentsize)),SEEKF_START);
    if(!(*(unsigned long *)&sh.sh_offset && *(unsigned long *)&sh.sh_size)) continue;
    if(pa >= *start && pa < *(unsigned long *)&sh.sh_offset)
    {
      /* means between two objects */
      *end = *(unsigned long *)&sh.sh_offset;
      ret = 0;
      break;
    }
    if(pa >= *(unsigned long *)&sh.sh_offset &&
       pa < *(unsigned long *)&sh.sh_offset + *(unsigned long *)&sh.sh_size)
    {
      *start = *(unsigned long *)&sh.sh_offset;
      *end = *(unsigned long *)&sh.sh_offset + *(unsigned long *)&sh.sh_size;
      if(*(unsigned long *)sh.sh_type == SHT_PROGBITS)
      {
        if((*(unsigned long *)&sh.sh_flags & SHF_EXECINSTR) == SHF_EXECINSTR)
                        *class = OC_CODE;
        else            *class = OC_DATA;
      }
      else              *class = OC_NOOBJECT;
      elf386_readnametable(*((long *)&sh.sh_name),name,cb_name);
      ret = i+1;
      break;
    }
    *start = *(unsigned long *)&sh.sh_offset + *(unsigned long *)&sh.sh_size;
  }
  BMSeek(cfpos,SEEK_SET);
  return ret;
}

REGISTRY_BIN elf386Table =
{
  "ELF-i386 (Executable and Linkable Format)",
  { "ELFhlp", "DynInf", "DynSec", NULL, NULL, NULL, "SymTab", NULL, "SecHdr", "PrgDef" },
  { ELFHelp, ShowELFDynInfo, ShowELFDynSec, NULL, NULL, NULL, ShowELFSymTab, NULL, ShowSecHdrElf, ShowPrgHdrElf },
  IsELF32, ELFinit, ELFdestroy,
  ShowELFHeader,
  AppendELFRef,
  fmtSetState,
  ELFbitness,
  ELFAddrResolv,
  elfVA2PA,
  elfPA2VA,
  elfGetPubSym,
  elfGetObjAttr,
  NULL,
  NULL
};
