/****************************************************************************\
*   LE - this source file is part of Binary vIEW project (BIEW)              *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1995, 2000 Nick Kurshev                            *
*   License:              See below                                          *
*   Author and developer: Nick Kurshev                                       *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   plugins/bin/le.c                                   *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  This file contain implementation of LE (Linear Executable) file format    *
* decoder.                                                                   *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>

#include "plugins/bin/lx_le.h"
#include "bin_util.h"
#include "bmfile.h"
#include "biewhelp.h"
#include "kbd_code.h"
#include "tstrings.h"
#include "biewutil.h"
#include "bconsole.h"
#include "reg_form.h"
#include "__os_dep.h"

static BGLOBAL le_cache = &bNull;

unsigned long ShowNewHeaderLE( void )
{
  return ShowNewHeaderLX();
}

extern char * lxeGetMapAttr(unsigned long attr);

static int __ReadMapTblLE(BGLOBAL handle,memArray * obj,unsigned n)
{
 int i;
  bioSeek(handle,lxe.le.leObjectPageMapTableOffset + headshift,SEEKF_START);
  for(i = 0;i < n;i++)
  {
    LE_PAGE lep;
    char stmp[80];
    if(IsKbdTerminate()) break;
    bioReadBuffer(handle,&lep,sizeof(LE_PAGE));
    sprintf(stmp,"#=%08lXH Flags: %04hX = %s",(long)lep.number,lep.flags,(char *)lxeGetMapAttr(lep.flags));
    if(!ma_AddString(obj,stmp,1)) break;
  }
  return 1;
}

static unsigned long __NEAR__ __calcPageEntryLE(LE_PAGE *mt,unsigned long idx)
{
  unsigned long ret;
  unsigned long dataoff;
  dataoff = idx*lxe.le.lePageSize;
  if(mt->flags == 1 || mt->flags == 5) ret = lxe.le.leObjectIterDataMapOffset + dataoff;
  else
    if(mt->flags == 0) ret = lxe.le.leDataPagesOffset + dataoff;
    else                      ret = -1;
  return ret;
}

unsigned long CalcPageEntryLE(unsigned long pageidx)
{
  BGLOBAL handle;
  int found;
  unsigned i;
  LE_PAGE mt;
  if(!pageidx) return -1;
  handle = le_cache;
  bioSeek(handle,lxe.le.leObjectPageMapTableOffset + headshift,SEEK_SET);
  found = 0;
  for(i = 0;i < lxe.le.lePageCount;i++)
  {
    bioReadBuffer(handle,(void *)&mt,sizeof(LE_PAGE));
    if(mt.number == pageidx)
    {
      found = 1;
      break;
    }
  }
  if(found) return __calcPageEntryLE((void *)&mt,pageidx - 1);
  else      return BMGetCurrFilePos();
}

unsigned long CalcEntryPointLE(unsigned long objnum,unsigned long offset)
{
  BGLOBAL handle;
  unsigned long i,start,ret,pageoff,pidx,j;
  LX_OBJECT lo;
  LE_PAGE mt;
  if(!objnum) return -1;
  handle = le_cache;
  bioSeek(handle,lxe.le.leObjectTableOffset + headshift,SEEK_SET);
  bioSeek(handle,sizeof(LX_OBJECT)*(objnum - 1),SEEKF_CUR);
  bioReadBuffer(handle,(void *)&lo,sizeof(LX_OBJECT));
/*  if((lo.o32_flags & 0x00002000L) == 0x00002000L) USE16 = 0;
  else                                            USE16 = 0xFF; */
  pageoff = lxe.le.leObjectPageMapTableOffset + headshift;
  start = 0;
  ret = -1;
  for(i = 0;i < lo.o32_mapsize;i++)
  {
    if(offset >= start && offset < start + lxe.le.lePageSize)
    {
      int found;
      bioSeek(handle,pageoff,SEEKF_START);
      pidx = i + lo.o32_pagemap;
      found = 0;
      for(j = 0;j < lxe.le.lePageCount;j++)
      {
        bioReadBuffer(handle,(void *)&mt,sizeof(LE_PAGE));
        if(mt.number == pidx) { found = 1; break; }
      }
      if(found) ret = __calcPageEntryLE((void *)&mt,pidx - 1) + offset - start;
      else      ret = BMGetCurrFilePos();
      break;
    }
    start += lxe.le.lePageSize;
  }
  return ret;
}

unsigned long CalcEntryLE(LX_ENTRY *lxent)
{
  unsigned long ret;
  ret = BMGetCurrFilePos();
      switch(lxent->b32_type)
      {
        case 1: ret = CalcEntryPointLE(lxent->b32_obj,lxent->entry.e32_variant.e32_offset.offset16);
                      break;
        case 2: ret = CalcEntryPointLE(lxent->b32_obj,lxent->entry.e32_variant.e32_callgate.offset);
                      break;
        case 3: ret = CalcEntryPointLE(lxent->b32_obj,lxent->entry.e32_variant.e32_offset.offset32);
                      break;
        case 4: ShowFwdModOrdLX(lxent);
        case 5:
        default: break;
      }
  return ret;
}

static unsigned long __NEAR__ CalcEntryBungleLE(unsigned ordinal,int dispmsg)
{
  BGLOBAL handle;
  int found;
  unsigned i;
  unsigned char j;
  unsigned char cnt,type;
  unsigned short numobj = 0;
  LX_ENTRY lxent;
  unsigned long ret;
  ret = BMGetCurrFilePos();
  handle = le_cache;
  bioSeek(handle,lxe.le.leEntryTableOffset + headshift,SEEK_SET);
  i = 0;
  found = 0;
  while(1)
  {
   cnt = bioReadByte(handle);
   type = bioReadByte(handle);
   if(!cnt) break;
   if(type) numobj = bioReadWord(handle);
   for(j = 0;j < cnt;j++,i++)
   {
     char size;
     switch(type)
     {
       case 0: size = 0; break;
       case 1: size = 2; break;
       case 2:
       case 0x80:
       case 3: size = 4; break;
       default:
       case 4: size = 6; break;
     }
     if(i == ordinal - 1)
     {
       lxent.b32_type = type;
       found = 1;
       if(size)
       {
         lxent.b32_obj = numobj;
         lxent.entry.e32_flags = bioReadByte(handle);
         bioReadBuffer(handle,(void *)&lxent.entry.e32_variant,size);
       }
       break;
     }
     else if(size) bioSeek(handle,size + sizeof(char),SEEKF_CUR);
   }
   if(found == 1) break;
 }
 if(found) ret = CalcEntryLE((LX_ENTRY *)&lxent);
 else      if(dispmsg) ErrMessageBox("No such entry in the Table",NULL);
 return ret;
}

static unsigned leMapTblNumEntries(BGLOBAL handle) { return (unsigned)lxe.le.lePageCount; }

unsigned long ShowMapTableLE( void )
{
 unsigned long fpos;
 int ret;
 fpos = BMGetCurrFilePos();
 ret = fmtShowList(leMapTblNumEntries,__ReadMapTblLE,
                   " Map of pages ", LB_SELECTIVE, NULL);
 if(ret != -1)
 {
   fpos = CalcPageEntryLE(ret + 1);
 }
 return fpos;
}

unsigned long ShowResNamLE( void )
{
  unsigned long fpos = BMGetCurrFilePos();
  int ret;
  unsigned ordinal;
  ret = fmtShowList(LXRNamesNumItems,LXRNamesReadItems,
                    RES_NAMES,
                    LB_SELECTIVE | LB_SORTABLE,&ordinal);
  if(ret != -1)
  {
    fpos = CalcEntryBungleLE(ordinal,1);
  }
  return fpos;
}

unsigned long ShowNResNmLE( void )
{
  unsigned long fpos;
  fpos = BMGetCurrFilePos();
  {
    int ret;
    unsigned ordinal;
    ret = fmtShowList(LXNRNamesNumItems,LXNRNamesReadItems,
                      NORES_NAMES,
                      LB_SELECTIVE | LB_SORTABLE,&ordinal);
    if(ret != -1)
    {
      fpos = CalcEntryBungleLE(ordinal,1);
    }
  }
  return fpos;
}


int isLE( void )
{
   char id[2];
   headshift = IsNewExe();
   if(headshift)
   {
     BMReadBufferEx(id,sizeof(id),headshift,SEEKF_START);
     if(id[0] == 'L' && id[1] == 'E') return 1;
   }
   return 0;
}

void LEinit( void )
{
   BGLOBAL main_handle;
   LXType = FILE_LE;
   BMReadBufferEx(&lxe.le,sizeof(LEHEADER),headshift,SEEKF_START);
   main_handle = BMbioHandle();
   if((le_cache = bioDup(main_handle)) == &bNull) le_cache = main_handle;
}

void LEdestroy( void )
{
   BGLOBAL main_handle;
   main_handle = BMbioHandle();
   if(le_cache != &bNull && le_cache != main_handle) bioClose(le_cache);
}

unsigned long LEHelp( void )
{
  hlpDisplay(10004);
  return BMGetCurrFilePos();
}

void leAddressResolv(char *addr,unsigned long cfpos)
{
  if(cfpos >= headshift && cfpos < headshift + sizeof(LEHEADER))
  {
     sprintf(addr,"LEH :%s",Get4Digit(cfpos - headshift));
  }
  else
  if(cfpos >= headshift + lxe.le.leObjectTableOffset &&
     cfpos <  headshift + lxe.le.leObjectTableOffset + sizeof(LX_OBJECT)*lxe.le.leObjectCount)
  {
     sprintf(addr,"LEOD:%s",Get4Digit(cfpos - headshift - lxe.le.leObjectTableOffset));
  }
  else
  if(cfpos >= headshift + lxe.le.leObjectPageMapTableOffset &&
     cfpos <  headshift + lxe.le.leObjectPageMapTableOffset + sizeof(LE_PAGE)*lxe.le.lePageCount)
  {
    sprintf(addr,"LEPD:%s",Get4Digit(cfpos - headshift - lxe.le.leObjectPageMapTableOffset));
  }
}

REGISTRY_BIN leTable =
{
  "LE (Linear Executable)",
  { "LEHelp", "Import", "ResNam", "NRsNam", "ImpNam", "Entry ", NULL, "LEHead", "MapTbl", "Object" },
  { LEHelp, ShowModRefLX, ShowResNamLE, ShowNResNmLE, ShowImpProcLXLE, ShowEntriesLX, NULL, ShowNewHeaderLE, ShowMapTableLE, ShowObjectsLX },
  isLE, LEinit, LEdestroy,
  NULL,
  NULL,
  NULL,
  NULL,
  leAddressResolv,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL
};
