/****************************************************************************\
*   LX - this source file is part of Binary vIEW project (BIEW)              *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1995, 2000 Nick Kurshev                            *
*   License:              See below                                          *
*   Author and developer: Nick Kurshev                                       *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   plugins/bin/lx.c                                   *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  This file contain implementation of LX (Linear eXecutable) file format    *
* decoder.                                                                   *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "plugins/bin/lx_le.h"
#include "bin_util.h"
#include "bmfile.h"
#include "biewhelp.h"
#include "kbd_code.h"
#include "tstrings.h"
#include "biewutil.h"
#include "bconsole.h"
#include "reg_form.h"
#include "__os_dep.h"
#include "pmalloc.h"

union LX_LE lxe;
int LXType;

static BGLOBAL lx_cache = &bNull;

static unsigned long __NEAR__ CalcEntryPointLX(unsigned long objnum,unsigned long offset);
extern unsigned long CalcEntryPointLE(unsigned long objnum,unsigned long offset);

static char * LXordering[] =
{
 "little endian",
 "big endian"
};

char * LXcputype[] =
{
 "Unknown",
 "80286",
 "80386",
 "80486",
 "80586",
 "80686",
 "80787",
 "80887"
};

char * LXostype[] =
{
  "Unknown",
  "OS/2",
  "Windows",
  "DOS 4.x",
  "Windows/386",
  "Unknown",
  "Unknown",
  "Unknown"
};

char * __osModType[] =
{
  "PROGRAM",
  "LIBRARY",
  "PROT. MODE PROGRAM",
  "PROT. MODE LIBRARY",
  "PHYSICAL DEVICE DRIVER",
  "VIRTUAL DEVICE DRIVER",
  "PROT. MODE PHYSICAL DEVICE DRIVER",
  "PROT. MODE VIRTUAL DEVICE DRIVER"
};

static unsigned long LXEntryPoint = 0;

extern char * GetPMWinAPI(unsigned flag);
extern char * __nedata[];

static char * __NEAR__ GetOrderingLX(unsigned char type)
{
 if(type < 2) return LXordering[type];
 else         return "";
}

static char * __NEAR__ GetCPUTypeLX(int num)
{
 if(num > 5) num = 0;
 return LXcputype[num];
}

static char * __NEAR__ GetOSTypeLX(int num)
{
  if(num > 4) num = 0;
  return LXostype[num];
}

static char * __NEAR__ __getOSModType(char type)
{
  return __osModType[type & 0x07];
}

static void __NEAR__ PaintNewHeaderLX_1(void)
{
  twPrintF("Signature                        = '%c%c'\n"
           "Byte order                       = %02XH (%s)\n"
           "Word order                       = %02XH (%s)\n"
           "Format level                     = %08lXH\n"
           "OS Type                          = %s\n"
           "CPU Type                         = %s\n"
           "Module version                   = %d.%04d\n"
           "Linear flags :                     [%08lXH]\n"
           "  Contest DATA in EXE: %s\n"
           "  [%c] Per-process library initializtion\n"
           "  [%c] Internal fixups have been applied\n"
           "  [%c] External fixups have been applied\n"
           "  %s\n"
           "  [%c] Module is not loadable (contain errors)\n"
           "  Module type is : %s\n"
           "  [%c] Per-process library termination\n"
           "Number of pages                  = %08lXH\n"
           "EIP objects number               = %08lXH\n"
           "EIP                              = %08lXH\n"
           "ESP objects number               = %08lXH\n"
           "ESP                              = %08lXH"
           ,lxe.lx.lxSignature[0],lxe.lx.lxSignature[1]
           ,(int)lxe.lx.lxByteOrdering,GetOrderingLX(lxe.lx.lxByteOrdering)
           ,(int)lxe.lx.lxWordOrdering,GetOrderingLX(lxe.lx.lxWordOrdering)
           ,lxe.lx.lxFormatLevel
           ,GetOSTypeLX(lxe.lx.lxOSType)
           ,GetCPUTypeLX(lxe.lx.lxCPUType)
           ,(unsigned)(lxe.lx.lxModuleVersion >> 16),(unsigned)(lxe.lx.lxModuleVersion)
           ,lxe.lx.lxModuleFlags
           ,__nedata[lxe.lx.lxModuleFlags & 0x0000003]
           ,GetBool((lxe.lx.lxModuleFlags & 0x00000004L) == 0x00000004L)
           ,GetBool((lxe.lx.lxModuleFlags & 0x00000010L) == 0x00000010L)
           ,GetBool((lxe.lx.lxModuleFlags & 0x00000020L) == 0x00000020L)
           ,GetPMWinAPI((unsigned)(lxe.lx.lxModuleFlags))
           ,GetBool((lxe.lx.lxModuleFlags & 0x00002000L) == 0x00002000L)
           ,__getOSModType(((lxe.lx.lxModuleFlags & 0x00038000L) >> 15) & 0x07)
           ,GetBool((lxe.lx.lxModuleFlags & 0x40000000L) == 0x40000000L)
           ,lxe.lx.lxPageCount
           ,lxe.lx.lxEIPObjectNumbers
           ,lxe.lx.lxEIP
           ,lxe.lx.lxESPObjectNumbers
           ,lxe.lx.lxESP);
}

static void __NEAR__ PaintNewHeaderLX_2( void )
{
  twPrintF("Page size                        = %08lXH\n"
           "Page offset shift                = %08lXH\n"
           "Fixup section size               = %08lXH\n"
           ,lxe.lx.lxPageSize
           ,lxe.lx.lxPageOffsetShift
           ,lxe.lx.lxFixupSectionSize);
  if(LXType == FILE_LX) twPrintF("Fixup section checksum           = %08lXH\n",lxe.lx.lxFixupSectionChecksum);
  else                  twPrintF("Page checksum                    = %08lXH\n",lxe.lx.lxFixupSectionChecksum);
  twPrintF("Loader section size              = %08lXH\n"
           "Loader section checksum          = %08lXH\n"
           "Object table offset              = %08lXH\n"
           "Number of objects in module      = %08lXH\n"
           "Object page table offset         = %08lXH\n"
           "Object iter page  offset         = %08lXH\n"
           "Resource table offset            = %08lXH\n"
           "Number of resource table entries = %08lXH\n"
           "Resident name table offset       = %08lXH\n"
           "Entry table table offset         = %08lXH\n"
           "Module directives offset         = %08lXH\n"
           "Number module directives         = %08lXH\n"
           "Fixup page table offset          = %08lXH\n"
           "Fixup record table offset        = %08lXH\n"
           "Import module table offset       = %08lXH\n"
           "Import module table entries      = %08lXH\n"
           "Import procedure table offset    = %08lXH"
           ,lxe.lx.lxLoaderSectionSize
           ,lxe.lx.lxLoaderSectionChecksum
           ,lxe.lx.lxObjectTableOffset
           ,lxe.lx.lxObjectCount
           ,lxe.lx.lxObjectPageTableOffset
           ,lxe.lx.lxObjectIterPageOffset
           ,lxe.lx.lxResourceTableOffset
           ,lxe.lx.lxNumberResourceTableEntries
           ,lxe.lx.lxResidentNameTableOffset
           ,lxe.lx.lxEntryTableOffset
           ,lxe.lx.lxModuleDirectivesOffset
           ,lxe.lx.lxNumberModuleDirectives
           ,lxe.lx.lxFixupPageTableOffset
           ,lxe.lx.lxFixupRecordTableOffset
           ,lxe.lx.lxImportModuleTableOffset
           ,lxe.lx.lxImportModuleTableEntries
           ,lxe.lx.lxImportProcedureTableOffset);
}

static void __NEAR__ PaintNewHeaderLX_3( void )
{
  twPrintF("Per - page checksum  offset      = %08lXH\n"
           "Data pages offset                = %08lXH\n"
           "Number of preload pages          = %08lXH\n"
           "Non resident name table offset   = %08lXH\n"
           "Non resident name table length   = %08lXH\n"
           "Non resident name table checksum = %08lXH\n"
           ,lxe.lx.lxPerPageChecksumOffset
           ,lxe.lx.lxDataPagesOffset
           ,lxe.lx.lxNumberPreloadPages
           ,lxe.lx.lxNonResidentNameTableOffset
           ,lxe.lx.lxNonResidentNameTableLength
           ,lxe.lx.lxNonResidentNameTableChecksum);
  if(LXType == FILE_LX)
  {
    twPrintF("Auto DS objects number           = %08lXH\n"
             "Debug info offset                = %08lXH\n"
             "Debug info length                = %08lXH\n"
             "Number instance preload          = %08lXH\n"
             "Number instance demand           = %08lXH\n"
             "Heap size                        = %08lXH\n"
             "Stack size                       = %08lXH\n"
             ,lxe.lx.lxAutoDSObjectNumber
             ,lxe.lx.lxDebugInfoOffset
             ,lxe.lx.lxDebugInfoLength
             ,lxe.lx.lxNumberInstancePreload
             ,lxe.lx.lxNumberInstanceDemand
             ,lxe.lx.lxHeapSize
             ,lxe.lx.lxStackSize);
  }
  else
  {
    twPrintF("Debug info offset                = %08lXH\n"
             "Debug info length                = %08lXH\n"
             ,lxe.lx.lxAutoDSObjectNumber
             ,lxe.lx.lxDebugInfoOffset);
  }
  twSetColor(White,Black);
  twPrintF(">Entry Point                     = %08lXH",LXEntryPoint);
  twClrEOL();
  twSetColor(Black,LightGray);
}

static void (__NEAR__ * lxphead[])( void ) =
{
  PaintNewHeaderLX_1,
  PaintNewHeaderLX_2,
  PaintNewHeaderLX_3
};

void PaintNewHeaderLX(TWindow * win,void * *ptr,unsigned npage,unsigned tpage)
{
  char text[80];
  twUseWin(win);
  twFreezeWin(win);
  twClearWin();
  sprintf(text," Linear eXecutable Header [%d/%d] ",npage + 1,tpage);
  twSetTitle(win,text,TW_TMODE_CENTER,Black,LightGray);
  twSetFooter(win,PAGEBOX_SUB,TW_TMODE_RIGHT,White,Black);
  if(npage < 3) (*(lxphead[npage]))();
  twRefreshFullWin(win);
}

unsigned long ShowNewHeaderLX( void )
{
  unsigned long fpos;
  LXEntryPoint = LXType == FILE_LX ? CalcEntryPointLX(lxe.lx.lxEIPObjectNumbers,lxe.lx.lxEIP) : CalcEntryPointLE(lxe.lx.lxEIPObjectNumbers,lxe.lx.lxEIP);
  if(LXEntryPoint == -1) LXEntryPoint = 0;
  fpos = BMGetCurrFilePos();
  if(PageBox(70,21,NULL,3,PaintNewHeaderLX) != -1)
  {
    if(LXEntryPoint) fpos = LXEntryPoint;
  }
  return fpos;
}

extern unsigned GetNamCountNE(BGLOBAL,unsigned long);
extern int RNamesReadItems(BGLOBAL,memArray *,unsigned,unsigned long);

unsigned LXModRefNumItems(BGLOBAL handle)
{
  return (unsigned)lxe.lx.lxImportModuleTableEntries;
}

unsigned LXRNamesNumItems(BGLOBAL handle)
{
  return GetNamCountNE(handle,headshift + lxe.lx.lxResidentNameTableOffset);
}

unsigned LXNRNamesNumItems(BGLOBAL handle)
{
  return GetNamCountNE(handle,lxe.lx.lxNonResidentNameTableOffset);
}

int LXRNamesReadItems(BGLOBAL handle,memArray * obj,unsigned nnames)
{
   return RNamesReadItems(handle,obj,nnames,lxe.lx.lxResidentNameTableOffset + headshift);
}

unsigned LXImpNamesNumItems(BGLOBAL handle)
{
  unsigned long fpos;
  unsigned char len;
  unsigned count;
  bioSeek(handle,lxe.lx.lxImportProcedureTableOffset + headshift,SEEKF_START);
  fpos = bioTell(handle);
  count = 0;
  while(fpos < lxe.lx.lxFixupSectionSize + lxe.lx.lxFixupPageTableOffset + headshift)
  {
    len = bioReadByte(handle);
    bioSeek(handle,len,SEEKF_CUR);
    fpos = bioTell(handle);
    count++;
  }
  return count;
}

int LXImpNamesReadItems(BGLOBAL handle,memArray * obj,unsigned nnames)
{
 unsigned i;
 unsigned char byte;
 bioSeek(handle,lxe.lx.lxImportProcedureTableOffset + headshift,SEEKF_START);
 for(i = 0;i < nnames;i++)
 {
   char nam[256];
   byte = bioReadByte(handle);
   if(IsKbdTerminate()) break;
   bioReadBuffer(handle,nam,byte);
   nam[byte] = 0;
   if(!ma_AddString(obj,nam,1)) break;
 }
 return 1;
}


int LXNRNamesReadItems(BGLOBAL handle,memArray * obj,unsigned nnames)
{
   return RNamesReadItems(handle,obj,nnames,lxe.lx.lxNonResidentNameTableOffset);
}

extern unsigned RNameReadFull(BGLOBAL handle,char * names,unsigned nindex,unsigned long offset);

unsigned LXRNamesReadFullName(BGLOBAL handle,char * names,unsigned index)
{
   return RNameReadFull(handle,names,index,lxe.lx.lxResidentNameTableOffset + headshift);
}

unsigned LXNRNamesReadFullName(BGLOBAL handle,char * names,unsigned index)
{
   return RNameReadFull(handle,names,index,lxe.lx.lxNonResidentNameTableOffset);
}

int  __ReadModRefNamesLX(BGLOBAL handle,memArray * obj,unsigned nnames)
{
 unsigned i;
 unsigned char byte;
 bioSeek(handle,lxe.lx.lxImportModuleTableOffset + headshift,SEEKF_START);
 for(i = 0;i < nnames;i++)
 {
   char nam[256];
   byte = bioReadByte(handle);
   if(IsKbdTerminate()) break;
   bioReadBuffer(handle,nam,byte);
   nam[byte] = 0;
   if(!ma_AddString(obj,nam,1)) break;
 }
 return 1;
}

static void __NEAR__ objpaintLX(LX_OBJECT *nam)
{
 twGotoXY(1,1);
 twPrintF("Virtual Size                         = %lX bytes\n"
          "BVA (base virtual address)           = %08lX\n"
          "FLAGS: %lX\n"
          "   [%c] Readable object\n"
          "   [%c] Writable object\n"
          "   [%c] Executable object\n"
          "   [%c] Resource object\n"
          "   [%c] Object is discardable\n"
          "   [%c] Object is shared\n"
          "   [%c] Object has preload pages\n"
          "   [%c] Object has invalid pages\n"
          "   [%c] Object is permanent and swappable\n"
          "   [%c] Object is permanent and resident\n"
          "   [%c] Object is permanent and long lockable\n"
          "   [%c] 16:16 alias required (80x86 specific)\n"
          "   [%c] Big/Default bit setting (80x86 specific)\n"
          "   [%c] Object is conforming for code (80x86 specific)\n"
          "   [%c] Object I/O privilege level (80x86 specific)\n"
          "Object page map index                = %lu\n"
          "Number of entries in object page map = %lu"
          ,nam->o32_size
          ,nam->o32_base
          ,nam->o32_flags
          ,GetBool((nam->o32_flags & 0x00000001L) == 0x00000001L)
          ,GetBool((nam->o32_flags & 0x00000002L) == 0x00000002L)
          ,GetBool((nam->o32_flags & 0x00000004L) == 0x00000004L)
          ,GetBool((nam->o32_flags & 0x00000008L) == 0x00000008L)
          ,GetBool((nam->o32_flags & 0x00000010L) == 0x00000010L)
          ,GetBool((nam->o32_flags & 0x00000020L) == 0x00000020L)
          ,GetBool((nam->o32_flags & 0x00000040L) == 0x00000040L)
          ,GetBool((nam->o32_flags & 0x00000080L) == 0x00000080L)
          ,GetBool((nam->o32_flags & 0x00000100L) == 0x00000100L)
          ,GetBool((nam->o32_flags & 0x00000200L) == 0x00000200L)
          ,GetBool((nam->o32_flags & 0x00000400L) == 0x00000400L)
          ,GetBool((nam->o32_flags & 0x00001000L) == 0x00001000L)
          ,GetBool((nam->o32_flags & 0x00002000L) == 0x00002000L)
          ,GetBool((nam->o32_flags & 0x00004000L) == 0x00004000L)
          ,GetBool((nam->o32_flags & 0x00008000L) == 0x00008000L)
          ,nam->o32_pagemap
          ,nam->o32_mapsize);
}

void ObjPaintLX(TWindow * win,void * * names,unsigned start,unsigned nlist)
{
 char buffer[81];
 LX_OBJECT * * nam = (LX_OBJECT * *)names;
 twUseWin(win);
 twFreezeWin(win);
 twClearWin();
 sprintf(buffer," Object Table [ %u / %u ] ",start + 1,nlist);
 twSetTitle(win,buffer,TW_TMODE_CENTER,Black,LightGray);
 twSetFooter(win,PAGEBOX_SUB,TW_TMODE_RIGHT,White,Black);
 objpaintLX(nam[start]);
 twRefreshFullWin(win);
}

static int __NEAR__ __ReadObjectsLX(BGLOBAL handle,memArray * obj,unsigned n)
{
 int i;
  for(i = 0;i < n;i++)
  {
    LX_OBJECT lxo;
    if(IsKbdTerminate()) break;
    bioReadBuffer(handle,&lxo,sizeof(LX_OBJECT));
    if(!ma_AddData(obj,&lxo,sizeof(LX_OBJECT),1)) break;
  }
  return 1;
}

static int __NEAR__ __ReadEntriesLX(BGLOBAL handle,memArray *obj)
{
 unsigned i;
 unsigned char j;
 unsigned char cnt,type;
 unsigned short numobj = 0;
 LX_ENTRY lxe;
 i = 0;
 while(1)
 {
   cnt = bioReadByte(handle);
   type = bioReadByte(handle);
   if(!cnt) break;
   if(type) numobj = bioReadWord(handle);
   for(j = 0;j < cnt;j++,i++)
   {
     char size;
     switch(type)
     {
       case 0: size = 0; break;
       case 1: size = 2; break;
       case 2:
       case 0x80:
       case 3: size = 4; break;
       default:
       case 4: size = 6; break;
     }
     if(IsKbdTerminate()) goto exit;
     lxe.b32_type = type;
     if(size)
     {
       lxe.b32_obj = numobj;
       lxe.entry.e32_flags = bioReadByte(handle);
       bioReadBuffer(handle,&lxe.entry.e32_variant,size);
     }
     if(!ma_AddData(obj,&lxe,sizeof(LX_ENTRY),1)) goto exit;
   }
 }
 exit:
 return 1;
}

static unsigned long __NEAR__ __calcPageEntry(LX_MAP_TABLE *mt)
{
  unsigned long dataoff;
  unsigned long ret;
  dataoff = mt->o32_pagedataoffset << lxe.lx.lxPageOffsetShift;
  if(mt->o32_pageflags == 1 || mt->o32_pageflags == 5) ret = lxe.lx.lxObjectIterPageOffset + dataoff;
  else
    if(mt->o32_pageflags == 0) ret = lxe.lx.lxDataPagesOffset + dataoff;
    else                      ret = -1;
  return ret;
}

static unsigned long __NEAR__ CalcPageEntry(unsigned long pageidx)
{
  BGLOBAL handle;
  LX_MAP_TABLE mt;
  if(!pageidx) return -1;
  handle = lx_cache;
  bioSeek(handle,lxe.lx.lxObjectPageTableOffset + headshift,SEEK_SET);
  bioSeek(handle,sizeof(LX_MAP_TABLE)*(pageidx - 1),SEEKF_CUR);
  bioReadBuffer(handle,(void *)&mt,sizeof(LX_MAP_TABLE));
  return __calcPageEntry((void *)&mt);
}

static unsigned long __NEAR__ CalcEntryPointLX(unsigned long objnum,unsigned long offset)
{
  BGLOBAL handle;
  unsigned long ret,pshift;
  LX_OBJECT lo;
  LX_MAP_TABLE mt;
  if(!objnum) return BMGetCurrFilePos();
  handle = lx_cache;
  bioSeek(handle,lxe.lx.lxObjectTableOffset + headshift,SEEK_SET);
  bioSeek(handle,sizeof(LX_OBJECT)*(objnum - 1),SEEKF_CUR);
  bioReadBuffer(handle,(void *)&lo,sizeof(LX_OBJECT));
/*  if((lo.o32_flags & 0x00002000L) == 0x00002000L) USE16 = 0;
  else                                            USE16 = 0xFF; */
  bioSeek(handle,lxe.lx.lxObjectPageTableOffset + headshift,SEEKF_START);
  pshift = offset % lxe.lx.lxPageSize;
  bioSeek(handle,sizeof(LX_MAP_TABLE)*(lo.o32_pagemap - 1 + offset/lxe.lx.lxPageSize),SEEKF_CUR);
  bioReadBuffer(handle,(void *)&mt,sizeof(LX_MAP_TABLE));
  ret = __calcPageEntry((void *)&mt) + pshift;
  return ret;
}

void ReadLXLEImpMod(unsigned long offtable,unsigned num,char *str)
{
  BGLOBAL handle;
  unsigned i;
  unsigned char len;
  char buff[256];
  handle = lx_cache;
  bioSeek(handle,offtable,SEEK_SET);
  for(i = 1;i < num;i++)
  {
    len = bioReadByte(handle);
    bioSeek(handle,len,SEEKF_CUR);
  }
  len = bioReadByte(handle);
  bioReadBuffer(handle,(void *)buff,len);
  buff[len] = 0;
  strcat(str,buff);
}

void ReadLXLEImpName(unsigned long offtable,unsigned num,char *str)
{
  BGLOBAL handle;
  unsigned char len;
  char buff[256];
  handle = lx_cache;
  bioSeek(handle,offtable+num,SEEK_SET);
  len = bioReadByte(handle);
  bioReadBuffer(handle,(void *)buff,len);
  buff[len] = 0;
  strcat(str,buff);
}

void ShowFwdModOrdLX(LX_ENTRY *lxent)
{
  char buff[513];
  buff[0] = 0;
  ReadLXLEImpMod(lxe.lx.lxImportModuleTableOffset + headshift,lxent->entry.e32_variant.e32_fwd.modord,buff);
  strcat(buff,".");
  if((lxent->entry.e32_flags & 0x01) == 0x01)
  {
    sprintf(&buff[strlen(buff)],"@%u",(unsigned)lxent->entry.e32_variant.e32_fwd.value);
  }
  else ReadLXLEImpName(lxe.lx.lxImportProcedureTableOffset + headshift,(unsigned)lxent->entry.e32_variant.e32_fwd.value,buff);
  TMessageBox(buff," Forwarder entry point ");
}

static unsigned long __NEAR__ CalcEntryLX(LX_ENTRY *lxent)
{
  unsigned long ret;
  ret = BMGetCurrFilePos();
      switch(lxent->b32_type)
      {
        case 1: ret = CalcEntryPointLX(lxent->b32_obj,lxent->entry.e32_variant.e32_offset.offset16);
                      break;
        case 2: ret = CalcEntryPointLX(lxent->b32_obj,lxent->entry.e32_variant.e32_callgate.offset);
                      break;
        case 3: ret = CalcEntryPointLX(lxent->b32_obj,lxent->entry.e32_variant.e32_offset.offset32);
                      break;
        case 4: ShowFwdModOrdLX(lxent);
        case 5:
        default: break;
      }
  return ret;
}

static unsigned long __NEAR__ CalcEntryBungleLX(unsigned ordinal,int dispmsg)
{
  BGLOBAL handle;
  int found;
  unsigned i;
  unsigned char j;
  unsigned char cnt,type;
  unsigned short numobj = 0;
  LX_ENTRY lxent;
  unsigned long ret;
  ret = BMGetCurrFilePos();
  handle = lx_cache;
  bioSeek(handle,lxe.lx.lxEntryTableOffset + headshift,SEEK_SET);
  {
    if(dispmsg) ErrMessageBox("Internal error !!!",NULL);
    return ret;
  }
  i = 0;
  found = 0;
  while(1)
  {
   cnt = bioReadByte(handle);
   type = bioReadByte(handle);
   if(!cnt) break;
   if(type) numobj = bioReadWord(handle);
   for(j = 0;j < cnt;j++,i++)
   {
     char size;
     switch(type)
     {
       case 0: size = 0; break;
       case 1: size = 2; break;
       case 2:
       case 0x80:
       case 3: size = 4; break;
       default:
       case 4: size = 6; break;
     }
     if(i == ordinal - 1)
     {
       lxent.b32_type = type;
       found = 1;
       if(size)
       {
         lxent.b32_obj = numobj;
         lxent.entry.e32_flags = bioReadByte(handle);
         bioReadBuffer(handle,(void *)&lxent.entry.e32_variant,size);
       }
       break;
     }
     else if(size) bioSeek(handle,size + sizeof(char),SEEKF_CUR);
   }
   if(found == 1) break;
 }
 if(found) ret = CalcEntryLX((LX_ENTRY *)&lxent);
 else      if(dispmsg) ErrMessageBox("No such entry in the Table",NULL);
 return ret;
}

extern unsigned long CalcPageEntryLE(unsigned long idx);

unsigned long ShowObjectsLX( void )
{
 BGLOBAL handle;
 unsigned long fpos;
 unsigned nnames;
 memArray * obj;
 fpos = BMGetCurrFilePos();
 nnames = (unsigned)lxe.lx.lxObjectCount;
 if(!nnames) { NotifyBox("Entry not found"," Objects Table "); return fpos; }
 if(!(obj = ma_Build(nnames,1))) return fpos;
 handle = lx_cache;
 bioSeek(handle,lxe.lx.lxObjectTableOffset + headshift,SEEK_SET);
 if(__ReadObjectsLX(handle,obj,nnames))
 {
  int ret;
    ret = PageBox(70,20,obj->data,obj->nItems,ObjPaintLX);
    if(ret != -1)  fpos = LXType == FILE_LX ? CalcPageEntry(((LX_OBJECT *)obj->data[ret])->o32_pagemap) : CalcPageEntryLE(((LX_OBJECT *)obj->data[ret])->o32_pagemap);
 }
 ma_Destroy(obj);
 return fpos;
}

char * mapattr[] =
{
"Valid Physical Page In .EXE",
"Iterated Data Page",
"Invalid Page",
"Zero Filled Page",
"Range Of Pages",
"Iterated Data Page Type II"
};

char * lxeGetMapAttr(unsigned long attr)
{
  if (attr > 5) return "Unknown";
  else  return mapattr[attr];
}

static void __NEAR__ iterpaintLX(LX_ITER *nam)
{
 twGotoXY(1,1);
 twPrintF("Number of iteration                  = %hu\n"
          "Number of bytes                      = %hu\n"
          "Iterated data bytes                  = %hu"
          ,nam->LX_nIter
          ,nam->LX_nBytes
          ,(int)nam->LX_Iterdata);
}

char *__e32type[] =
{
  "Empty",
  "Entry16",
  "Gate16",
  "Entry32",
  "EntryForwarder",
  "TypeInfo"
};

static char * __NEAR__ entryTypeLX(unsigned char type)
{
   if(type < 6) return __e32type[type];
   else         return "Unknown";
}

static void __NEAR__ entrypaintLX(LX_ENTRY *nam)
{
 if(!nam->b32_type)
 {
   twGotoXY(35,4);
   twPrintF("Unused");
 }
 else
 {
   twGotoXY(1,1);
   twPrintF("Entry type: %s\n"
            "Object number : %hd\n"
            "Flags: %02XH\n"
            ,entryTypeLX(nam->b32_type)
            ,nam->b32_obj
            ,(int)nam->entry.e32_flags);
   if(nam->b32_type != 4)
   {
     twPrintF("   [%c] Exported Entry\n"
              "   [%c] Used Shared Data\n"
              "   %02XH - parameter word count mask\n"
              ,GetBool((nam->entry.e32_flags & 0x01) == 0x01)
              ,GetBool((nam->entry.e32_flags & 0x02) == 0x02)
              ,(int)(nam->entry.e32_flags >> 2));
   }
   else
   {
     twPrintF("\n"
              "   [%c] Import by ordinal\n"
              "\n"
              ,GetBool((nam->entry.e32_flags & 0x01) == 0x01));
   }
   if(nam->b32_type == 1)
   {
     twPrintF("Entry offset : %04hXH\n"
              "\n"
              ,nam->entry.e32_variant.e32_offset.offset16);
   }
   else
      if(nam->b32_type == 3)
      {
        twPrintF("Entry offset : %08XH\n"
                 "\n"
                 ,nam->entry.e32_variant.e32_offset.offset32);
      }
      else
      if(nam->b32_type == 2)
      {
       twPrintF("Callgate offset : %04hXH\n"
                "Callgate selector : %04hXH\n"
                ,nam->entry.e32_variant.e32_callgate.offset
                ,nam->entry.e32_variant.e32_callgate.callgate);
      }
      else
       if(nam->b32_type == 4)
       {
         twPrintF("Module ordinal number : %04hXH\n"
                  "Proc name offset or ordinal : %04hXH\n"
                  ,nam->entry.e32_variant.e32_fwd.modord
                  ,nam->entry.e32_variant.e32_fwd.value);
       }
   }
}

void IterPaintLX(TWindow * win,void * * names,unsigned start,unsigned nlist)
{
 char buffer[81];
 LX_ITER * * nam = (LX_ITER * *)names;
 twUseWin(win);
 twFreezeWin(win);
 twClearWin();
 sprintf(buffer," Iter Table [ %u / %u ] ",start + 1,nlist);
 twSetTitle(win,buffer,TW_TMODE_CENTER,Black,LightGray);
 iterpaintLX(nam[start]);
 twRefreshFullWin(win);
}

void PaintEntriesLX(TWindow * win,void * * names,unsigned start,unsigned nlist)
{
 char buffer[81];
 LX_ENTRY * * nam = (LX_ENTRY * *)names;
 twUseWin(win);
 twFreezeWin(win);
 twClearWin();
 sprintf(buffer," Entries Table [ %u / %u ] ",start + 1,nlist);
 twSetTitle(win,buffer,TW_TMODE_CENTER,Black,LightGray);
 twSetFooter(win,PAGEBOX_SUB,TW_TMODE_RIGHT,White,Black);
 entrypaintLX(nam[start]);
 twRefreshFullWin(win);
}

static int __ReadMapTblLX(BGLOBAL handle,memArray * obj,unsigned n)
{
 int i;
  bioSeek(handle,lxe.lx.lxObjectPageTableOffset + headshift,SEEKF_START);
  for(i = 0;i < n;i++)
  {
    LX_MAP_TABLE mt;
    char stmp[80];
    if(IsKbdTerminate()) break;
    bioReadBuffer(handle,&mt,sizeof(LX_MAP_TABLE));
    sprintf(stmp,"Off=%08lXH Siz=%04hXH Flg:%04hXH=%s",
                 mt.o32_pagedataoffset,
                 mt.o32_pagesize,
                 mt.o32_pageflags,
                 (char *)lxeGetMapAttr(mt.o32_pageflags));
    if(!ma_AddString(obj,stmp,1)) break;
  }
  return 1;
}
/*
static int __NEAR__ __ReadIterTblLX(BGLOBAL handle,memArray * obj,unsigned n)
{
 int i;
  for(i = 0;i < n;i++)
  {
    LX_ITER lxi;
    if(IsKbdTerminate()) break;
    bioReadBuffer(handle,&lxi,sizeof(LX_ITER));
    if(!ma_AddData(obj,&lxi,sizeof(LX_ITER),1)) break;
  }
  return 1;
}
*/
static unsigned lxGetPageCount(BGLOBAL handle) { return (unsigned)lxe.lx.lxPageCount; }

unsigned long ShowMapTableLX( void )
{
 unsigned long fpos;
 int ret;
 fpos = BMGetCurrFilePos();
 ret = fmtShowList(lxGetPageCount,__ReadMapTblLX," Map of pages ",
                   LB_SELECTIVE,NULL);
 if(ret != -1)
 {
   fpos = CalcPageEntry(ret + 1);
 }
 return fpos;
}

extern unsigned long CalcEntryLE(LX_ENTRY *);

unsigned long ShowEntriesLX( void )
{
 BGLOBAL handle;
 unsigned long fpos;
 memArray * obj;
 fpos = BMGetCurrFilePos();
 if(!lxe.lx.lxEntryTableOffset) { NotifyBox("Entry not found"," Entry Table "); return fpos; }
 handle = lx_cache;
 bioSeek(handle,lxe.lx.lxEntryTableOffset + headshift,SEEK_SET);
 if(!(obj = ma_Build(0,1))) goto exit;
 if(__ReadEntriesLX(handle,obj))
 {
    int ret;
    if(!obj->nItems) { NotifyBox("Entry not found"," Entry Table "); goto bye; }
    ret = PageBox(70,8,obj->data,obj->nItems,PaintEntriesLX);
    if(ret != -1)  fpos = LXType == FILE_LX ? CalcEntryLX(obj->data[ret]) : CalcEntryLE(obj->data[ret]);
 }
 bye:
 ma_Destroy(obj);
 exit:
 return fpos;
}

char * ResourceGrNamesLX[] =
{
  "RESERVED 0",
  "Mouse pointer share",
  "BITMAP",
  "MENU",
  "DIALOG",
  "STRINGTABLE",
  "FONTDIR",
  "FONT",
  "ACCELTABLE",
  "RCDATA",
  "Err msg table",
  "DlgInclude",
  "VKey table",
  "Key to UGL table",
  "Glyph to char table",
  "Display info",
  "FKA short",
  "FKA long",
  "Help table",
  "Help subtable",
  "DBCS font drive dir",
  "DBCS font drive"
};


typedef struct tagLXResource
{
   unsigned short typeID;
   unsigned short nameID;
   unsigned long  resourceSize;
   unsigned short object;
   unsigned long  offset;
}LXResource;

static int __NEAR__ __ReadResourceGroupLX(BGLOBAL handle,memArray *obj,unsigned nitems,long * addr)
{
 unsigned i;
 LXResource lxr;
 for(i = 0;i < nitems;i++)
 {
    char stmp[81];
    bioReadBuffer(handle,&lxr,sizeof(LXResource));
    addr[i] = lxr.offset;
    if(IsKbdTerminate()) break;
    sprintf(stmp,"%6hu = ",lxr.nameID);
    if(lxr.typeID < sizeof(ResourceGrNamesLX)/sizeof(char *)) strcat(stmp,ResourceGrNamesLX[lxr.typeID]);
    else  sprintf(&stmp[strlen(stmp)],"Unknown < %04hXH >",lxr.typeID);
    sprintf(&stmp[strlen(stmp)]," obj=%04hXH.%08lXH",lxr.object,lxr.offset);
    if(!ma_AddString(obj,stmp,1)) break;
 }
 return 1;
}


unsigned long ShowResourceLX( void )
{
 unsigned long fpos;
 BGLOBAL handle;
 memArray * obj;
 long * raddr;
 unsigned nrgroup;
 fpos = BMGetCurrFilePos();
 handle = lx_cache;
 bioSeek(handle,(long)headshift + lxe.lx.lxResourceTableOffset,SEEK_SET);
 nrgroup = (unsigned)lxe.lx.lxNumberResourceTableEntries;
 if(!nrgroup) { NotifyBox(NOT_ENTRY," Resources "); return fpos; }
 if(!(obj = ma_Build(nrgroup,1))) goto exit;
 if(!(raddr  = PMalloc(nrgroup*sizeof(long)))) return fpos;
 if(__ReadResourceGroupLX(handle,obj,nrgroup,raddr))
 {
   ma_Display(obj," Resource groups : ",LB_SORTABLE,-1);
 }
 ma_Destroy(obj);
 free(raddr);
 exit:
 return fpos;
}

unsigned long ShowModRefLX( void )
{
  fmtShowList(LXModRefNumItems,
              __ReadModRefNamesLX,
              MOD_REFER,
              LB_SORTABLE,
              NULL);
  return BMGetCurrFilePos();
}

unsigned long ShowResNamLX( void )
{
  unsigned long fpos = BMGetCurrFilePos();
  int ret;
  unsigned ordinal;
  ret = fmtShowList(LXRNamesNumItems,LXRNamesReadItems,
                    RES_NAMES,
                    LB_SELECTIVE | LB_SORTABLE,&ordinal);
  if(ret != -1)
  {
    fpos = CalcEntryBungleLX(ordinal,1);
  }
  return fpos;
}

unsigned long ShowNResNmLX( void )
{
  unsigned long fpos;
  fpos = BMGetCurrFilePos();
  {
    int ret;
    unsigned ordinal;
    ret = fmtShowList(LXNRNamesNumItems,LXNRNamesReadItems,
                      NORES_NAMES,
                      LB_SELECTIVE | LB_SORTABLE,&ordinal);
    if(ret != -1)
    {
      fpos = CalcEntryBungleLX(ordinal,1);
    }
  }
  return fpos;
}

unsigned long ShowImpProcLXLE( void )
{
  fmtShowList(LXImpNamesNumItems,LXImpNamesReadItems,
              IMPPROC_TABLE,0,NULL);
  return BMGetCurrFilePos();
}


int isLX( void )
{
   char id[2];
   headshift = IsNewExe();
   if(headshift)
   {
     BMReadBufferEx(id,sizeof(id),headshift,SEEKF_START);
     if(id[0] == 'L' && id[1] == 'X') return 1;
   }
   return 0;
}

void LXinit( void )
{
   BGLOBAL main_handle;
   LXType = FILE_LX;
   BMReadBufferEx(&lxe.lx,sizeof(LXHEADER),headshift,SEEKF_START);
   main_handle = BMbioHandle();
   if((lx_cache = bioDup(main_handle)) == &bNull) lx_cache = main_handle;
}

void LXdestroy( void )
{
   BGLOBAL main_handle;
   main_handle = BMbioHandle();
   if(lx_cache != &bNull && lx_cache != main_handle) bioClose(lx_cache);
}

unsigned long LXHelp( void )
{
  hlpDisplay(10005);
  return BMGetCurrFilePos();
}

void lxAddressResolv(char *addr,unsigned long cfpos)
{
  if(cfpos >= headshift && cfpos < headshift + sizeof(LXHEADER))
  {
    sprintf(addr,"LXH :%s",Get4Digit(cfpos - headshift));
  }
  else
  if(cfpos >= headshift + lxe.lx.lxObjectTableOffset &&
     cfpos <  headshift + lxe.lx.lxObjectTableOffset + sizeof(LX_OBJECT)*lxe.lx.lxObjectCount)
  {
    sprintf(addr,"LXOD:%s",Get4Digit(cfpos - headshift - lxe.lx.lxObjectTableOffset));
  }
  else
  if(cfpos >= headshift + lxe.lx.lxObjectPageTableOffset &&
     cfpos <  headshift + lxe.lx.lxObjectPageTableOffset + sizeof(LX_MAP_TABLE)*lxe.lx.lxPageCount)
  {
    sprintf(addr,"LXPD:%s",Get4Digit(cfpos - headshift - lxe.lx.lxObjectPageTableOffset));
  }
}

REGISTRY_BIN lxTable =
{
  "LX (Linear eXecutable)",
  { "LXhelp", "Import", "ResNam", "NRsNam", "ImpNam", "Entry ", "ResTbl", "LXHead", "MapTbl", "Object" },
  { LXHelp, ShowModRefLX, ShowResNamLX, ShowNResNmLX, ShowImpProcLXLE, ShowEntriesLX, ShowResourceLX, ShowNewHeaderLX, ShowMapTableLX, ShowObjectsLX },
  isLX, LXinit, LXdestroy,
  NULL,
  NULL,
  NULL,
  NULL,
  lxAddressResolv,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL
};


