/****************************************************************************\
*   PHARLAP - this header file is part of Binary vIEW project (BIEW)         *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1995, 2000 Nick Kurshev                            *
*   License:              See below                                          *
*   Author and developer: Nick Kurshev                                       *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   plugins/bin/pharlap.h                              *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  This file contain PharLap executable file definitions                     *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/
#ifndef __PHARLAP_INC
#define __PHARLAP_INC

#include "biewtype.h"

#ifdef __GNUC__
#pragma pack(1)
#endif

/* Format of .old Phar Lap .EXP file header */

typedef struct tagoldPharLap
{
  tByte            plSignature[2];   /* "MP" 0x4D 0x50 */
  tWord            plSizeRemaind;    /* remainder of image size / page size (page size = 512h) */
  tWord            plImageSize;      /* size of image in pages */
  tWord            plNRelocs;        /* number of relocation items */
  tWord            plHeadSize;       /* header size in paragraphs */
  tWord            plMinExtraPages;  /* minimum number of extra 4K pages to be allocated at the end of program, when it is loaded */
  tWord            plMaxExtraPages;  /* maximum number of extra 4K pages to be allocated at the end of program, when it is loaded */
  tDWord           plESP;            /* initial ESP */
  tWord            plCheckSum;       /* check sum of file */
  tDWord           plEIP;            /* initial EIP */
  tWord            plFirstReloc;     /* offset of first relocation item */
  tWord            plNOverlay;       /* overlay number */
  tWord            plReserved;       /* (???) wants to be 1 */
}oldPharLap;

/* Format of new Phar Lap .EXP file header */

typedef struct tagnewPharLap
{
  tByte            plSignature[2];   /* "P2" for 286 .EXP and "P3" for 386 .EXP */
  tWord            plLevel;          /* 0x01 - flat model 0x02 - multisegmented file */
  tWord            plHeaderSize;
  tDWord           plFileSize;       /* Size of file in bytes */
  tWord            plCheckSum;
  tDWord           plRunTimeParms;   /* offset of run-time parameters within file */
  tDWord           plRunTimeSize;    /* size of run-time parameters in bytes */
  tDWord           plRelocOffset;    /* offset of relocation table within file */
  tDWord           plRelocSize;      /* size of relocation table in bytes */
  tDWord           plSegInfoOffset;  /* offset of segment information table within file */
  tDWord           plSegInfoSize;    /* size of segment information table in bytes */
  tWord            plSegEntrySize;   /* size of segment information table entry in bytes */
  tDWord           plImageOffset;    /* offset of load image within file */
  tDWord           plImageSize;      /* size of load image on disk */
  tDWord           plSymTabOffset;   /* offset of symbol table within file */
  tDWord           plSymTabSize;     /* size of symbol table in bytes */
  tDWord           plGDTOffset;      /* offset of GDT within load image */
  tDWord           plGDTSize;        /* size of GDT in bytes */
  tDWord           plLDTOffset;      /* offset of LDT within load image */
  tDWord           plLDTSize;        /* size of LDT in bytes */
  tDWord           plIDTOffset;      /* offset of IDT within load image */
  tDWord           plIDTSize;        /* size of IDT in bytes */
  tDWord           plTSSOffset;      /* offset of TSS within load image */
  tDWord           plTSSSize;        /* size of TSS in bytes */
  tDWord           plMinExtraPages;  /* minimum number of extra 4K pages to be allocated at the end of program, level 1 only */
  tDWord           plMaxExtraPages;  /* maximum number of extra 4K pages to be allocated at the end of program, level 1 only */
  tDWord           plBase;           /* base load offset (level 1 executables only) */
  tDWord           plESP;            /* initial ESP */
  tWord            plSS;             /* initial SS */
  tDWord           plEIP;            /* initial EIP */
  tWord            plCS;             /* initial CS */
  tWord            plLDT;            /* initial LDT */
  tWord            plTSS;            /* initial TSS */
  tWord            plFlags;          /* bit 0: load image is packed */
                                     /* bit 1: 32-bit checksum is present */
                                     /* bits 4-2: type of relocation table */
  tDWord           plMemReq;         /* memory requirements for load image */
  tDWord           plChecksum32;     /* 32-bit checksum (optional) */
  tDWord           plStackSize;      /* size of stack segment in bytes */
  tByte            plReserv[256];
}newPharLap;

typedef struct tagPLSegInfo
{
  tWord            siSelector;       /* selector number */
  tWord            siFlags;
  tDWord           siBaseOff;        /* base offset of selector */
  tDWord           siMinAlloc;       /* minimum number of extra bytes to be allocated to the segment */
}PLSegInfo;

typedef struct tagPLRunTimeParms
{
  tByte            rtSignature[2];   /* "DX" 44h 58h */
  tWord            rtMinRModeParms;  /* minimum number of real-mode params to leave free at run time */
  tWord            rtMaxRModeParms;  /* maximum number of real-mode params to leave free at run time */
  tWord            rtMinIBuffSize;   /* minimum interrupt buffer size in KB */
  tWord            rtMaxIBuffSize;   /* maximum interrupt buffer size in KB */
  tWord            rtNIStacks;       /* number of interrupt stacks */
  tWord            rtIStackSize;     /* size in KB of each interrupt stack */
  tDWord           rtEndRModeOffset; /* offset of byte past end of real-mode code and data */
  tWord            rtCallBuffSize;   /* size in KB of call buffers */
  tWord            rtFlags;
                                     /* bit 0: file is virtual memory manager */
                                     /* bit 1: file is a debugger */
  tWord            rtUnprivFlags;    /* unprivileged flag (if nonzero, executes at ring 1, 2, or 3) */
  tByte            rtReserv[104];
}PLRunTimeParms;

typedef struct tagPLRepeatBlock
{
  tWord            rbCount;          /* byte count */
  tByte            rbString[1];      /* repeat string length */
}PLRepeatBlock;

#ifdef __GNUC__
#pragma pack()
#endif

#endif
