/****************************************************************************\
*   IX86 - this header file is part of Binary vIEW project (BIEW)            *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1995, 2000 Nick Kurshev                            *
*   License:              See below                                          *
*   Author and developer: Nick Kurshev                                       *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   plugins/disasm/ix86/ix86.h                         *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  This file contain declaration of internal Intel x86 disassembler functions*
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/

#ifndef ____DISASM_H
#define ____DISASM_H

#include "disasm.h"

#define TAB_POS 10
#define TILE_SAFE 4000

#define DUMMY_PTR 0 /* "" */
#define BYTE_PTR  1 /*" <b>"*/
#define WORD_PTR  2 /*" <w>"*/
#define DWORD_PTR 3 /*" <d>"*/
#define PWORD_PTR 4 /*" <p>"*/
#define QWORD_PTR 5 /*" <q>"*/
#define TWORD_PTR 6 /*" <t>"*/

#define DIS_LENONLY 0x8000

typedef struct tagix86Param
{
  unsigned long pro_clone;
  unsigned char codelen;
  unsigned long DisasmPrefAddr; /* address of instruction with prefixes */
  unsigned long CodeAddress; /* without prefixes */
  MBuffer       CodeBuffer; /* buffer with source code */
  MBuffer       RealCmd; /* buffer without prefixes */
  unsigned      flags;
}ix86Param;

extern char Use32Addr,Use32Data,UseMMXSet,UseXMMXSet;
extern char * SJump[];
typedef void (*ix86_method)(char *encode_str,ix86Param *);

#define IX86_CPU086  0x00000000L
#define IX86_CPU186  0x00000001L
#define IX86_CPU286  0x00000002L
#define IX86_CPU386  0x00000003L
#define IX86_CPU486  0x00000004L
#define IX86_CPU586  0x00000005L
#define IX86_CPU686  0x00000006L
#define IX86_CPU786  0x00000007L

#define IX86_CPUMASK 0x000000FFL

#define IX86_FPU087 0x000000FFL
#define IX86_FPU187 0x000001FFL
#define IX86_FPU287 0x000002FFL
#define IX86_FPU387 0x000003FFL
#define IX86_FPU487 0x000004FFL
#define IX86_FPU587 0x000005FFL
#define IX86_FPU687 0x000006FFL
#define IX86_FPU787 0x000007FFL

#define IX86_FPUMASK 0x0000FF00L

#define IX86_MMX586 0x0005FFFFL
#define IX86_MMX686 0x0006FFFFL
#define IX86_MMX786 0x0007FFFFL

#define IX86_MMXMASK 0x00FF0000L

#define IX86_INTEL     0x00000000L
#define IX86_AMD       0x01000000L
#define IX86_CYRIX     0x02000000L
#define IX86_CLONEMASK 0xFF000000L

#define IX86_CKLAMATH  IX86_CPU686
#define IX86_CKATMAI   IX86_CPU786
#define IX86_FDECHUTES IX86_FPU787
#define IX86_MKATMAI   IX86_MMX786
#define IX86_K6        IX86_AMD | IX86_CPU586
#define IX86_3DNOW     IX86_AMD | IX86_MMX686
#define IX86_ATHLON    IX86_AMD | IX86_MMX786
#define IX86_CYRIX486  IX86_CYRIX | IX86_CPU486
#define IX86_CYRIX686  IX86_CYRIX | IX86_CPU586
#define IX86_CYRIX686MMX IX86_CYRIX | IX86_MMX586

typedef struct tag_ix86opcodes
{
  const char *  name16;
  const char *  name32;
  ix86_method   method;
  unsigned long pro_clone;
}ix86_Opcodes;

typedef struct tag_ix86ExOpcodes
{
  const char *  name;
  ix86_method   method;
  unsigned long pro_clone;
}ix86_ExOpcodes;

typedef struct tag_ix3dNowopcodes
{
  const char *  name;
  unsigned long pro_clone;
}ix86_3dNowopcodes;


extern ix86_Opcodes ix86_table[];
extern ix86_ExOpcodes ix86_extable[];
extern ix86_3dNowopcodes ix86_3dNowtable[];

extern char ix86_segpref[];
extern const char * ix86_sizes[];

extern const char * ix86_A16[];

extern const char * ix86_ByteRegs[];
extern const char * ix86_WordRegs[];
extern const char * ix86_DWordRegs[];
extern const char * ix86_MMXRegs[];
extern const char * ix86_XMMXRegs[];
extern const char * ix86_SegRegs[];
extern const char * ix86_CrxRegs[];
extern const char * ix86_DrxRegs[];
extern const char * ix86_TrxRegs[];
extern const char * ix86_XrxRegs[];

extern const char * ix86_Op1Names[];
extern const char * ix86_ShNames[];
extern const char * ix86_Gr1Names[];
extern const char * ix86_Gr2Names[];

extern const char * ix86_ExGrp0[];
extern const char * ix86_ExGrp1[];
extern const char * ix86_BitGrpNames[];

extern const char * ix86_MMXGr1[];
extern const char * ix86_MMXGr2[];
extern const char * ix86_MMXGr3[];

extern const char * ix86_KatmaiGr1Names[];
extern const char * ix86_KatmaiGr2Names[];

extern char * ix86_getModRM(unsigned char w,unsigned char mod,unsigned char rm,ix86Param *DisP);
extern void   ix86_setModifier(char *str,const char *modf);
extern const char * ix86_getREG(unsigned char reg,unsigned char w);
extern char * ix86_CStile(char *str,const char *arg2);
extern char * ix86_getTilenWnD( ix86Param *DisP );
extern char * ix86_getTileWnD( ix86Param *DisP );
extern char * ix86_getTilenWD( ix86Param *DisP );
extern char * ix86_getTileWD( ix86Param *DisP );
extern char * ix86_GetDigitTile(char w,MBuffer tile,ix86Param* DisP);
extern char * TabSpace(char * str,unsigned nSpace);


extern void          ix86_ArgES(char *str,ix86Param *);
extern void          ix86_ArgDS(char *str,ix86Param *);
extern void          ix86_ArgSS(char *str,ix86Param *);
extern void          ix86_ArgCS(char *str,ix86Param *);
extern void          ix86_ArgFS(char *str,ix86Param *);
extern void          ix86_ArgGS(char *str,ix86Param *);
extern void          ix86_ArgByte(char *str,ix86Param *);
extern void          ix86_ArgWord(char *str,ix86Param *);
extern void          ix86_ArgDWord(char *str,ix86Param *);
extern void          ix86_ArgShort(char *str,ix86Param *);
extern void          ix86_ArgNear(char *str,ix86Param *);
extern void          ix86_ArgFar(char *str,ix86Param *);
extern void          ix86_ArgModRM(char *str,ix86Param *);
extern void          ix86_ArgModRMDW(char *str,ix86Param *);
extern void          ix86_ArgModRMnDW(char *str,ix86Param *);
extern void          ix86_ArgModRMDnW(char *str,ix86Param *);
extern void          ix86_ArgModRMnDnW(char *str,ix86Param *);
extern void          ix86_ArgMod(char *str,ix86Param *);
extern void          ix86_ArgModB(char *str,ix86Param *);
extern void          ix86_SMov(char * str,ix86Param *DisP);
extern void          ix86_ArgSegRM(char * str,ix86Param *DisP);
extern void          ix86_ArgAXMem(char *str,ix86Param *DisP);
extern void          ix86_ArgAXDigit(char *str,ix86Param *);
extern void          ix86_ArgRMDigit(char *str,ix86Param *);
extern void          ix86_ArgIRegDigit(char *str,ix86Param *);
extern void          ix86_ArgIReg(char *str,ix86Param *);
extern void          ix86_ArgAXIReg(char *str,ix86Param *);
extern void          ix86_ArgSInt(char *str,ix86Param *);
extern void          ix86_ArgInt(char *str,ix86Param *);
extern void          ix86_ArgRegRMDigit(char *str,ix86Param *);
extern void          ix86_ArgRmDigit(char *str,ix86Param *DisP,char w);
extern void          ix86_ArgOp1(char *str,ix86Param *);
extern void          ix86_ArgOp2(char *str,ix86Param *);
extern void          ix86_ShOp2(char *str,ix86Param *);
extern void          ix86_ShOp1(char *str,ix86Param *);
extern void          ix86_ShOpCL(char *str,ix86Param *);
extern void          ix86_ArgGrp1(char *str,ix86Param *);
extern void          ix86_ArgGrp2(char *str,ix86Param *);
extern void          ix86_InOut(char *str,ix86Param *);
extern void          ix86_FPUCmd(char *str,ix86Param *);

extern void          ix86_ExOpCodes(char *str,ix86Param *);
extern void          ix86_3dNowOpCodes(char *str,ix86Param *);
extern void          ix86_BitGrp(char *str,ix86Param *);

extern void          ix86_ArgExGr0(char *str,ix86Param *);
extern void          ix86_ArgExGr1(char *str,ix86Param *);
extern void          ix86_ArgMovXRY(char *str,ix86Param *);
extern void          ix86_DblShift(char *str,ix86Param *);
extern void          ix86_ArgMMXD(char *str,ix86Param *);
extern void          ix86_ArgMMXnD(char *str,ix86Param *);
extern void          ix86_ArgMMXRD(char *str,ix86Param *);
extern void          ix86_ArgMMXRnD(char *str,ix86Param *);
extern void          ix86_ArgMMXGr1(char *str,ix86Param *);
extern void          ix86_ArgMMXGr2(char *str,ix86Param *);
extern void          ix86_ArgMMXGr3(char *str,ix86Param *);
extern void          ix86_ArgMMRMDigit(char *str,ix86Param *);
extern void          ix86_ArgRMMXRevD(char *str,ix86Param *);
extern void          ix86_ArgRMMXRevnD(char *str,ix86Param *);
extern void          ix86_ArgXMMXD(char *str,ix86Param *);
extern void          ix86_ArgXMMXnD(char *str,ix86Param *);
extern void          ix86_ArgXMMXRD(char *str,ix86Param *);
extern void          ix86_ArgXMMXRnD(char *str,ix86Param *);
extern void          ix86_ArgRXMMXD(char *str,ix86Param *);
extern void          ix86_ArgRXMMXnD(char *str,ix86Param *);
extern void          ix86_ArgRXMMXRevD(char *str,ix86Param *);
extern void          ix86_ArgRXMMXRevnD(char *str,ix86Param *);
extern void          ix86_ArgXMMXMMD(char *str,ix86Param *);
extern void          ix86_ArgXMMXMMnD(char *str,ix86Param *);
extern void          ix86_ArgMMXMMXD(char *str,ix86Param *);
extern void          ix86_ArgMMXMMXnD(char *str,ix86Param *);
extern void          ix86_ArgXMMXGr1(char *str,ix86Param *);
extern void          ix86_ArgXMMXGr2(char *str,ix86Param *);
extern void          ix86_ArgXMMXGr3(char *str,ix86Param *);
extern void          ix86_ArgKatmaiGrp1(char *str,ix86Param *);
extern void          ix86_ArgKatmaiGrp2(char *str,ix86Param *);
extern void          ix86_ArgXMMRMDigit(char *str,ix86Param *);
extern void          ix86_ArgRegXMMDigit(char *str,ix86Param *);
extern void          ix86_ArgXMMRegDigit(char *str,ix86Param *);
extern void          ix86_ArgMovYX(char *str,ix86Param *);

#endif
