/****************************************************************************\
*   IX86_0F - this source file is part of Binary vIEW project (BIEW)         *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1995, 2000 Nick Kurshev                            *
*   License:              See below                                          *
*   Author and developer: Nick Kurshev                                       *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   plugins/disasm/ix86/ix86_0F.c                      *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  This file contain implementation of Intel x86 disassembler with 0x0F      *
* leading instructions set.                                                  *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/
#include "plugins/disasm/ix86/ix86.h"
#include "biewutil.h"
#include <string.h>
#include <limits.h>

char *ix86_Katmai_buff;

void          ix86_ExOpCodes(char *str,ix86Param *DisP)
{
 unsigned char code;
 DisP->RealCmd = &DisP->RealCmd[1];
 code = DisP->RealCmd[0];
 DisP->codelen++;
 strcpy(str,ix86_extable[code].name);
 if(ix86_extable[code].method)
 {
   TabSpace(str,TAB_POS);
   ix86_extable[code].method(str,DisP);
 }
 if(DisP->pro_clone < ix86_extable[code].pro_clone) DisP->pro_clone = ix86_extable[code].pro_clone;
}

void           ix86_ArgExGr0(char *str,ix86Param *DisP)
{
    unsigned char rm = (DisP->RealCmd[1] & 0x38) >> 3;
    unsigned char reg = (DisP->RealCmd[1] & 0x07);
    unsigned char mod = (DisP->RealCmd[1] & 0xC0) >> 6;
    strcpy(str,ix86_ExGrp0[rm]);
    TabSpace(str,TAB_POS);
    DisP->codelen++;
    ix86_setModifier(str,ix86_sizes[WORD_PTR]);
    strcat(str,ix86_getModRM(1,mod,reg,DisP));
}

void           ix86_ArgExGr1(char *str,ix86Param *DisP)
{
    unsigned char rm = (DisP->RealCmd[1] & 0x38) >> 3;
    unsigned char reg = (DisP->RealCmd[1] & 0x07);
    unsigned char mod = (DisP->RealCmd[1] & 0xC0) >> 6;
    unsigned char buf,ptr;
    int word;
    strcpy(str,ix86_ExGrp1[rm]);
    TabSpace(str,TAB_POS);
    DisP->codelen++;
    buf = DisP->RealCmd[1];
    ptr = PWORD_PTR;
    word = 1;
    if(rm == 7)
    {
      word = 0;
      DisP->pro_clone = IX86_CPU486;
      buf = 1;
      ptr = DUMMY_PTR;
    }
    else  if(rm == 4 || rm == 6) ptr = WORD_PTR;
    ix86_setModifier(str,ix86_sizes[ptr]);
    strcat(str,ix86_getModRM(word ? 1 : buf & 0x01,mod,reg,DisP));
}

static const char ** xry[] = { ix86_CrxRegs, ix86_DrxRegs, ix86_TrxRegs, ix86_XrxRegs };

void           ix86_ArgMovXRY(char *str,ix86Param *DisP)
{
    unsigned char code = DisP->RealCmd[0];
    unsigned char xreg = (DisP->RealCmd[1] & 0x38) >> 3;
    unsigned char reg = DisP->RealCmd[1] & 0x07;
    unsigned char ridx = (code & 0x01) + ((code >> 1) & 0x02);
    unsigned char d = (code >> 1) & 0x01;
    DisP->codelen++;
    Use32Data = CHAR_MAX;
    strcat(str,d ? xry[ridx][xreg] : ix86_getREG(reg,1));
    ix86_CStile(str,d ? ix86_getREG(reg,1) : xry[ridx][xreg]);
}

void           ix86_DblShift(char *str,ix86Param *DisP)
{
    unsigned char code = DisP->RealCmd[0];
    char *a1,*a2;
    a1 = ix86_getTileWnD(DisP);
    strcat(str,a1);
    a2 = code & 0x01 ? "cl" : ix86_GetDigitTile(0,&DisP->RealCmd[DisP->codelen],DisP);
    ix86_CStile(str,a2);
    DisP->codelen++;
}

static void          ix86_ArgMM(char *str,ix86Param *DisP,unsigned char direct)
{
   if((DisP->RealCmd[1] & 0xC0) != 0xC0)
   {
     UseMMXSet = CHAR_MAX;
     direct ? ix86_ArgModRMnDW(str,DisP) : ix86_ArgModRMDW(str,DisP);
     UseMMXSet = 0;
   }
   else
   {
      const char *mmx,*exx;
      UseMMXSet = CHAR_MAX;
      mmx = ix86_getREG((DisP->RealCmd[1] >> 3) & 0x07,1);
      UseMMXSet = 0;
      exx = ix86_getREG(DisP->RealCmd[1] & 0x07,1);
      strcat(str,direct ? exx : mmx);
      strcat(str,",");
      strcat(str,direct ? mmx : exx);
      DisP->codelen++;
   }
}

static void          ix86_ArgMMRev(char *str,ix86Param *DisP,unsigned char direct)
{
   if((DisP->RealCmd[1] & 0xC0) != 0xC0)
   {
     UseMMXSet = CHAR_MAX;
     direct ? ix86_ArgModRMnDW(str,DisP) : ix86_ArgModRMDW(str,DisP);
     UseMMXSet = 0;
   }
   else
   {
      const char *mmx,*exx;
      UseMMXSet = CHAR_MAX;
      mmx = ix86_getREG(DisP->RealCmd[1] & 0x07,1);
      UseMMXSet = 0;
      exx = ix86_getREG((DisP->RealCmd[1] >> 3) & 0x07,1);
      strcat(str,direct ? exx : mmx);
      strcat(str,",");
      strcat(str,direct ? mmx : exx);
      DisP->codelen++;
   }
}

void         ix86_ArgMMXD(char *str,ix86Param *DisP)
{
     UseMMXSet = CHAR_MAX;
     ix86_ArgModRMnDW(str,DisP);
     UseMMXSet = 0;
}

void         ix86_ArgMMXnD(char *str,ix86Param *DisP)
{
     UseMMXSet = CHAR_MAX;
     ix86_ArgModRMDW(str,DisP);
     UseMMXSet = 0;
}

void         ix86_ArgMMXRnD(char *str,ix86Param *DisP)
{
   ix86_ArgMM(str,DisP,0);
}

void         ix86_ArgMMXRD(char *str,ix86Param *DisP)
{
   ix86_ArgMM(str,DisP,1);
}

void         ix86_ArgMMXGr1(char *str,ix86Param *DisP)
{
  unsigned char code1 = (DisP->RealCmd[1] >> 3) & 0x07;
  DisP->codelen++;
  strcpy(str,ix86_MMXGr1[code1]);
  TabSpace(str,TAB_POS);
  UseMMXSet = CHAR_MAX;
  strcat(str,ix86_getREG(DisP->RealCmd[1] & 0x07,1));
  UseMMXSet = 0;
  strcat(str,",");
  strcat(str,Get2Digit(DisP->RealCmd[2]));
  DisP->codelen++;
}

void         ix86_ArgMMXGr2(char *str,ix86Param *DisP)
{
  unsigned char code1 = (DisP->RealCmd[1] >> 3) & 0x07;
  DisP->codelen++;
  strcpy(str,ix86_MMXGr2[code1]);
  TabSpace(str,TAB_POS);
  UseMMXSet = CHAR_MAX;
  strcat(str,ix86_getREG(DisP->RealCmd[1] & 0x07,1));
  UseMMXSet = 0;
  strcat(str,",");
  strcat(str,Get2Digit(DisP->RealCmd[2]));
  DisP->codelen++;
}

void         ix86_ArgMMXGr3(char *str,ix86Param *DisP)
{
  unsigned char code1 = (DisP->RealCmd[1] >> 3) & 0x07;
  DisP->codelen++;
  strcpy(str,ix86_MMXGr3[code1]);
  TabSpace(str,TAB_POS);
  UseMMXSet = CHAR_MAX;
  strcat(str,ix86_getREG(DisP->RealCmd[1] & 0x06,1));
  UseMMXSet = 0;
  strcat(str,",");
  strcat(str,Get2Digit(DisP->RealCmd[2]));
  DisP->codelen++;
}

static void          ix86_ArgXMM(char *str,ix86Param *DisP,unsigned char direct)
{
   if((DisP->RealCmd[1] & 0xC0) != 0xC0)
   {
     UseXMMXSet = CHAR_MAX;
     direct ? ix86_ArgModRMnDW(str,DisP) : ix86_ArgModRMDW(str,DisP);
     UseXMMXSet = 0;
   }
   else
   {
      const char *mmx,*exx;
      UseXMMXSet = CHAR_MAX;
      mmx = ix86_getREG(DisP->RealCmd[1] & 0x07,1);
      UseXMMXSet = 0;
      exx = ix86_getREG((DisP->RealCmd[1] >> 3) & 0x07,1);
      strcat(str,direct ? exx : mmx);
      strcat(str,",");
      strcat(str,direct ? mmx : exx);
      DisP->codelen++;
   }
}

static void          ix86_ArgRXMM(char *str,ix86Param *DisP,unsigned char direct)
{
   if((DisP->RealCmd[1] & 0xC0) != 0xC0)
   {
     UseXMMXSet = CHAR_MAX;
     direct ? ix86_ArgModRMnDW(str,DisP) : ix86_ArgModRMDW(str,DisP);
     UseXMMXSet = 0;
   }
   else
   {
      const char *mmx,*exx;
      UseXMMXSet = CHAR_MAX;
      mmx = ix86_getREG((DisP->RealCmd[1] >> 3) & 0x07,1);
      UseXMMXSet = 0;
      exx = ix86_getREG(DisP->RealCmd[1] & 0x07,1);
      strcat(str,direct ? exx : mmx);
      strcat(str,",");
      strcat(str,direct ? mmx : exx);
      DisP->codelen++;
   }
}

static void          ix86_ArgRXMMRev(char *str,ix86Param *DisP,unsigned char direct)
{
   if((DisP->RealCmd[1] & 0xC0) != 0xC0)
   {
     direct ? ix86_ArgModRMnDW(str,DisP) : ix86_ArgModRMDW(str,DisP);
   }
   else
   {
      const char *mmx,*exx;
      UseXMMXSet = CHAR_MAX;
      mmx = ix86_getREG(DisP->RealCmd[1] & 0x07,1);
      UseXMMXSet = 0;
      exx = ix86_getREG((DisP->RealCmd[1] >> 3) & 0x07,1);
      strcat(str,direct ? mmx : exx);
      strcat(str,",");
      strcat(str,direct ? exx : mmx);
      DisP->codelen++;
   }
}

static void          ix86_ArgRMMRev(char *str,ix86Param *DisP,unsigned char direct)
{
   if((DisP->RealCmd[1] & 0xC0) != 0xC0)
   {
     direct ? ix86_ArgModRMnDW(str,DisP) : ix86_ArgModRMDW(str,DisP);
   }
   else
   {
      const char *mmx,*exx;
      UseMMXSet = CHAR_MAX;
      mmx = ix86_getREG(DisP->RealCmd[1] & 0x07,1);
      UseMMXSet = 0;
      exx = ix86_getREG((DisP->RealCmd[1] >> 3) & 0x07,1);
      strcat(str,direct ? mmx : exx);
      strcat(str,",");
      strcat(str,direct ? exx : mmx);
      DisP->codelen++;
   }
}

static void          ix86_ArgXMMMM(char *str,ix86Param *DisP,unsigned char direct)
{
   if((DisP->RealCmd[1] & 0xC0) != 0xC0)
   {
     UseXMMXSet = CHAR_MAX;
     direct ? ix86_ArgModRMnDW(str,DisP) : ix86_ArgModRMDW(str,DisP);
     UseXMMXSet = 0;
   }
   else
   {
      const char *mmx,*exx;
      UseXMMXSet = CHAR_MAX;
      mmx = ix86_getREG((DisP->RealCmd[1] >> 3) & 0x07,1);
      UseXMMXSet = 0;
      UseMMXSet = CHAR_MAX;
      exx = ix86_getREG(DisP->RealCmd[1] & 0x07,1);
      UseMMXSet = 0;
      strcat(str,direct ? exx : mmx);
      strcat(str,",");
      strcat(str,direct ? mmx : exx);
      DisP->codelen++;
   }
}

static void          ix86_ArgMMXMM(char *str,ix86Param *DisP,unsigned char direct)
{
   if((DisP->RealCmd[1] & 0xC0) != 0xC0)
   {
     UseMMXSet = CHAR_MAX;
     direct ? ix86_ArgModRMnDW(str,DisP) : ix86_ArgModRMDW(str,DisP);
     UseMMXSet = 0;
   }
   else
   {
      const char *mmx,*exx;
      UseMMXSet = CHAR_MAX;
      mmx = ix86_getREG((DisP->RealCmd[1] >> 3) & 0x07,1);
      UseMMXSet = 0;
      UseXMMXSet = CHAR_MAX;
      exx = ix86_getREG(DisP->RealCmd[1] & 0x07,1);
      UseXMMXSet = 0;
      strcat(str,direct ? exx : mmx);
      strcat(str,",");
      strcat(str,direct ? mmx : exx);
      DisP->codelen++;
   }
}

void         ix86_ArgXMMXD(char *str,ix86Param *DisP)
{
     UseXMMXSet = CHAR_MAX;
     ix86_ArgModRMnDW(str,DisP);
     UseXMMXSet = 0;

}

void         ix86_ArgXMMXnD(char *str,ix86Param *DisP)
{
     UseXMMXSet = CHAR_MAX;
     ix86_ArgModRMDW(str,DisP);
     UseXMMXSet = 0;
}

void         ix86_ArgXMMXRnD(char *str,ix86Param *DisP)
{
   ix86_ArgXMM(str,DisP,0);
}

void         ix86_ArgXMMXRD(char *str,ix86Param *DisP)
{
   ix86_ArgXMM(str,DisP,1);
}

void         ix86_ArgRXMMXnD(char *str,ix86Param *DisP)
{
   ix86_ArgRXMM(str,DisP,0);
}

void         ix86_ArgRXMMXD(char *str,ix86Param *DisP)
{
   ix86_ArgRXMM(str,DisP,1);
}

void         ix86_ArgRXMMXRevnD(char *str,ix86Param *DisP)
{
   ix86_ArgRXMMRev(str,DisP,0);
}

void         ix86_ArgRXMMXRevD(char *str,ix86Param *DisP)
{
   ix86_ArgRXMMRev(str,DisP,1);
}

void         ix86_ArgRMMXRevnD(char *str,ix86Param *DisP)
{
   ix86_ArgRMMRev(str,DisP,0);
}

void         ix86_ArgRMMXRevD(char *str,ix86Param *DisP)
{
   ix86_ArgRMMRev(str,DisP,1);
}

void         ix86_ArgXMMXMMnD(char *str,ix86Param *DisP)
{
   ix86_ArgXMMMM(str,DisP,0);
}

void         ix86_ArgXMMXMMD(char *str,ix86Param *DisP)
{
   ix86_ArgXMMMM(str,DisP,1);
}

void         ix86_ArgMMXMMXnD(char *str,ix86Param *DisP)
{
   ix86_ArgMMXMM(str,DisP,0);
}

void         ix86_ArgMMXMMXD(char *str,ix86Param *DisP)
{
   ix86_ArgMMXMM(str,DisP,1);
}

void         ix86_ArgXMMXGr1(char *str,ix86Param *DisP)
{
  unsigned char code1 = (DisP->RealCmd[1] >> 3) & 0x07;
  DisP->codelen++;
  strcpy(str,ix86_MMXGr1[code1]);
  TabSpace(str,TAB_POS);
  UseXMMXSet = CHAR_MAX;
  strcat(str,ix86_getREG(DisP->RealCmd[1] & 0x07,1));
  UseXMMXSet = 0;
  strcat(str,",");
  strcat(str,Get2Digit(DisP->RealCmd[2]));
  DisP->codelen++;
}

void         ix86_ArgXMMXGr2(char *str,ix86Param *DisP)
{
  unsigned char code1 = (DisP->RealCmd[1] >> 3) & 0x07;
  DisP->codelen++;
  strcpy(str,ix86_MMXGr2[code1]);
  TabSpace(str,TAB_POS);
  UseXMMXSet = CHAR_MAX;
  strcat(str,ix86_getREG(DisP->RealCmd[1] & 0x07,1));
  UseXMMXSet = 0;
  strcat(str,",");
  strcat(str,Get2Digit(DisP->RealCmd[2]));
  DisP->codelen++;
}

void         ix86_ArgXMMXGr3(char *str,ix86Param *DisP)
{
  unsigned char code1 = (DisP->RealCmd[1] >> 3) & 0x07;
  DisP->codelen++;
  strcpy(str,ix86_MMXGr3[code1]);
  TabSpace(str,TAB_POS);
  UseXMMXSet = CHAR_MAX;
  strcat(str,ix86_getREG(DisP->RealCmd[1] & 0x06,1));
  UseXMMXSet = 0;
  strcat(str,",");
  strcat(str,Get2Digit(DisP->RealCmd[2]));
  DisP->codelen++;
}

void ix86_ArgMovYX(char *str,ix86Param *DisP)
{
      const char *dst,*src;
      unsigned char u32data,mod;
      mod = (DisP->RealCmd[1] >> 6) & 0x03;
      ix86_setModifier(str,DisP->RealCmd[0] & 0x01 ? Use32Data ? ix86_sizes[DWORD_PTR] : ix86_sizes[WORD_PTR] : ix86_sizes[BYTE_PTR]);
      dst = ix86_getREG((DisP->RealCmd[1] >> 3) & 0x07,1);
      u32data = Use32Data;
      Use32Data = 0;
      src = ix86_getModRM(DisP->RealCmd[0] & 0x01,mod,DisP->RealCmd[1] & 0x07,DisP);
      Use32Data = u32data;
      strcat(str,dst);
      strcat(str,",");
      strcat(str,src);
      DisP->codelen++;
}

void ix86_BitGrp(char *str,ix86Param *DisP)
{
   unsigned char idx;
   char mod = (DisP->RealCmd[1] >> 6) & 0x03;
   char rm = (DisP->RealCmd[1] & 0x07);
   idx = (DisP->RealCmd[1] & 0x38) >> 3;
   strcpy(str,ix86_BitGrpNames[idx]);
   TabSpace(str,TAB_POS);
   strcat(str,ix86_getModRM(1,mod,rm,DisP));
   strcat(str,",");
   strcat(str,ix86_GetDigitTile(0,&DisP->RealCmd[DisP->codelen++],DisP));
}

void          ix86_ArgKatmaiGrp1(char *str,ix86Param *DisP)
{
   unsigned char code2 = ( DisP->RealCmd[1] & 0x38 ) >> 3;
   strcpy(str,ix86_KatmaiGr1Names[code2]);
   if(DisP->RealCmd[1] != 0xF8)
   {
     TabSpace(str,TAB_POS);
     ix86_ArgMod(str,DisP);
   }
   else DisP->codelen++;
}

void          ix86_ArgKatmaiGrp2(char *str,ix86Param *DisP)
{
   unsigned char code2 = ( DisP->RealCmd[1] & 0x38 ) >> 3;
   strcpy(str,ix86_KatmaiGr2Names[code2]);
   TabSpace(str,TAB_POS);
   ix86_ArgMod(str,DisP);
}

void          ix86_ArgXMMRMDigit(char *str,ix86Param *DisP)
{
   char * a1,*a2;
   UseXMMXSet = CHAR_MAX;
   DisP->codelen++;
   a1 = ix86_getTileWD(DisP);
   strcat(str,a1);
   a2 = ix86_GetDigitTile(0,&DisP->RealCmd[DisP->codelen-1],DisP);
   ix86_CStile(str,a2);
   UseXMMXSet = 0;
}

void          ix86_ArgMMRMDigit(char *str,ix86Param *DisP)
{
   char * a1,*a2;
   DisP->codelen++;
   UseMMXSet = CHAR_MAX;
   a1 = ix86_getTileWD(DisP);
   strcat(str,a1);
   a2 = ix86_GetDigitTile(0,&DisP->RealCmd[DisP->codelen-1],DisP);
   ix86_CStile(str,a2);
   UseMMXSet = 0;
}

void          ix86_ArgRegXMMDigit(char *str,ix86Param *DisP)
{
   char * a1;
   ix86_ArgMMRev(str,DisP,1);
   a1 = ix86_GetDigitTile(0,&DisP->RealCmd[DisP->codelen-1],DisP);
   ix86_CStile(str,a1);
}

void          ix86_ArgXMMRegDigit(char *str,ix86Param *DisP)
{
   char * a1;
   ix86_ArgMMXRnD(str,DisP);
   a1 = ix86_GetDigitTile(0,&DisP->RealCmd[DisP->codelen-1],DisP);
   ix86_CStile(str,a1);
}

void ix86_3dNowOpCodes( char *str,ix86Param *DisP)
{
 unsigned char code;
 UseMMXSet = CHAR_MAX;
 ix86_Katmai_buff[0] = 0;
 ix86_ArgModRMDW(ix86_Katmai_buff,DisP);
 UseMMXSet = 0;
 code = DisP->RealCmd[DisP->codelen-1];
 DisP->codelen++;
 strcpy(str,ix86_3dNowtable[code].name);
 TabSpace(str,TAB_POS);
 strcat(str,ix86_Katmai_buff);
 DisP->pro_clone = ix86_3dNowtable[code].pro_clone;
}

