/****************************************************************************\
*   TEXTMODE - this source file is part of Binary vIEW project (BIEW)        *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1995, 2000 Nick Kurshev                            *
*   License:              See below                                          *
*   Author and developer: Nick Kurshev                                       *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   plugins/mod/textmode.c                             *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  This file contain implementation of text viewer with different submodes   *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <limits.h>

#include "plugins/mod/textmode.h"
#include "pmalloc.h"
#include "kbd_code.h"
#include "bmfile.h"
#include "file_ini.h"
#include "biewutil.h"
#include "biewhelp.h"
#include "bconsole.h"
#include "biewlib.h"
#include "reg_form.h"

extern REGISTRY_NLS RussianNLS;

static REGISTRY_NLS *nls_set[] =
{
  &RussianNLS
};

static REGISTRY_NLS *activeNLS = &RussianNLS;
static unsigned      defNLSSet = 0;

#define TEXT_TAB 8

#define MOD_PLAIN    0
#define MOD_BINARY   1

#define MOD_MAXMODE  1

typedef struct tagTSTR
{
  unsigned long st;
  unsigned long end;
}TSTR;

static char * mod_names[] =
{
   "~Plain text",
   "~All characters 8bit (as is)"
};

static unsigned bin_mode = MOD_PLAIN;
static BGLOBAL txtHandle = &bNull;

#define MAX_STRLEN 1000
TSTR *tlines,*ptlines;
static unsigned int maxstrlen = MAX_STRLEN;
unsigned int strmaxlen = __TVIO_MAXSCREENWIDTH-2;
static unsigned char wmode; /* wrap mode */

static unsigned long PrevPageSize,CurrPageSize,PrevStrLen,CurrStrLen;

void  txt_cvt_full(char * str,int size,const unsigned char *tmpl)
{
  int i;
  for(i = 0;i < size;i++) str[i] = tmpl[(unsigned char)str[i]];
}

void txt_cvt_hi80(char * str,unsigned size,const unsigned char *tmpl)
{
 int i;
 unsigned char cc,ch;
 for(i = 0;i < size;i++)
 {
   cc = str[i];
   if(cc >= 0x80) ch = tmpl[cc-0x80];
   else           ch = cc;
  str[i] = ch;
 }
}

void txt_cvt_lo80(char * str,unsigned size,const unsigned char *tmpl)
{
 int i;
 unsigned char cc,ch;
 for(i = 0;i < size;i++)
 {
   cc = str[i];
   if(cc < 0x80)  ch = tmpl[cc];
   else           ch = cc;
  str[i] = ch;
 }
}

static unsigned char __NEAR__ nlsReadByte(unsigned long cp)
{
 char nls_buff[256];
 unsigned sym_size;
 sym_size = activeNLS->get_symbol_size();
 bioSeek(txtHandle,cp,SEEK_SET);
 bioReadBuffer(txtHandle,nls_buff,sym_size);
 activeNLS->convert_buffer(nls_buff,sym_size);
 return (unsigned char)nls_buff[0];
}

static unsigned long __NEAR__ BackScanCR( unsigned long cp )
{
 unsigned long lval;
 unsigned int freq;
 unsigned cp_symb_len;
 char ch;
 cp_symb_len = activeNLS->get_symbol_size();
 ch = nlsReadByte(cp);
 if(ch == '\n' || ch == '\r')
 {
     char ch1;
     if(cp >= cp_symb_len) cp-=cp_symb_len;
     ch1 = nlsReadByte(cp);
     if((ch1 == '\n' || ch1 == '\r') && ch != ch1)
       if(cp >= cp_symb_len)
         cp-=cp_symb_len;
 }
 for(lval = freq = 0;cp;cp-=cp_symb_len)
 {
    ch = nlsReadByte(cp);
    if(ch == '\n' || ch == '\r')
    {
      lval = cp + cp_symb_len;
      break;
    }
    freq++;
    if(freq > maxstrlen) { lval = cp; break; }
 }
 return lval;
}

static unsigned long __NEAR__ ForwardScanCR( unsigned long cp ,unsigned long flen)
{
 unsigned long lval;
 unsigned int freq = 0;
 unsigned cp_symb_len;
 cp_symb_len = activeNLS->get_symbol_size();

 for(;cp < flen;cp+=cp_symb_len)
 {
    unsigned char ch;
    ch = nlsReadByte(cp);
    if(ch == '\n' || ch == '\r')
    {
      char ch1;
      lval = cp + cp_symb_len;
      ch1 = nlsReadByte(lval);
      if((ch1 == '\n' || ch1 == '\r') && ch != ch1)  lval+=cp_symb_len;
      return lval;
    }
    freq++;
    if(freq > maxstrlen) return cp;
 }
 return flen;
}

static void __NEAR__ FillPrevPage(unsigned long lval)
{
 unsigned cp_symb_len;
 int i;
 cp_symb_len = activeNLS->get_symbol_size();
 for(i = Height - 1;i >= 0;i--)
 {
    ptlines[i].end = lval;
    if(lval >= cp_symb_len) lval = BackScanCR(lval - cp_symb_len);
    ptlines[i].st = lval;
 }
}

static void __NEAR__ FillCurrPage(unsigned long lval,unsigned long flen)
{
 int i;
 for(i = 0;i < Height;i++)
 {
   tlines[i].st = lval;
   if(lval < flen) lval = ForwardScanCR(lval,flen);
   tlines[i].end = lval;
 }
}

unsigned long tmocpos = 0;

static void __NEAR__ PrepareLines(int keycode)
{
 int size,size1,h;
 unsigned cp_symb_len;
 unsigned long lval,flen,cp = BMGetCurrFilePos();
 cp_symb_len = activeNLS->get_symbol_size();
 flen = BMGetFLength();
 /* search begin of first string */
 h = Height - 1;
 size = sizeof(TSTR)*h;
 size1 = size + sizeof(TSTR);
 if(keycode == KE_UPARROW)
 {
   /* i.e. going down */
   if(tlines[0].st)
   {
     memmove(&tlines[1],tlines,size);
     tlines[0] = ptlines[h];
     memmove(&ptlines[1],ptlines,size);
     lval = ptlines[1].st;
     ptlines[0].end = lval;
     if(lval >= cp_symb_len) lval-=cp_symb_len;
     ptlines[0].st = BackScanCR(lval);
   }
 }
 else
   if(keycode == KE_DOWNARROW)
   {
     /* i.e. going up */
     if(tlines[0].end < flen)
     {
       memmove(ptlines,&ptlines[1],size);
       ptlines[h] = tlines[0];
       memmove(tlines,&tlines[1],size);
       lval = tlines[h - 1].end;
       tlines[h].st = lval;
       tlines[h].end = ForwardScanCR(lval,flen);
     }
   }
   else
     if(keycode == KE_PGUP)
     {
      if(ptlines[0].st)
      {
       memcpy(tlines,ptlines,size1);
       lval = tlines[0].st;
       FillPrevPage(lval);
      }
      else goto CommonPart;
     }
     else
       if(keycode == KE_PGDN)
       {
        if(tlines[Height - 1].end < flen)
        {
          memcpy(ptlines,tlines,size1);
          lval = ptlines[h].end;
          FillCurrPage(lval,flen);
        }
        else goto CommonPart;
       }
       else
         if(cp != tmocpos || keycode == KE_SUPERKEY)
         {
          CommonPart:
           lval = BackScanCR(cp);
           if(cp != lval) cp = lval;
           if(cp) FillPrevPage(lval);
           else { PrevStrLen = PrevPageSize = 2; memset(ptlines,0,sizeof(TSTR)*Height); }
           /* scan forward */
           FillCurrPage(lval,flen);
         }
 tmocpos = cp;
 PrevStrLen = ptlines[h].end - ptlines[h].st + cp_symb_len;
 PrevPageSize = ptlines[h].end - ptlines[0].st + cp_symb_len;
}

static int __NEAR__ Tab2Space(unsigned short * dest,unsigned int alen,char * str,unsigned int len,unsigned int shift,unsigned *n_tabs)
{
  unsigned char ch,defcol;
  int i,size,j,k;
  unsigned int freq;
  if(n_tabs) *n_tabs = 0;
  for(i = 0,freq = 0,k = 0;i < len;i++,freq++)
  {
    defcol = (Blue << 4) | LightCyan;
    ch = str[i];
    if(ch < 32)
    {
      switch(ch)
      {
         default:
         case 0: /*NUL  end string*/
         case 1: /*SOH  start of heading*/
         case 2: /*SOT  start of text*/
         case 3: /*ETX  end of text*/
         case 4: /*EOT  end of transmission*/
         case 5: /*ENQ  enquiry*/
         case 6: /*ACK  acknowledge*/
         case 7: /*BEL  bell*/
         case 11: /*VT  virtical tab*/
         case 12: /*FF  form feed*/
         case 14: /*SO  shift out*/
         case 15: /*SI  shift in*/
         case 16: /*DLE data line escape*/
         case 17: /*DC1 dev ctrl 1 (X-ON)*/
         case 18: /*DC2 dev ctrl 2*/
         case 19: /*DC3 dev ctrl 3 (X-OFF)*/
         case 20: /*DC4 dev ctrl 4*/
         case 21: /*NAK negative acknowledge*/
         case 22: /*SYN synhronous idel*/
         case 23: /*ETB end transmission block*/
         case 24: /*CAN cancel*/
         case 25: /*EM  end of medium*/
         case 26: /*SUB substitude*/
         case 27: /*ESC escape*/
         case 28: /*FS  file separator*/
         case 29: /*GS  group separator*/
         case 30: /*RS  record separator*/
         case 31: /*US  unit separator*/
                  break;
         case 9: /*HT   horizontal tab*/
               {
                  size = TEXT_TAB - (freq%TEXT_TAB);
                  for(j = 0;j < size;j++,freq++)
                  {
                    if(k < alen && freq >= shift)
                    {
                      if(dest) dest[k] = (' ' & 0x00FF) | (defcol << 8);
                      k++;
                    }
                  }
                  if(n_tabs) (*n_tabs)++;
               }
               freq--;
               break;
         case 10: /*LF  line feed*/
         case 13: /*CR  cariage return*/
                  goto End;
         case 8:  /*BS  backspace*/
               {
                  char pch;
                  pch = i ? str[i-1] : str[0];
                  switch(pch)
                  {
                    case '_': defcol = (Blue << 4) | Yellow; break;
                    default:  defcol = (Blue << 4) | White;
                  }
                  if(i < len) ch = str[++i];
                  if(k) k--;
                  freq--;
               }
               if(freq >= shift) goto DefChar;
               break;
      }
    }
    else
    {
      DefChar:
      if(k < alen && freq >= shift)
      {
        if(dest) dest[k] = (ch & 0x00FF) | (defcol << 8);
        k++;
      }
    }
  }
  End:
  return k;
}

static char *buff;
static unsigned buff1[__TVIO_MAXSCREENWIDTH];

static void __NEAR__ txtPaintSearch(char * cptr,unsigned int shift,int i,int size,int bin_mode)
{
    int sh,she;
    long save,savee;
    unsigned cp_symb_len;
    cp_symb_len = activeNLS->get_symbol_size();
    if(bin_mode)
    {
      sh = (int)(FoundTextSt - tlines[i].st) - shift;
      she = (int)(FoundTextEnd - tlines[i].st) - shift;
    }
    else
    {
      sh = Tab2Space(NULL,UINT_MAX,buff,(int)(FoundTextSt - tlines[i].st)/cp_symb_len,shift,NULL)*cp_symb_len;
      she= Tab2Space(NULL,UINT_MAX,buff,(int)(FoundTextEnd - tlines[i].st)/cp_symb_len,shift,NULL)*cp_symb_len;
    }
    save = FoundTextSt;
    savee= FoundTextEnd;
    FoundTextSt = tlines[i].st + shift + sh/cp_symb_len;
    FoundTextEnd = tlines[i].st + shift + she/cp_symb_len;
    HiLightSearch(MainWnd,tlines[i].st + shift,0,size,i,cptr,bin_mode == MOD_BINARY ? HLS_NORMAL : HLS_USE_BUFFER_AS_VIDEO);
    FoundTextSt = save;
    FoundTextEnd = savee;
}

static void __NEAR__ drawBound(int x,int y,char ch)
{
  twGotoXY(x,y);
  twTextColor(Yellow);
  twPutChar(ch);
  twTextColor(LightCyan);
}

unsigned txtConvertBuffer(char *buff,unsigned size)
{
  return activeNLS->convert_buffer(buff,size);
}

unsigned drawText( unsigned keycode , unsigned shift )
{
  int i,tmp,hilightline;
  int len,size,rsize,rshift,textmaxlen;
  unsigned long cpos;
  unsigned cp_symb_len;
  char * cptr;
  cp_symb_len = activeNLS->get_symbol_size();
  if(shift%cp_symb_len)
  {
    if(keycode == KE_RIGHTARROW || keycode == KE_PGDN)
                                     shift+=shift%cp_symb_len;
    else
                                     shift=(shift/cp_symb_len)*cp_symb_len;
  }
  maxstrlen = wmode ? tvioWidth : MAX_STRLEN;
  cpos = BMGetCurrFilePos();
  if(cpos%cp_symb_len)
  {
    if(keycode == KE_RIGHTARROW || keycode == KE_PGDN)
                                    cpos+=cpos%cp_symb_len;
    else
                                    cpos=(cpos/cp_symb_len)*cp_symb_len;
  }
  BMSeek(cpos,BIO_SEEK_SET);
  if(!(keycode == KE_LEFTARROW || keycode == KE_RIGHTARROW))
                                     PrepareLines(keycode);
  hilightline = -1;
  for(i = 0;i < Height;i++)
  {
    len = (int)(tlines[i].end - tlines[i].st);
    if(len > strmaxlen) strmaxlen = len;
    if(isHOnLine(tlines[i].st,len))
    {
      hilightline = i;
      if(bin_mode != MOD_BINARY && keycode == KE_JUSTFIND)
      {
        len = min(MAX_STRLEN,(int)(FoundTextSt - tlines[i].st));
        BMReadBufferEx((void *)buff,len,tlines[i].st,BM_SEEK_SET);
        len = txtConvertBuffer(buff,len);
        shift = Tab2Space(NULL,UINT_MAX,buff,len,0,NULL);
      }
      if(!isHOnLine((tlines[i].st+shift)*cp_symb_len,min(len,tvioWidth)) &&
          keycode == KE_JUSTFIND)
      {
          shift = (unsigned)(FoundTextSt - tlines[i].st)/cp_symb_len;
          if(bin_mode != MOD_BINARY)
          {
            unsigned n_tabs;
            shift = Tab2Space(NULL,UINT_MAX,buff,len,0,&n_tabs);
            if(shift) shift-=cp_symb_len;
          }
          else shift -= (tvioWidth/2)/cp_symb_len;
          if((int)shift < 0) shift = 0;
          if(shift%cp_symb_len) shift+=shift%cp_symb_len;
          PrepareLines(KE_SUPERKEY);
      }
    }
  }
  textmaxlen = maxstrlen - 2;
  for(i = 0;i < Height;i++)
  {
    len = (int)(tlines[i].end - tlines[i].st);
    if(bin_mode != MOD_BINARY) rshift = 0;
    else                       rshift = shift;
    rsize = size = len - rshift;
    if(size > 0)
    {
        BMReadBufferEx((void *)buff,size,tlines[i].st + rshift,BM_SEEK_SET);
        rsize = size = txtConvertBuffer(buff,size);
        if(bin_mode == MOD_BINARY)  cptr = buff;
        else
        {
             rsize = size = Tab2Space((unsigned short *)buff1,sizeof(buff1)/sizeof(unsigned),buff,size,shift,NULL);
             cptr = (char *)buff1;
             tmp = size + shift;
             if(strmaxlen < tmp) strmaxlen = tmp;
             if(textmaxlen < tmp) textmaxlen = tmp;
        }
        if(size > tvioWidth) size = tvioWidth;
        if(i == hilightline)
                    txtPaintSearch(cptr,shift,i,size,bin_mode == MOD_BINARY);
        else
        {
          if(bin_mode == MOD_BINARY) twDirectWrite(1,i+1,cptr,size);
          else                       twWriteBuffer(twUsedWin(),1,i + 1,cptr,size);
        }
        if(rsize < tvioWidth)
        {
          twGotoXY(1 + rsize,i + 1);
          twClrEOL();
        }
        else
          if(rsize > tvioWidth)
             drawBound(tvioWidth,i + 1,'>');
    }
    else
    {
       twGotoXY(1,i + 1);
       twClrEOL();
    }
    if(shift) drawBound(1,i + 1,'<');
    lastbyte = tlines[i].st;
    if(bin_mode == MOD_BINARY) lastbyte += shift + size;
    else                       lastbyte += rshift + len;
  }
  tmp = textmaxlen - tvioWidth + 2;
  if(shift > tmp) shift = tmp;
  if(!tlines[1].st) tlines[1].st = tlines[0].end;
  CurrStrLen = tlines[0].end - tlines[0].st;
  CurrPageSize = lastbyte - tlines[0].st;
  BMSeek(cpos,BM_SEEK_SET);
  return shift;
}

void HelpTxt( void )
{
   hlpDisplay(1001);
}

unsigned long txtPrevPageSize( void ) { return PrevPageSize; }
unsigned long txtCurrPageSize( void ) { return CurrPageSize; }
unsigned long txtPrevLineWidth( void ) { return PrevStrLen; }
unsigned long txtCurrLineWidth( void ) { return CurrStrLen; }
char *        txtMiscKeyName( void ) { return wmode ? "Unwrap" : "Wrap  "; }
void          txtMiscKeyAction( void ) { wmode = wmode ? 0 : UCHAR_MAX; }

int txtSelectCP( void )
{
  return activeNLS->select_table();
}

int txtSelectMode( void )
{
  unsigned nModes,i;
  nModes = sizeof(mod_names)/sizeof(char *);
  i = SelBoxA(mod_names,nModes," Select cyrillic code page: ",bin_mode);
  if(i != -1)
  {
    bin_mode = i;
    return 1;
  }
  return 0;
}

int txtSelectNLS( void )
{
  char *modeName[sizeof(nls_set)/sizeof(REGISTRY_NLS *)];
  unsigned nModes;
  int i;

  nModes = sizeof(nls_set)/sizeof(REGISTRY_NLS *);
  for(i = 0;i < nModes;i++) modeName[i] = nls_set[i]->set_name;
  i = SelBoxA(modeName,nModes," Select NLS set: ",defNLSSet);
  if(i != -1)
  {
    if(activeNLS->term) activeNLS->term();
    activeNLS = nls_set[i];
    if(activeNLS->init) activeNLS->init();
    defNLSSet = i;
    return 1;
  }
  return 0;
}


static int __NEAR__ isBinByte(unsigned char ch)
{
  return ch < 32 && !isspace(ch) && ch != 0x08 && ch != 0x1A;
}

int txtDetect( void )
{
  int maxl,i,bin = 0;
  unsigned long flen;
  maxl = 1000;
  flen = BMGetFLength();
  if(maxl > flen) maxl = (int)flen;
  for(i = 0;i < maxl;i++)
  {
   char ch;
   ch = BMReadByteEx(i,BM_SEEK_SET);
   if(isBinByte(ch & 0xFF)) { bin = 1; break; }
  }
  if(!bin) bin_mode = MOD_PLAIN;
  return bin ? 0 : 1;
}

void txtReadIni( hIniProfile *ini )
{
  char tmps[10];
  if(isValidIniArgs())
  {
    int w_m;
    biewReadProfileString(ini,"Biew","Browser","SubSubMode4","0",tmps,sizeof(tmps));
    defNLSSet = (unsigned)strtoul(tmps,NULL,10);
    if(defNLSSet > sizeof(nls_set)/sizeof(REGISTRY_NLS *)) defNLSSet = 0;
    activeNLS = nls_set[defNLSSet];
    if(activeNLS->init) activeNLS->init();
    activeNLS->read_ini(ini);
    biewReadProfileString(ini,"Biew","Browser","SubSubMode3","0",tmps,sizeof(tmps));
    bin_mode = (unsigned)strtoul(tmps,NULL,10);
    if(bin_mode > MOD_MAXMODE) bin_mode = 0;
    biewReadProfileString(ini,"Biew","Browser","MiscMode","0",tmps,sizeof(tmps));
    w_m = (int)strtoul(tmps,NULL,10);
    wmode = w_m ? UCHAR_MAX : 0;
 }
}

void txtSaveIni( hIniProfile *ini )
{
  char tmps[10];
  sprintf(tmps,"%i",bin_mode);
  biewWriteProfileString(ini,"Biew","Browser","SubSubMode3",tmps);
  biewWriteProfileString(ini,"Biew","Browser","MiscMode",wmode ? "1" : "0");
  sprintf(tmps,"%i",defNLSSet);
  biewWriteProfileString(ini,"Biew","Browser","SubSubMode4",tmps);
  activeNLS->save_ini(ini);
}

void txtInit( void )
{
   buff = PMalloc(MAX_STRLEN);
   tlines = PMalloc(sizeof(TSTR)*__TVIO_MAXSCREENWIDTH);
   ptlines = PMalloc(sizeof(TSTR)*__TVIO_MAXSCREENWIDTH);
   if((!buff) || (!tlines) || !(ptlines))
   {
     MemOutBox("Text mode initialization");
     exit(EXIT_FAILURE);
   }
   if((txtHandle = bioDup(BMbioHandle())) == &bNull) txtHandle = BMbioHandle();
}

void txtTerm( void )
{
   PFREE(buff);
   PFREE(tlines);
   PFREE(ptlines);
   if(txtHandle != BMbioHandle()) { bioClose(txtHandle); txtHandle = &bNull; }
}

unsigned txtCharSize( void ) { return activeNLS->get_symbol_size(); }

REGISTRY_MODE textMode =
{
  "~Text mode",
  { NULL, "CodPag", "TxMode", "NLSSet", NULL, NULL, NULL, NULL, NULL, NULL },
  { NULL, txtSelectCP, txtSelectMode, txtSelectNLS, NULL, NULL, NULL, NULL, NULL, NULL },
  txtDetect,
  1,
  drawText,
  txtConvertBuffer,
  txtCharSize,
  txtMiscKeyName,
  txtMiscKeyAction,
  txtPrevPageSize,
  txtCurrPageSize,
  txtPrevLineWidth,
  txtCurrLineWidth,
  HelpTxt,
  txtReadIni,
  txtSaveIni,
  txtInit,
  txtTerm
};


