/****************************************************************************\
*   REG_FORM - this header file is part of Binary vIEW project (BIEW)        *
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*
*   Copyrights:           1995, 2000 Nick Kurshev                            *
*   License:              See below                                          *
*   Author and developer: Nick Kurshev                                       *
*   Requirement:          ANSI C compatible development system               *
*   Original file name:   reg_form.h                                         *
*---------------------- [ D e s c r i p t i o n ] ---------------------------*
*  This file contain structure prototypes for embededding new file formats,  *
* translation modes e t.c. in BIEW project.                                  *
*-------------------------- [ L i c e n c e ] -------------------------------*
* The Binary vIEW (BIEW) is copyright (C) 1995 Nick Kurshev.                 *
* All rights reserved. This software is redistributable under the licence    *
* given in the file "Licence.en" ("Licence.ru" in russian translation)       *
* distributed in the BIEW archive.                                           *
\****************************************************************************/

#ifndef __REG_FORM__H
#define __REG_FORM__H

#ifdef __cplusplus
#define extern "C" {
#endif

#ifndef __FILE_INI_RUNTIME_SUPPORT_SYSTEM__
#include "file_ini.h"
#endif

typedef unsigned long (*BinFunc)( void );
typedef int           (*ModFunc)( void );

#define APREF_NORMAL    0x0000 /* Append references in short form if it really present in binary */
#define APREF_USE_TYPE  0x0001 /* Append references in detail form if it really present in binary */
#define APREF_TRY_LABEL 0x0002 /* Append references in short form even if it not present in binary (smart method) */

/* Arguments:
   str    - string buffer for write to
   shift  - physical address of field, that required of binding
   flags  - see above
   codelen- length of field, that required binding
   r_shift- used only if APPREF_TRY_LABEL flag is set contain real value
            of field, that required binding
   return   0 - if not refs is not appended !0 - otherwise
*/

typedef int           (*AppRefs)(char *str,unsigned long shift,int flags,int codelen,unsigned long r_shift);

/***************************************************************\
*  This form registry binary file formats                       *
\***************************************************************/

/* List of DisAssembler Bitness */

#define DAB_USE16   0
#define DAB_USE32   1
#define DAB_USE64   2
#define DAB_USE128  3
#define DAB_USE256  4
#define DAB_AUTO   -1 /* never must return! Only for selection */

/* Public symbols classes */

#define SC_LOCAL    0 /* means: present as entry but not exported */
#define SC_GLOBAL   1 /* measn: exported entry point */

/* object classes */

#define OC_CODE      0 /* for code objects */
#define OC_DATA      1 /* for any data objects */
#define OC_NOOBJECT -1 /* for non objects (means: relocs, resources, tables ...) */

/* Binary plugins state constants: */
/* Note:
     First called 'check_format' - plugin in UNPLUGGED state.
     Second called 'init' - plugin turn into PLUGGED IN state.
     Last called 'destroy' - plugin turn into UNPLUGGED state.
     These foundamental state is not defined through PS_XXX constants,
     because it meaningless.
     General purpose for PS_XXX state is activating of preemptive memory
     allocation mechanism. Means: when plugin in PS_INACTIVE state (when
     references resolving is 'None') then on low memory it can freed some
     buffers. (This need for prevent of infinity loop of memory
     allocation/freing, that possible in some case).
*/

#define PS_INACTIVE    0
#define PS_ACTIVE      1

typedef struct tag_REGISTRY_BIN
{
  char *    name;        /* name of binary format */
  char *    prompt[10];  /* on ALT-Fx selection */
  BinFunc   action[10];  /* action on ALT-Fx selection */
  int     (*check_format)( void ); /* check for format */
  void    (*init)( void );         /* init plugin (if check o'k) (constructor) */
  void    (*destroy)( void );      /* destroy plugin (destructor) */
  BinFunc   showHdr;     /* if not an MZ style format */
  AppRefs   bind;        /* for show references */
  void    (*set_state)(int state); /* called, when activity of plugin is
                                      changed. See above for PS_XXX
                                      constants.
                                      Note: Plugin must support counter of
                                      states. (For multiple call purpose) */

                         /* must return DAB_XXX. Quick version for
                            disassembler */
  int     (*query_bitness)(unsigned long);

                         /* for displaying offset within struct in left
                            address column */
  void    (*AddressResolving)(char *,unsigned long);

                         /* convert virtual address to physical (mean file
                            offset). Return 0 if not possible */
  unsigned long (*va2pa)(unsigned long va);

                         /* convert physical address to virtual. seg pointer
                            can be NULL */
  unsigned long (*pa2va)(unsigned long pa);


/*-- Below placed functions for 'put structures' method of save as dialog --*/

                         /* return public symbol (like: 'GetVersion')
                            to str at file offset - pa.
                            class is filled as print above with SC_XXX constants.
                            return 0 - if no symbol name available
                                       in given direction (as_prev)
                                  !0 - if success return physical address
                                       of public name in given direction */
  unsigned long (*GetPubSym)(char *str,unsigned cb_str,unsigned *class,
                             unsigned long pa,int as_prev);

                         /* Determine attributes of object at given physical
                            address (equ file offset).
                            Output:
                            name - filled with object name if possible,
                            start- contain begin of object
                            end  - contain end of object
                            WARNING!!! start of next object always equal of
                            end of current object.
                            class - see above for OC_XXX constants.
                            start and end arguments must fill anyway, and must
                            describe boundaries of current area. (For example:
                            if exe-format - new exe i.e. contain MZ and NEW
                            header and given file offset in old exe stub then
                            start = 0, end = begin of first data or code
                            object).
                            Return: logical number of object or 0 if no object
                            */
  unsigned      (*GetObjAttr)(unsigned long pa,char *name,unsigned cb_name,
                              unsigned long *start,unsigned long *end,int *class,int *bitness);

                         /* Prepare internal buffers for put structures
                            method. Called before GetPubSym and GetObjAttr
                            Arguments:
                            start - start position in the file, requiring for
                            structural disassembler.
                            end - final offset in the file, requiring for
                            structural disassembler.
                            Return:
                            0 - success
                            !0 - an error occured (sample: out of memory)
                            */
  int           (*prepare_structs)(unsigned long start,unsigned long end);
                        /* Clean internal buffers after structural
                           disassembler has stop */
  void          (*drop_structs)( void );
}REGISTRY_BIN;

extern REGISTRY_BIN *detectedFormat;

/***************************************************************\
* This form registry modes of translation file                  *
\***************************************************************/

typedef struct tag_REGISTRY_MODE
{
  char * name;
  char *        prompt[10];  /* on Ctrl-Fx selection */
  ModFunc       action[10];  /* action on Ctrl-Fx selection */
  int           (*detect)(void); /* detect - is file can be displayed in given
                                    mode */
  int           is_text;         /* return !0 if file is text mode */

                                 /* Paint the screen */
  unsigned      (*paint)(unsigned keycode,unsigned textshift);

                                 /* NLS codepage conversion.
                                    Return new size of blocks */
  unsigned      (*convert_cp)(char *,unsigned len);
                                 /* NLS support:
                                    return size of symbol in bytes
                                    for selected codepage */
  unsigned      (*get_symbol_size)( void );
  char *        (*misckey_name)( void ); /* F4 key name */
  void          (*misc_action)( void ); /* F4 action */
  unsigned long (*PrevPageSize)(void);  /* Get previous page size */
  unsigned long (*CurPageSize)(void);   /* Get current page size */
  unsigned long (*PrevLineWidth)( void ); /* Get previous line width */
  unsigned long (*CurLineWidth)( void ); /* Get current line width */
  void          (*help)( void );         /* display help about mode */
  void          (*read_ini)( hIniProfile * );     /* read biew.ini file if need */
  void          (*save_ini)( hIniProfile * );     /* write to biew.ini if need */
  void          (*init)( void );         /* initialize mode (constructor) */
  void          (*term)( void );         /* destroy mode (destructor) */
}REGISTRY_MODE;

extern REGISTRY_MODE *activeMode;
extern int  SelectMode( void );
extern void QuickSelectMode( void );

typedef struct tag_REGISTRY_TOOL
{
  char *        name;                /* Tool name */
  void          (*tool)( void );     /* Tool body */
  void          (*read_ini)( void ); /* read biew.ini if need */
  void          (*save_ini)( void ); /* write to biew.ini if need */
}REGISTRY_TOOL;

typedef struct tag_REGISTRY_SYSINFO
{
  char *        name;                /* System depended information name */
  void          (*sysinfo)( void );  /* System depended information body */
  void          (*read_ini)( void ); /* read biew.ini if need */
  void          (*save_ini)( void ); /* write to biew.ini if need */
}REGISTRY_SYSINFO;

#ifdef __cplusplus
}
#endif

#endif

